
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche, Thomas Genet and Erwan Houssay

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.5. Last modification 23/09/08
*)
(* ..................................................................
    Module ......: Unify
    Initial Date : 03 January 2005
    Last Update .: 07 March 2006
    Developer ...: Mathieu Turani and Max Tuengerthal
    Compiler ....: OCAML 3.07
    Comments ....: Combination algorithm for unification
    Project .....: The CL-Atse Protocol Analyser, AVISPA.
 .................................................................... *)

open Types_unify;;
open Utils_unify;;
open Unifystd;;
open Unifyxor;;
(* open Debugprinting;;    Called directly with Debugprinting.xxxx  *)


(* ==============
   Help Functions
   ============== *)


(* returns the intersection of l1 and l2
 * takes every element x in l1 and tests whether it is in l2 or not.
 * if yes x will be in the output
 *)
let intersection l1 l2 =
  it_list (fun l x -> if List.mem x l2 then x::l else l) [] l1
;;




(* ==========================
   Purification and Splitting
   ========================== *)


(* The returned substituion is like the given one, but without the assignment
 * of n, if n appears.
 * If n does not occur in ls the given substitution is returned.
 *)
let remove_from_subs n (ls,fv,ctr) =
  ((List.remove_assoc n ls),fv,ctr)
;;


(* Splits the substitution sub into three parts (std, xor, rem) with the
 * following properties:
 * - rem contains all the assignments of variables which are independend of u
 *   and v,
 *   i.e. if x -> t is in rem, then x does not appear in std, xor, u or v
 * - std contains only assignments of variables that are dependend of u and v
 *   and that belong to the standard theory,
 *   i.e. if x -> t is in std, then x does occur in u or v (maybe not directly
 *   but modulo the substitution sub) and t is of the kind Atm(n) (n<>0),
 *   SCrypt(_,_), PCrypt(_,_), Uplet(_) or PInv(_)
 * - xor contains only assignments of variables that are dependend of u and v
 *   and that belong to the xor theory,
 *   i.e. if x -> t is in std, then x does occur in u or v (maybe not directly
 *   but modulo substitution sub) and t is of the kind Atm(0) or Xor(_)
 *
 * If x -> y (y = Var(_)) is in sub then every occurance of x will be replaced
 * by y. x -> y will be stored in rem. This is used to keep std and xor small.
 * Also if x -> Atm(n) is in sub, then x will be replaced by Atm(n) in std
 * terms, if n<>0, and in xor terms, if n=0.
 *
 * Therefore what split_subst has to do is a mixture of a map and a fold on the
 * terms u and v. Because of these replacements the original u and v can change
 * and the new versions u2 and v2 are returned.
 *
 * Assumptions:
 * - u and v have to be purified and normalized
 * - u and v have to be in the same theory (a variable is encountered to be in
 *   both theories, i.e. u = Var(_) and v = Xor(_) is as well allowed as u =
 *   Var(_) and v = Uplet(_))
 *
 * Output: u2, v2, (std, xor, rem, fv, ctr), std_vars, xor_vars
 * - u2 is the possibly changed value of u
 * - v2 is the possibly changed value of v
 * - std is of type (int * term) and contains bindings of the std theory
 * - xor is of type (term list list) and contains the equations representing the
 *   bindings of xor terms.
 *   e.g. if x -> xor(y,z,u,v) is in sub and x is dependent of u, then
 *        [x,y,z,u,v] will be in xor
 * - rem is of type (int * term) and contains bindings of variables that are not
 *   dependent of u or v (may contain std and xor values)
 * - fv is the same free variable as the one given in sub
 * - ctr are the negative constraints of sub (unchanged)
 * - std_vars contains all the variables which occure in std (not only on the
 *   left hand side but also on the right)
 * - xor_vars contains all the variables which occure in xor
 *
 * Remarks:
 * - std combined with fv and some constraints is still a "valid" substitution.
 *   Valid here means in particular, that all the terms are purified and
 *   normalized. This holds as long as the folling is true:
 *   "After removing a binding from a substitution all bindings x -> t where t
 *    is a std term are still purified and normalized."
 *   This holds for std terms, but you have to be careful, because it is not
 *   true for xor terms. The problem is that the variables in xor terms are
 *   sorted. Free variables for example have to appear first and are sorted by
 *   theire index. Since variables will become free, if one removes bindings,
 *   they would have to be sorted differently.
 *
 * TODOforEXP:
 * - to make use of the exponential operator, of course, split_subst has to
 *   split the given substitution sub into four parts: std, xor, exp and rem
 *)
let split_subst (sub : t_subst) (u : term) (v : term)
  : term * term * t_split_subst * int list * int list =

  (* add_var adds n clever to (std_vs,xor_vs). If n has a type different from 0,
   * it will be added to std_vs, no matter what to_std is.
   * It will not produce duplicates in std_vs or xor_vs (insert is used).
   *
   * Input: n - an index of a variable
   *        to_std - true, if n should be added to std_vs
   *        to_xor - true, if n should be added to xor_vs
   *        std_vs - list of variables
   *        xor_vs - list of variables
   *)
  let add_var n (to_std,to_xor) (std_vs,xor_vs) =
    let to_std = to_std || (var_type n <> 0)
    in
    match (to_std,to_xor) with
    | (true ,true ) -> (insert n std_vs,insert n xor_vs)
    | (true ,false) -> (insert n std_vs,         xor_vs)
    | (false,true ) -> (         std_vs,insert n xor_vs)
    | (false,false) -> (         std_vs,         xor_vs)
  in

  (* Returns true, iff n is in std_vs or in xor_vs
   *)
  let mem_var n (std_vs,xor_vs) =
    List.mem n std_vs || List.mem n xor_vs
  in

  (* applies "f" to a list of terms "t::ts" instead of just a single term
   *)
  let rec apply_f_to_list (in_std : bool) tas = function
    | t::ts ->
        let tas, t  = f in_std tas t in
        let tas, ts = apply_f_to_list in_std tas ts in
        (tas, t::ts)
    | [] -> tas, []
  (* applies "f" to a list of terms "t::ts" instead of just a single term
   *)
  and apply_f_to_list_sign (in_std : bool) tas = function
    | (t,si)::ts ->
        let tas, t  = f in_std tas t in
        let tas, ts = apply_f_to_list_sign in_std tas ts in
        (tas, (t,si)::ts)
    | [] -> tas, []
  and
  (* f takes a term t and returns t2 which has the following properties:
   * - t2 contains only variables that are free or have a value different
   *   from a variable and an atom (except Atm(0) if t is std and Atm(n)
   *   (n<>0) if t is xor)
   * - otherwise t2 is the same as t
   *
   * In tas it collects all standard equations (in std), all xor
   * equations (in xor), all variables (in vars) and the remaining
   * substitution (in s).
   * f goes into the structure of t and collects all variables.
   * For every variable t it goes into the structure of the value of
   * this variable (if there is any). If there is a value t2 different
   * from Atm or Var it is either a standard or an xor expression and the
   * equation t = t2 will be added to std or xor, respectively.
   * The value for the variable will also be removed from s because the
   * information is now stored in std or xor and it is not necessary
   * to store it twice.
   *
   * The value of in_std is only relevant if t is a variable, otherwise it
   * is ignored. It indicates weather the appearance of t is in a std term
   * or an xor term.
   *)
  f (in_std : bool) (std,xor,vars,s as tas) (t : term) =
    match t with
    | Var(n) ->
        if mem_var n vars then
          (* n in vars means, that we have already handled it. We do not need to
           * do it twice, but if in_std is false we have to add it to xor_vs and
           * vice versa *)
          ((std,xor,(add_var n (in_std, not in_std) vars),s), t)
        else
          (* We found a variable that is not handled yet. If there is a value
           * for it in s, add this as an equation to std or xor, respectively.
           *
           * TODO: Here could be done some optimization, because most time we
           * look for the value of n in s and remove this from the substitution
           * afterwards. You could do this in one step. *)
          (match valeur n s with
          | Var(m) as t2 ->
              if m=n then
                (* m=n means, that n is free in s *)
                ((std,xor,add_var n (in_std, not in_std) vars, s), t)
              else
                (* forget that we saw Var(n) and replace it with Var(m) (= t2)
                 *)
                f in_std tas t2
          | Xor l ->
              (* add n to standard vars if in_std is true and add n to xor vars
               * in any case *)
              let vars = add_var n (in_std,true) vars in
              (* apply f to l *)
              let ((std, xor, vars, s), l) =
                apply_f_to_list false (std, xor, vars, remove_from_subs n s) l
              in
              (* add the equation 0 = t+l to xor *)
              ((std, (t::l)::xor, vars, s), t)
          | Atm(0) ->
              (* add n to standard vars if in_std is true and add n to xor vars
               * in any case *)
              let vars = add_var n (in_std,true) vars in
              (* do not replace an Atm(0), but add "0 = Var(n) (= t)" to xor *)
              ((std, [t]::xor, vars, remove_from_subs n s), t)
          | Atm(_) as t2 ->
              if in_std then
                (* there is an atom t2 assigned to n so we can replace n by t2
                 * and forget about n *)
                ((std, xor, vars, s), t2)
              else
                (* There is an atom t2 assigned to n, but because we are in an
                 * xor term we cannot replace n by t2. *)
                (* Add n to std_vars and to xor_vars. *)
                let vars = add_var n (true,true) vars in
                (* Add n -> t2 to std. *)
                (((n,t2)::std, xor, vars, remove_from_subs n s), t)
          | u ->
              (* u has to be standard. *)
              (* Add n to xor vars if in_std is false and add n to standard vars
               * in any case. *)
              let vars = add_var n (true,not in_std) vars in
              (* Apply f on u *)
              let (std,xor,vars,s),u =
                f true (std, xor, vars, remove_from_subs n s) u in
              (* Add n -> u to std. *)
              (((n,u)::std, xor, vars, s), t)
          )
    | Atm(_) ->
        tas, t
    | Xor l ->
        let tas, l = apply_f_to_list false tas l in
        tas, (Xor l)
    | Uplet l ->
        let tas, l = apply_f_to_list true tas l in
        tas, (Uplet l)
    | PCrypt(m,k) ->
        let tas,m = f true tas m in
        let tas,k = f true tas k in
        tas, PCrypt(m,k)
    | SCrypt(m,k) ->
        let tas,m = f true tas m in
        let tas,k = f true tas k in
        tas, SCrypt(m,k)
    | PInv(t) ->
        let tas,t = f true tas t in
        tas, PInv(t)
    | Exp(t,l) ->
	let tas,t = f true tas t in
	let tas,l = apply_f_to_list_sign true tas l in
	tas, Exp(t,l)
  in
  (* in_std will be true, iff u and v are both std *)
  let in_std = match (u,v) with
               | (Atm(0),_     )
               | (Xor(_),_     )
               | (_     ,Atm(0))
               | (_     ,Xor(_)) -> false
               | _               -> true
  in
  (* Apply f on u with an empty tas. *)
  let tas,u = f in_std ([],[],([],[]),sub) u in
  (* Apply f on v with the tas obtained before. *)
  let (std_ls,xor_ls,(std_vars,xor_vars),(rem_ls,fv,ctr)),v = f in_std tas v in
  u,v,(std_ls,xor_ls,rem_ls,fv,ctr),std_vars,xor_vars
;;


(* purify_pair purifies the input pair (u,v) such that u or v is a variable or
 * they are both std or both xor. It is not possible that one is a std term
 * while the other is Atm(0).
 *
 * It cannot happen that both terms are Xor[_::_] because this will be
 * simplified to one Xor[...] that equals Atm(0).
 *
 * If to_purify is true u and v are purified. Otherwise it is assumed that they
 * are already purified.
 *
 * The output can only be of one of the following shapes:
 * - Atm(0), Xor(_)
 * - Atm(0), Var(_)
 * - Atm(0), Atm(0)
 * - u     , v       where u and v are both standard or variables,
 *                   i.e. not Atm(0) and not Xor(_)
 *
 * (e.g. it is not possible that one is Var(_) and the other Xor[_::_])
 *)
let rec purify_pair (to_purify : bool) (sub : t_subst) (u,v : term * term)
  : term * term * t_subst =

  match (u,v) with
  | (Atm(0), Atm(0)) -> (u, v, sub)
  | ((Xor _ as u), (Atm 0 as v))
  | ((Atm 0 as v), (Xor _ as u)) ->
      let u, sub = if to_purify then purify sub u else u, sub in
      (v, u, sub)
  | (Xor l1, Xor l2) ->
      let t, sub = purify sub (Xor (merge l1 l2))
      in
      (match t with
      | Atm(0)
      | Xor(_)
      | Var(_) -> Atm(0), t, sub
      | _      ->
          (* we have to introduce a new variable *)
          let (newvar,sub) = add_newvalue sub (Atm(0)) in
          newvar, t, sub
      )
  | ((Xor l     ), (Atm _ as w))
  | ((Atm _ as w), (Xor l     ))
  | (Xor l       , (Var _ as w))
  | ((Var _ as w), Xor l       ) ->
      let t, sub = purify sub (Xor (w::l))
      in
      (match t with
      | Atm(0)
      | Xor(_)
      | Var(_) -> Atm(0), t, sub
      | _      ->
          (* we have to introduce a new variable *)
          let (newvar,sub) = add_newvalue sub (Atm(0)) in
          newvar, t, sub
      )
  | ((Atm(0) as v), u            )
  | (u            , (Atm(0) as v))
  | ((Xor(_) as v), u            )
  | (u            , (Xor(_) as v)) ->
      (* u has to be standard and is not a variable *)
      if to_purify then
        (* purify u and v if necessary *)
        let (u2,sub2) = purify sub  u in
        let (v2,sub3) = purify sub2 v in
        (* Call purify_pair again, this time with to_purify = false. This is
         * necessary because u or v could have changed their topsymbol. *)
        purify_pair false sub3 (u2,v2)
      else
        (* we know that u is standard and v not, so replace v by a new variable
         * x, add x -> v to the substitution and return (x,u) *)
        let (newvar,sub) = add_newvalue sub v in
        (newvar, u, sub)
  | (u,v) ->
      (* u and v are both standard or variables *)
      if to_purify then
        (* purify u and v if necessary *)
        let (u,sub) = purify sub u in
        let (v,sub) = purify sub v in
        (* Call purify_pair again, this time with to_purify = false. This is
         * necessary because u or v could have changed their topsymbol. *)
        purify_pair false sub (u,v)
      else
        (u, v, sub)
;;




(* =================
   Query a partition
   ================= *)


(* Returns the representative of n in p.
   Raises the exception Not_found if n is not in partition. *)
let rec rep (p : t_partition) (n : int) : int =
  match p with
  | c::cs -> if List.mem n c then List.hd c else rep cs n
  | [] -> raise Not_found
;;


(* Returns the list of the  representatives in p.
 * Raises the exception (Failure "hd") if p is not a valid partition,
 * i.e. if p contains an empty class.
 *)
let reps (p : t_partition) : int list =
  let rec reps_rec tas = function
    | c::cs -> reps_rec ((List.hd c)::tas) cs
    | [] -> tas
  in
  reps_rec [] p
;;


(* If ls = [1;2;3;4;...;n] it returns [[n];...;[4];[3];[2];[1]],
 * i.e. the partition of ls where each element is in its own class.
 *)
let trivial_part (ls : int list) : t_partition =
  let rec trivial_part_rec tas = function
    | x::xs -> trivial_part_rec ([x]::tas) xs
    | [] -> tas
  in
  trivial_part_rec [] ls
;;




(* ============================
   Pre Choice of Theory Indices
   ============================ *)


(* remove_from_list fulfils two functions:
 * - it tests, if n is in l
 * - it removes n from l, if it was in l
 *
 * Output:
 * - (false, rev(l) ) iff n is not in l
 * - (true , rev(l2)) iff n is in l, where l2 is produced from l by removing the
 *                                   first occurrence of n
 * (rev(l) means the reverse of list l)
 *)
let remove_from_list n l =
  let rec remove_from_list_rec tas = function
    | m::ms -> if m=n then
                 (true, merge tas ms)
               else
                 remove_from_list_rec (m::tas) ms
    | [] -> (false, tas)
  in
  remove_from_list_rec [] l
;;


(* pre_choice_std takes the shared variables (shared_vars) and splits it into
 * three parts std, xor and uncertain.
 *
 * std will contain all the variables that have to be variables in the std
 * theory, otherwise unification will fail in the std theory immediately.
 *
 * xor will contain the variables that do not change the solution of the std
 * theory. These are basically the variables that are free in std_ls and do not
 * occure in bindings like x -> PInv(y).
 *
 * uncertain will contain the variables where it is unsure to which theory they
 * will belong.
 *
 * Assumptions:
 * - std_ls must not contain n -> PInv(Var(m)) and n2 -> PInv(Var(m)),
 *   if n <> n2 and both are in shared_vars.
 *
 * Remarks:
 * - All variables that have a type different from 0 will be in std.
 * - uncertain is organized as a (int * int) list, because it is always the case
 *   that two uncertain variables are dependent to each other. Only one of them
 *   can be xor, not both.
 * - std, xor and uncertain are a partition of shared_vars
 *)
let pre_choice_std (shared_vars : int list) (std_ls : t_equal)
  : int list * int list * (int * int) list =

  (* A help function which will collect std, xor, uncertain and not_visited as
   * tas.
   * not_visited contains all the variables of shared_vars, that are not already
   * put into std, xor or uncertain. So, they four build a partition of
   * shared_vars
   *)
  let rec pre_choice_std_rec (std, xor, uncertain, not_visited as tas) =
    function
    | (n,t)::tail ->
        (* Test if n is in not_visited. In fact it cannot happen that n is not
         * in not_visited and it is in std, xor or uncertain. This is due to the
         * fact that std_ls must not contain two values for a variable.
         * What is acctually tested here is, if n is a shared variable. *)
        let (n_is_shared, not_visited_without_n) =
          remove_from_list n not_visited
        in
        if n_is_shared then
          match t with
          | Var(_) ->
              (* A binding x -> y can happen because of variable identification.
               * Act as if n is free. *)
              pre_choice_std_rec tas tail
          | PInv(Var(mm)) ->
              (* mm is not neccesarily free *)
              (match valeur_elem mm std_ls with
              | Var(m) ->
                  (* m is free in std_ls *)
                  (* Test if m is a shared variable *)
                  let (m_is_shared, not_visited_without_nm) =
                    remove_from_list m not_visited_without_n
                  in
                  if m_is_shared then
                    (* Only if n and m are shared we add them to uncertain.
                     * It is possible to remove m from not_visited, because, due
                     * to assumptions, it is not allowed, that there is another
                     * rule n2 -> PInv(m) in std_ls with n2 in shared_vars. *)
                    pre_choice_std_rec
                      (std, xor, (n,m)::uncertain, not_visited_without_nm) tail
                  else
                    (* m is not shared *)
                    (* Test, if it is possible that m and n change the sides. *)
                    if acceptable_value ([],FreeVar 0,[]) m (PInv(Var(n))) then
                      (* It is possible that n and m change sides, so we can say
                       * that n is xor, because m is not shared and therefore
                       * std. *)
                      pre_choice_std_rec
                        (std, n::xor, uncertain, not_visited_without_n) tail
                    else
                      (* It is not possible that n and m change sides, because
                       * of their types. So, n has to be std. *)
                      pre_choice_std_rec
                        (n::std, xor, uncertain, not_visited_without_n) tail
              | _ ->
                  (* m is not free in std_ls and cannot be PInv(_), due to
                   * normalisation, so n has to be standard. *)
                  pre_choice_std_rec
                    (n::std, xor, uncertain, not_visited_without_n) tail
              )
          | _ ->
              (* t is a standard term with a top symbol that cannot vanish,
               * i.e. n has to be standard *)
              pre_choice_std_rec
                (n::std, xor, uncertain, not_visited_without_n) tail
        else
          (* n is not in not_visited, i.e. it is not in shared_vars,
           * i.e. ignore it *)
          pre_choice_std_rec tas tail
    | [] -> tas
  in

  (* distributes variables with type 0 to xor and the others to std
   *)
  let rec distribute std xor = function
    | n::ns -> if var_type n = 0 then
                 distribute std (n::xor) ns
               else
                 distribute (n::std) xor ns
    | [] -> (std, xor)
  in

  (* apply pre_choice_std_rec with an initial tas, where not_visited is
   * shared_vars. *)
  let (std, xor, uncertain, not_visited) =
    pre_choice_std_rec ([],[],[],shared_vars) std_ls
  in
  (* All variables that are not_visited are free in std_ls and do not appear in
   * equations like x -> PInv(y) or if they do appear in such equations they
   * cannot change the side because of their types.
   * We can say that all variables in not_visited that have type 0 belong to
   * xor, because this will not change the std solution. *)
  let (std, xor) = distribute std xor not_visited
  in
  (std, xor, uncertain)
;;


(* Assumptions:
 * - ls contains only Var terms
 *
 * Output:
 * - None  , iff ls has more than one variable not in no_vars or ls is empty
 * - Some x, iff ls has exactly one variable x not in no_vars
 *
 * If ls contains no variables (not in no_vars) and is not empty the exception
 * No_Solution is raised.
 *)
let get_onlyvar (no_vars : int list) (ls : term list) : int option =

  (* foundvar indicates, if we already found a variable and which. It does not
   * matter what is the int value, if the bool value is false. *)
  let rec get_onlyvar_rec (foundvar : bool * int) = function
    | Var(n)::tail ->
        if List.mem n no_vars then
          (* treat Var(n) like an atom *)
          get_onlyvar_rec foundvar tail
        else
          if fst foundvar then
            (* at least two variables not in no_vars found *)
            None
          else
            get_onlyvar_rec (true,n) tail
    | _::tail -> get_onlyvar_rec foundvar tail
    | [] ->
        if fst foundvar then
          (* only one variable not in no_vars found *)
          Some (snd foundvar)
        else
          (* no variables not in no_vars found *)
          raise No_Solution
  in

  if ls = [] then
    None
  else
    get_onlyvar_rec (false,-1) ls
;;


(* Tests, if (x,y) or (y,x) is in ls and returns (y,rev(ls2)) where ls2 is ls
 * without the first occurence of the binding (x,y) or (y,x).
 *
 * Raises the exception Not_found, if x is neither on the left hand side, nor on
 * the right.
 *)
let lookup_bidir_remove (x : 'a) (ls : ('a * 'a) list) : 'a * ('a * 'a) list =
  let rec lookup_bidir_remove_rec tas = function
    | (y,z as p)::tail ->
        if x=y then (z, merge tas tail)
        else
          if x=z then (y, merge tas tail)
          else lookup_bidir_remove_rec (p::tas) tail
    | [] -> raise Not_found
  in
  lookup_bidir_remove_rec [] ls
;;


(* pre_choice_xor is used to limit the possible choices of theory indices. It
 * takes the xor equations xor_eqns and the a set of variables that is already
 * split in three parts: std, xor and uncertain.
 *
 * The xor equations are analysed to find contradictions and raise the exception
 * No_Solution and to try to decide the choice for uncertain variables.
 *
 * Raises the exception No_Solution, if a variable is found, that has to be std
 * and xor at the same time, or if there are xor eqns with only constants.
 *)
let pre_choice_xor std_xor_uncertain (xor_eqns : term list list)
  : int list * int list * (int * int) list =

  let rec  pre_choice_xor_rec (std, xor, uncertain as sxu) = function
    | ls::lss ->
        (match get_onlyvar std ls with
        | Some n ->
            (* n is the only variable in ls, so n has to be xor *)
            (try
              let (m, uncertain_without_nm) = lookup_bidir_remove n uncertain in
              (* n has to be xor, so m has to be std *)
              (* Add n to xor and m to std.
               * Start from the beginning (xor_eqns), because adding m to std
               * can effect equations that were already visited.
               *
               * TODO: check if it is better to run through all eqns first and
               * check at the end if there were any changes and if yes then run
               * again over all eqns, or if this is not necessary and proceed
               * with lss, although we could possibly eliminate more choices, if
               * we start from beginning again. *)
              pre_choice_xor_rec (m::std, n::xor, uncertain_without_nm) xor_eqns
            with
              Not_found ->
                (* n is not in uncertain *)
                if List.mem n std then
                  (* n is in std, but it has to be in xor as well *)
                  raise No_Solution
                else
                  (* n is not in uncertain and not in std, so n is already in
                   * xor or n is not a shared variable. *)
                  pre_choice_xor_rec sxu lss
            )
        | None ->
            (* There are more than one variable in ls or ls is [], so we cannot
             * easily derive information from this *)
            pre_choice_xor_rec sxu lss
        )
    | [] -> sxu
  in
  pre_choice_xor_rec std_xor_uncertain xor_eqns
;;


(* pre_choice_of_thy is used to limit the Pre Choice of Theory Indices step. It
 * tries to determine the theorie index for as many variables as possible.
 *
 * Assumptions:
 * - std_ls contains no variable identifications between shared variables,
 *   i.e. x -> Var(y) in std_ls is not allowed if x and y are both in
 *   shared_vars.
 * - std_ls does not contain any xor terms (Atm(0) or Xor(_))
 * - xor_eqns contains only Atm or Var terms
 *
 * Output: (std, xor, uncertain)
 * std - contains variables that have to be standard,
 *       otherwise unification in the standard theory will fail
 * xor - contains variables that have to be xor, otherwise unification in the
 *       xor theory will fail
 *       and variables that would not change std_ls if they would not be std
 * uncertain - contains the rest, i.e. uncertain = (shared_vars \ std) \ xor
 *
 * std, xor and uncertain are pairwise disjoint.
 *
 * If unification is not possible in any case, e.g. because a variable has to be
 * in std and xor at the same time, then the exception No_Solution is raised.
 *
 *
 * TODOforEXP:
 * - pre_choice_xor is not anymore needed, if we replace it by exponential.
 * - you could provide another function, like pre_choice_exp, to limit the
 *   possible choices of theory indices and to find contradictions earlier.
 *   But this is only optimisation and it would also work without it.
 *)
let pre_choice_of_thy (xor_eqns : term list list) (std_ls : t_equal)
  (shared_vars : int list) : int list * int list * (int * int) list =

  (* do pre choice of theory with std_ls *)
  let std_xor_uncertain = pre_choice_std shared_vars std_ls in

  (* do pre choice of theory with xor_eqns *)
  (* we could think about skipping this, in most of the cases this does not
   * help to limit the possible choices *)
  let std_xor_uncertain = pre_choice_xor std_xor_uncertain xor_eqns in

  std_xor_uncertain
;;




(* ========================
   Choice of Theory Indices
   ======================== *)


(* For every pair (x,y) in uncertain it is choosen either that x belongs to std
 * and y belongs to xor, or the other way around.
 *
 * E.g.:
 * choose_theory [7;8] [9] [(1,2);(3,4)] = [([3;1;7;8], [4;2;9]);
 *                                          ([4;1;7;8], [3;2;9]);
 *                                          ([3;2;7;8], [4;1;9]);
 *                                          ([4;2;7;8], [3;1;9])]
 *)
let choose_theory (std : 'a list) (xor : 'a list) (uncertain : ('a * 'a) list)
  : ('a list * 'a list) list =

  let rec choose_theory_rec tas (l1,l2) = function
    | (n,m)::tail ->
        let tas2 = choose_theory_rec tas  (n::l1, m::l2) tail in
        let tas3 = choose_theory_rec tas2 (m::l1, n::l2) tail in
        tas3
    | [] -> (l1,l2)::tas
  in
  choose_theory_rec [] (std, xor) uncertain
;;


(* apply_choice_std modifies the substitution std_ls such that it contains no
 * binding x -> t, where x is in xor_vars.
 *
 * Assumptions:
 * - xor_vars has to be choosen in such a way that it is possible to turn a
 *   binding like x -> PInv(y) into y -> PInv(x), if x is in xor_vars.
 *)
let rec apply_choice_std vi (xor_vars : int list) (std_ls : t_equal) : t_equal =

  (* remove bindings like n -> PInv(Var(m)) from std_ls if n is an xor variable
   *)
  let rec remove_bindings tas to_add = function
    | (n,PInv(Var(m)) as b)::tail ->
        let nn = try rep vi n with Not_found -> n in
        if List.mem nn xor_vars then
          (* n is xor, so remove b and add m -> PInv(Var(n)) to to_add *)
          match valeur_elem m std_ls with
          | Var(mm) ->
              (* do not replace n by nn, nn is only used to test, if n is in
               * xor_vars *)
              remove_bindings tas ((mm,PInv(Var(n)))::to_add) tail
          | _       -> failwith "apply_choice_std: assumptions do not hold"
        else
          (* do not remove b *)
          remove_bindings (b::tas) to_add tail
    | (n,Var(m) as b)::tail ->
        let nn = try rep vi n with Not_found -> n in
        if List.mem nn xor_vars then
          (* do not replace n by nn, nn is only used to test, if n is in
           * xor_vars *)
          remove_bindings tas ((m,Var(n))::to_add) tail
        else
          remove_bindings (b::tas) to_add tail
    | b::tail -> remove_bindings (b::tas) to_add tail
    | [] -> (tas, to_add)
  in

  (* adds (n,t) to std_ls using add_elem for every (n,t) in the input
   *)
  let rec add_bindings (std_ls : t_equal) : t_equal -> t_equal = function
    | (n,t)::tail ->
        (try
          let (std_ls,_,_) =
            add_elem true(*false*) n t (std_ls,FreeVar(0),[])
          in
          add_bindings std_ls tail
        with
	  | No_Solution -> failwith ("apply_choice_std: "^
                           "add_elem raised No_Solution, which must not happen\n"^
			   "Add var "^(Debugprinting.str_term_dbg !localtab (Var n))^
			   " with value "^(Debugprinting.str_term_dbg !localtab t)^" to subst \n"^
			   (Debugprinting.str_sub_dbg !localtab (std_ls,FreeVar(0),[]))
				    ) 
	  | NeedXorUnification(_,_,_) -> failwith ("apply_choice_std: "^
                           "add_elem raised NeedXorUnification, which must not happen") 
          | _ -> failwith ("apply_choice_std: "^                                  
                           "add_elem raised an exception, which must not happen") 
        )
    | [] -> std_ls
  in

  (* first remove the bindings that have to be changed *)
  let (std_ls, to_add) = remove_bindings [] [] std_ls in
  (* then add the changed bindings again *)
  add_bindings std_ls to_add
;;




(* ========================
   Extract Partial Ordering
   ======================== *)


(* To a given substitution ls get_po generates a partial order, such that
 * x is a subterm of y, iff x < y.
 *
 * If ls contains bindings like n -> Var(m) then these are ignored. This is due
 * to the assumption, that in this case n and m have to be identified anyway.
 *
 * If ls contains bindings like n -> t and n is not a representative, i.e. n is
 * not the first element in its class in vi, then this binding is ignored.
 *
 * Therefore we have to have the following assumptions:
 *
 * Assumptions:
 * - vi should be a valid identification with respect to ls, i.e. if n and m are
 *   equal, then either n -> m (resp. m -> n) is in ls and m (resp. n) is free
 *   in ls, or n -> t and m -> t' are in ls and t is equal t' (modulo ls)
 *   (m free in ls means that there must not be a binding like m -> t in ls)
 * - if n -> Var(m) is in ls, then n and m have to be identified (i.e. they are
 *   in the same equivalence class in vi)
 * - if n -> Var(m) is in ls, then m has to be free in ls, i.e. there must not
 *   be a binding like m -> t' in ls.
 *
 * From this assumption it follows that if n and m are idenified, either
 * n -> m is in ls and m is free in ls, or n -> t and m -> t' is in ls and t
 * equals t' (modulo ls), because vi is a valid identification.
 *
 * This explains, why we can ignore bindings like n -> t, if n is not a
 * representative.
 *)
let get_po vi (ls : t_equal) : t_po =

  (* help takes a partial order po, a variable n and a term list.
   * For every variable m in the term list m < n is added to po.
   *)
  let rec help po n : term list -> t_po = function
    | Atm(_)     ::tail -> help po n tail
    | Var(m)     ::tail ->
        (* get the representative of m *)
        let mm = try rep vi m with Not_found -> m in
        (* mm is a subterm of n, so add mm < n to po *)
        help (add_to_po po mm n) n tail
    | Uplet(ll)  ::tail
    | Xor(ll)    ::tail -> help po n (merge tail ll)
    | PCrypt(t,k)::tail
    | SCrypt(t,k)::tail -> help po n (t::k::tail)
    | PInv(t)    ::tail -> help po n (t::tail)
    | Exp(t,l)   ::tail -> help po n (t::(merge_first tail l))
    | [] -> po
  in

  (* help2 applies help on every binding in (n,t)::tail, but not on
   * identifications.
   *)
  let rec help2 po = function
    | (_,Var(_))::tail ->
        (* skip identifications *)
        help2 po tail
    | (n,t)::tail ->
        (* get the representative of n *)
        let nn = try rep vi n with Not_found -> n in
        if n=nn then
          (* n is a representative, so add unequations to po *)
          help2 (help po nn [t]) tail
        else
          (* skip (n,t), because n is not a representative *)
          help2 po tail
    | [] -> po
  in

  help2 [] ls
;;




(* ========================================
   Combination of STD and XOR Substitutions
   ======================================== *)


(* Removes bindings like (n, t) from (n,b)::tail, if n is not a representative
 * (i.e. if n is not the first element its class in var_identif).
 * It also removes bindings like (n,Var(m)) if n and m are in the same
 * equivalence class.
 *)
let rec remove_identif var_identif tas = function
  | (n,Var(m) as b)::tail ->
      (try
        if rep var_identif n = rep var_identif m then
          (* remove binding, because n and m are identified *)
          remove_identif var_identif tas tail
        else
          remove_identif var_identif (b::tas) tail
      with
        Not_found -> remove_identif var_identif (b::tas) tail
      )
  | (n,_ as b)::tail ->
      (try
        if rep var_identif n = n then
          remove_identif var_identif (b::tas) tail
        else
          (* remove binding, because n is not a represantative *)
          remove_identif var_identif tas tail
      with
        Not_found -> remove_identif var_identif (b::tas) tail
      )
  | [] -> tas
;;


(* combine_subs combines std_ls, xor_ls and rem_ls to one substitution.
 *
 * The exception No_Solution is raised, if the negative constraints ctr do not
 * allow the combined substitution.
 * 
 * TODOforEXP:
 * - xor_ls is of type t_equal here, because the xor theory is unitary, i.e.
 *   unif_xor produces a mgu.
 * - The exponential theory is not unitary, i.e. the unif alg for it produces a
 *   complete set of unifiers. So, the type of exp_subs would have to be
 *   something like t_equal list.
 * - also the type of combine_subs would change to t_subst list
 * - what has to be done then is 
 *   - combining rem_ls and std_ls, like before
 *   - combining every exp_ls of exp_subs with the combination of rem and std,
 *     finally add ctr.
 *     It is not clear to me now, what has to be done if some combinations fail,
 *     because of adding ctr and others not.
 *)
let combine_subs (std_ls : t_equal) (xor_ls : t_equal) (rem_ls : t_equal) fv ctr
  : t_subst =

  (* add_help adds (n,t) to sub. If to_purify is false it is assumed that t is
   * purified and normalized with respect to sub.
   *)
  let add_help to_purify sub (n,t) = add_elem to_purify n t sub
  in

  (* This could raise the exception No_Solution or NeedXorUnification but it
   * must not due to assumptions.
   *)
  let combined_sub =
    (try
      (* std_ls is purified and normalized with respect to std_ls, because no
       * variable in std_ls is instantiated in rem_ls. Hence, we can use
       * (add_help false).
       * But the xor terms in xor_ls are not yet normalized, so we have to use
       * (add_help true) here. *)
      it_list (add_help true)
              (it_list (add_help false) (rem_ls,fv,[]) std_ls)
              xor_ls
    with
    | No_Solution ->
        failwith ("combine_subs: combining subs failed in add_elem with the "^
                  "exception No_Solution")
    | NeedXorUnification _ ->
        failwith ("combine_subs: combining subs failed in add_elem with the "^
                  "exception NeedXorUnification")
    )
  in

  (* And finally add the negative constraints ctr.
   * This may raise the exception No_Solution. *)
  add_ctr combined_sub ctr
;;




(* ==========================================
   Variable Identification Constraint Testing
   ========================================== *)


(* is_allowed returns true, iff it is allowed to identify n and m according to
 * the negative constraint neg_ctrs.
 *
 * Assumptions:
 * - n and m are both representatives
 *)
let is_allowed vi (n : int) (m : int) (neg_ctrs, _ : t_vi_ctrs) =
  let rec is_allowed_rec = function
    | (nn,mm)::tail ->
        (* TODO:
         * could be done faster, by looking up nn and mm at the same time *)
        let nn = rep vi nn in
        let mm = rep vi mm in
        if (nn=n && mm=m) || (nn=m && mm=n) then
          false
        else
          is_allowed_rec tail
    | [] -> true
  in
  is_allowed_rec neg_ctrs
;;


(* returns true, iff l1 is a subset of l2 *)
let is_subset l1 l2 =
  List.for_all (fun x -> List.mem x l2) l1
;;


(* returns true, iff p1 is equal to (modulo order) or more general than p2
 *)
let is_more_general_eq (p1 : t_partition) (p2 : t_partition) =
  List.for_all (fun x -> List.exists (is_subset x) p2) p1
;;


(* returns true, iff there is a partition p2 in ps that is equal to
 * (modulo order) or more general than p
 *)
let exists_more_general_eq (p : t_partition) (ps : t_partition list) =
  List.exists (fun p2 -> is_more_general_eq p2 p) ps
;;


(* is_allowed_vi returns true, iff vi is an allowed variable identification
 * according to the layer constraint layer_ctrs.
 *
 * A variable identification is allowed by layer_ctrs, iff it is not less
 * general than or equivalent to a variable identification in layer_ctrs.
 *)
let rec is_allowed_vi vi (_,layer_ctrs : t_vi_ctrs) =
  not (exists_more_general_eq vi layer_ctrs)
;;


(* is_allowed_sub returns true, iff ls does not identify variables, that are not
 * allowed by the negative constraints neg_ctrs.
 *)
let is_allowed_sub (neg_ctrs,_ : t_vi_ctrs) ls =
  (* a dummy substitution, to apply egal *)
  let sub = (ls,FreeVar 0,[])
  in
  let rec is_allowed_sub_rec = function
    | (n,m)::tail ->
        (* n and m are not allowed to be equal *)
        if egal sub (Var n) (Var m) then
          (* n and m must not be equal, but they are, so ls is not an allowed
           * substitution. *)
          false
        else
          is_allowed_sub_rec tail
    | [] -> true
  in
  is_allowed_sub_rec neg_ctrs
;;




(* ===========================
   Pre Variable Identification
   =========================== *)


(* Pre Variable Identification is used to determine variables, that have the
 * same value. If variables x and y are found with x equal y (according to
 * egal) in std_ls then their equivalence classes in var_identif are
 * combined.
 *
 * This is an optimisation, because we can then find the right variable
 * identification earlier. The combination alg would also work without this
 * step, except that e.g. pre_choice_of_thy or get_po have assumptions on the
 * partition. If two variables are equal, they have to be in one equivalence
 * class.
 *
 * If two variables are equal, but it is not allowed by vi_ctrs, the exception
 * Invalid_Var_Identif is raised.
 *)
let pre_identification (vi_ctrs : t_vi_ctrs) (std_ls : t_equal)
  (var_identif : t_partition) : t_partition =

  (* a dummy substitution, to apply egal *)
  let sub = (std_ls,FreeVar 0,[])
  in
  
  (* ns are the elements that are equivalent to n, but not n,
   * i.e. the equivalence class of n is (n::ns) *)
  let rec help n ns tas = function
    | ms::tail ->
        (* m is the representativ of the class ms *)
        let m = List.hd ms in
        if egal sub (Var(n)) (Var(m)) then
          (* combine the classes of n and m, if n=m does not violates vi_ctr,
           * the class of m is ms and the class of n is n::ns *)
          if is_allowed var_identif n m vi_ctrs then
            help n (merge ns ms) tas tail
          else
            raise Invalid_Var_Identif
        else
           help n ns (ms::tas) tail
    | [] -> ((n::ns),tas)
  in

  let rec help2 tas = function
    | (n::ns)::tail ->
        let (new_nns,new_tail) = help n ns [] tail in
        help2 (new_nns::tas) new_tail
    | [] -> tas
    | _ -> failwith ("pre_identification: var_identif is not a valid partition"^
                     ", it contains an empty class")
  in

  let var_identif = help2 [] var_identif
  in
  (* test, if var_identif is a valid variable identification, according to
   * vi_ctrs. *)
  if is_allowed_vi var_identif vi_ctrs then
    var_identif
  else
    raise Invalid_Var_Identif
;;




(* =======================
   Variable Identification
   ======================= *)


(* If x, y are two identified variables in the input partition,
 * the equation 0 = x + y is added to xor_eqns and this combined equation system
 * is returned.
 *)
let rec add_identification_ctr_xor xor_eqns : t_partition -> term list list =
  let rec help n xor_eqns = function
    | m::ms -> help n ([Var(m);Var(n)]::xor_eqns) ms
    | []    -> xor_eqns
  in
  function
    | (n::ns)::tail -> add_identification_ctr_xor (help n xor_eqns ns) tail
    | [] -> xor_eqns
    | _ -> failwith ("add_identification_ctr_xor: input is not a valid "^
                     "partition, it contains an empty class")
;;


(* combine_classes combines the equivalence classes of the two given
 * representatives m and n.
 *)
let combine_classes (p : t_partition) m n =
  let rec help2 n xs tas = function
    | (y::_ as ys)::tail ->
        if y=n then
          (* combine xs and ys *)
          merge ((merge xs ys)::tas) tail
        else
          help2 n xs (ys::tas) tail
    | [] -> failwith "combine_classes: m and n have to be representatives in p"
    | _  -> failwith ("combine_classes: p is not a valid partition, it "^
                      "contains an empty class")
  in
  let rec help m n tas = function
    | (x::_ as xs)::tail ->
        if x=m then
          help2 n xs tas tail
        else
          if x=n then
            help2 m xs tas tail
          else
            help m n (xs::tas) tail
    | [] -> failwith "combine_classes: m and n have to be representatives in p"
    | _  -> failwith ("combine_classes: p is not a valid partition, it "^
                      "contains an empty class")
  in
  help m n [] p
;;


(* For each two classes with representatives x and y in var_identif the
 * following is done:
 * - test, if it is allowed (by vi_ctrs) that x equals y,
 *   if not: proceed with the next pair of classes
 * - combine the classes of x and y
 * - apply unif_std, to add the constraint x = y to std_subs, obtain a set of
 *   substitutions std_subs2,
 *   if this failes (i.e. std_subs2 is empty): proceed with the next pair of
 *   classes
 * - add std_subs2 to tas
 * - proceed with the next pair of classes
 *
 * Remarks:
 * - If there is only one class in var_identif then tas is returned unchanged.
 *)
let identify_variables unif_std tas (std_subs : t_subst list)
  (var_identif : t_partition) (vi_ctrs : t_vi_ctrs)
  : (t_subst list * t_partition * (int * int) list) list =

  (* adds the constraint, that n must not be equal to m to the vi constraints.
   *)
  let add_vi_ctr ((equations, vis) : t_vi_ctrs) (n : int) (m : int) =
    ((n,m)::equations, vis)
  in

  let rec help n tas (vi_ctrs : t_vi_ctrs) = function
    | (m::_)::tail ->
        (* identify n and m, if it is allowed by vi_ctrs *)
        (* test, if n equal m is allowed by the negative ctrs *)
        (if is_allowed var_identif n m vi_ctrs then
          (* combine the classes of n and m *)
          let var_identif = combine_classes var_identif m n
          in
          (* test, if var_identif is allowed by the layer ctrs *)
          (if is_allowed_vi var_identif vi_ctrs then
            (* all requirements are fulfilled, var_identif is a variable
             * identification, that is not forbidden by vi_ctrs *)
            let std_subs2 = unif_std false (Var(m)) (Var(n)) std_subs
            in
            (if std_subs2 = [] then
              (* identifing m and n failed in the std theory *)
              help n tas (add_vi_ctr vi_ctrs n m) tail
            else
              let std_subs2_vi = (std_subs2, var_identif, fst vi_ctrs)
              in
              (* Add n <> m to the identification constraints,
               * because we do not want to test the same var identifications
               * twice. *)
              help n (std_subs2_vi::tas) (add_vi_ctr vi_ctrs n m) tail
            )
          else
            (* n must not be equal to m, proceed with tail *)
            help n tas vi_ctrs tail
          )
        else
          (* n must not be equal to m, proceed with tail *)
          help n tas vi_ctrs tail
        )
    | [] -> (tas, vi_ctrs)
    | _  -> failwith ("identify_variables: var_identif is not a valid "^
                      "partition, it contains an empty class")
  in

  let rec help2 tas vi_ctrs = function
    | (n::_)::tail ->
        let (tas,vi_ctrs2) = help n tas vi_ctrs tail in
        help2 tas vi_ctrs2 tail
    | [] -> tas
    | _  -> failwith ("identify_variables: var_identif is not a valid "^
                      "partition, it contains an empty class")
  in

  help2 tas vi_ctrs var_identif
;;




(* =====================
   Combination Algorithm
   ===================== *)


(* Input:
 * - (std_vars,xor_vars) is a choice of theory indices, each variable in
 *   std_vars is handled like an constant in the xor theory and each variable in
 *   xor_vars as an constant in the std theory.
 *
 * - xor_eqns is the xor unification problem.
 *
 * - std_ls is one solution for the std unification problem.
 *
 * - rem_ls and ctr are the remaining parts from the originally given
 *   substitution, which are not relevant for solving the active unification
 *   problem. (Except that ctr might declare some substitutions as invalid.)
 *
 * - unify_nonstd is the unification alg for the xor theory
 *
 * - vi_ctrs contains the constraints for var identifications, for that we
 *   already have unifiers.
 *
 *
 * Remarks:
 * - Raises No_Solution, if there is no solution of xor_eqns for this choice of
 *   theory.
 * - Raises Invalid_Var_Identif, if unif_nonstd identifies variables that are
 *   not allowed to be identified.
 *
 *
 * TODOforEXP:
 * - unif_nonstd has to be a unification alg for the exp theory.
 *   It will return a list of ls, so combine_subs has to be able to handle this.
 *)
let rec solve_xor_eqns
   unif_nonstd
  (std_ls,xor_eqns,rem_ls,fv,ctr              )
  (var_identif                   : t_partition)
  (vi_ctrs                       : t_vi_ctrs  )
  (std_vars, xor_vars                         )
                                 : t_subst    =

  (* apply the choice of theory to the std substitution std_ls *)
  let std_ls = apply_choice_std var_identif xor_vars std_ls
  in
  (* extract partial order on variables from std_ls *)
  let po = get_po var_identif std_ls
  in

  (* Try to solve xor_eqns with the partial order po.
   * This may raise the exception No_Solution. *)
  let xor_ls = unif_nonstd xor_eqns po std_vars
  in

  (* test, if xor_ls identifies variables that are not allowed to be identified,
   * by vi_ctrs *)
  (if is_allowed_sub vi_ctrs xor_ls then
    (* remove the identification constraints from xor_ls, before combining
     * xor_ls and std_ls *)
    let xor_ls = remove_identif var_identif [] xor_ls
    in

    (* combining the substitutions possibly raises No_Solution, namely
     * if the negative constraints ctr do not allow this substitution. *)
    let combined_sub = combine_subs std_ls xor_ls rem_ls fv ctr in

    combined_sub
  else
    (* The solution we get from unif_nonstd identifies variables, that are not
     * allowed to be identified. This means, that we would get this solution
     * also somewhere else. So, this variable identification is invalid. *)
    raise Invalid_Var_Identif
  )
;;


(* Input:
 * - The last argument (choice::choices) is a list of choices of theory indices
 *   for the shared variables.
 *
 * Description:
 * - Try to solve the xor problem with the first choice of theory (choice).
 *
 *   If this suceeds, we do not have to check other possible choices, because
 *   they would only produce equivalent or less general unifiers.
 *
 *   If it failes, there are two possibilities, how it failed:
 *   - Invalid_Var_Identif: ignore the choice, because it lead to a invalid
 *     variable identification, and procced with the other choices
 *   - No_Solution: we can not solve the xor problem with this choice, so
 *     proceed with the other choices. Also remember, that at least one choice
 *     of theory produced No_Solution.
 *
 * - If all choices failed, i.e. if the last argument is [], the boolean
 *   no_solution indicates, if one of the choices produced No_Solution or if all
 *   choices produced Invalid_Var_Identif.
 *   If at least one produced No_Solution, raise the exception No_Solution,
 *   otherwise raise the exception Invalid_Var_Identif.
 *)
let rec solve_xor_eqns_choices unif_nonstd (no_solution : bool)
  (std_ls,xor_eqns,rem_ls,fv,ctr) (var_identif : t_partition) vi_ctrs
  : (int list * int list) list -> t_subst =

  function
    | choice::choices ->
        (try
          (* If we find a solution for this choice of theory indices, we do not
           * need to check other choices of variable indices, because they would
           * only lead to equivalent solutions. *)
          solve_xor_eqns unif_nonstd
                         (std_ls,xor_eqns,rem_ls,fv,ctr)
                         var_identif vi_ctrs choice
        with
        | Invalid_Var_Identif ->
            (* Trying this choice of theory indices failed. Try the others. *)
            solve_xor_eqns_choices unif_nonstd no_solution
                                   (std_ls,xor_eqns,rem_ls,fv,ctr)
                                   var_identif vi_ctrs choices
        | No_Solution ->
            (* No solution for this choice of theory indices, try the other
             * ones, but remember, that at least one failed with No_Solution. *)
            solve_xor_eqns_choices unif_nonstd true
                                   (std_ls,xor_eqns,rem_ls,fv,ctr)
                                   var_identif vi_ctrs choices
        )
    | [] ->
        (* If at least one choice failed with No_Solution, then raise
         * No_Solution.
         * If all choices failed with Invalid_Var_Identif, then raise
         * Invalid_Var_Identif. *)
        if no_solution then
          raise No_Solution
        else
          raise Invalid_Var_Identif
;;


(* Input:
 * - The last argument (std_sub::tail) is a list of std substitutions
 *
 * - vi_ctrs are the constraints for the variable identification, both layer and
 *   negative constraints.
 *
 * - tas is a list, where the combined unifiers are collected.
 *
 * - failed is a list, where the std substitutions, which failed, are collected.
 *
 * - vi is the active variable identification.
 *
 *
 * Descritption:
 * - For each std substitution, try to solve the xor problem (xor_eqns)
 *
 *   If this suceeds (i.e. no exception) variable_identif_single returns the
 *   combined substitution. Add it to tas.
 *
 *   It can fail because of two reasons:
 *   - Invalid_Var_Identif: solving xor_eqns failed because of vi constraints,
 *     this means, that we have already solutions for this case and we can just
 *     skip this std substitution. Do not add it to failed.
 *   - No_Solution: solving xor_eqns failed, but not because of vi constraints.
 *     So, add std substitution to failed.
 *)
let rec process_layer_elem unif_std unif_nonstd vi vi_ctrs xor_eqns rem_ls
  ctr (tas : t_subst list) (failed : t_subst list) =

  function
  | (std_ls,fv,_ as std_sub)::tail ->
      (try
        (* Pre Identification, i.e. identify variables that have already the
         * same value, i.e. identify x and y where egal returns true.
         *
         * pre_identification raises Invalid_Var_Identif, if this var
         * identification is not allowed, according to vi_ctrs. *)
        let var_identif = pre_identification vi_ctrs std_ls vi
        in

        (* Add the constraints for the pre identification to xor_eqns, i.e.
         * if x and y are identified, add 0 = x + y to xor_eqns. *)
        let xor_eqns = add_identification_ctr_xor xor_eqns var_identif
        in

        (* Pre Choice of Theory.
         * Possibly raises No_Solution. *)
        let (std_vars, xor_vars, uncertain_vars) =
          pre_choice_of_thy xor_eqns std_ls (reps var_identif)
        in

        (* Choice of Theory.
         * choose_theory generates a list of all possibilities. *)
        let choices_of_theory = choose_theory std_vars xor_vars uncertain_vars
        in

        (* Solve the xor problem with the choices of theories.
         * If we can compute a substitution for at least one choice of theory,
         * we know that all subst for other choices of theories are equivalent.
         * So, just keep one.
         *
         * Possibly raises No_Solution or Invalid_Var_Identif. *)
        let sub = solve_xor_eqns_choices unif_nonstd false
          (std_ls,xor_eqns,rem_ls,fv,ctr) var_identif vi_ctrs choices_of_theory
        in

        (* add sub to tas and proceed with tail *)
        process_layer_elem unif_std unif_nonstd
          vi vi_ctrs xor_eqns rem_ls ctr (sub::tas) failed tail
      with
      | Invalid_Var_Identif ->
          (* This variable identification is invalid for std_sub and vi_ctrs *)
          (* just skip this vi for this std_sub and proceed with tail *)
          process_layer_elem unif_std unif_nonstd
            vi vi_ctrs xor_eqns rem_ls ctr tas failed tail
      | No_Solution ->
          process_layer_elem unif_std unif_nonstd
            vi vi_ctrs xor_eqns rem_ls ctr tas (std_sub::failed) tail
      )
  | [] -> (tas,failed)
;;


(* Input:
 * - The given layer is the last argument ((std_subs,vi,vi_unequal_ctrs)::tail)
 *
 *   A layer element (std_subs,vi,vi_unequal_ctrs) consists of a list of
 *   substitutions for the std problem (std_subs), a variable identification
 *   (vi) and a negative constraints for the variable identification
 *   (vi_unequal_ctrs).
 *
 *   vi is the corresponding var identif to std_subs, i.e. if x and y are equal
 *   in vi, than they have to be equal in all substitutions in std_subs.
 *
 *   The neg constraints (vi_unequal_ctrs) are represented by a list of pairs of
 *   variables ((int * int) list). (x,y) in vi_unequal_ctrs means, that x must
 *   not be identified with y, because we have handled this case already
 *   somewhere else in the tree. This avoids producing less general or
 *   equivalent unifiers, like vi_layer_ctrs.
 *
 * - A layer consists also of constraints on the variable identification
 *   (vi_layer_ctrs).
 *   This is a list of variable identifications (t_partition list).
 *
 *   vi in vi_layer_ctrs means, that we already have unifiers for this and less
 *   general variable identifications. This avoids producing less general or
 *   equivalent unifiers.
 *
 * - unif_std and unif_nonstd are the unification alg for the std and xor
 *   theory, respectivaly.
 *
 * - xor_eqns the xor unification problem
 *
 * - rem_ls and ctr are the parts that are left from the originally given
 *   substitution. Do not confuse ctr with vi_ctrs.
 *
 * - tas a list of substitutions (t_subst list), all produced combined unifiers
 *   are added to tas.
 *
 *
 * Description:
 * - Apply process_layer_elem to the first layer element.
 *   This produces two list of substitutions subs and failed.
 *   subs   contains combined unifiers for the problem.
 *   failed containes standard substitutions, for that unification of the xor
 *          problem failed.
 *
 * - add subs to tas
 *
 * - Do variable identification (identify_variables) for the substitutions that
 *   failed (if there are any).
 *   This generates a list of layer elements. Add them to nextlayer.
 *
 * - If failed is empty, i.e. solving the xor problem suceeded for every std
 *   substitution, then add the active vi to nextlayer_ctrs, because we have a
 *   complete set of unifiers for this and all less general var identifications.
 *
 * - Proceed with the next layer element.
 *
 * - If all layer elements of the active layer are handled, proceed with the
 *   next layer (nextlayer).
 *
 *
 * Remarks:
 * - What process_layer does in fact is doing breath first search on the tree of
 *   all possible variable identifications.
 *)
let rec process_layer
   unif_std
   unif_nonstd
  (xor_eqns       : term list list                                      )
  (rem_ls         : t_equal                                             )
  (ctr            : t_unequal                                           )
  (nextlayer      : (t_subst list * t_partition * (int * int) list) list)
  (nextlayer_ctrs : t_partition list                                    )
  (vi_layer_ctrs  : t_partition list                                    )
  (tas            : t_subst list                                        )
                  : (t_subst list * t_partition * (int * int) list) list
                 -> t_subst list =

  function
    | (std_subs,vi,vi_unequal_ctrs)::tail ->
        (* solve problem for this vi, add solutions to tas and add the subst
         * that fail to failed *)
        let (subs,failed) =
          process_layer_elem unif_std unif_nonstd vi
            (vi_unequal_ctrs,vi_layer_ctrs)
            xor_eqns rem_ls ctr [] [] std_subs
        in
        let tas = merge tas subs
        in

        (* do identification for failed, add it to nextlayer and
           add vi to nextlayer_ctrs, if no substituition failed *)
        let nextlayer =
          if failed=[] then
            nextlayer
          else
            identify_variables unif_std nextlayer failed vi
              (vi_unequal_ctrs,vi_layer_ctrs)
        in
        let nextlayer_ctrs =
          if failed=[] && subs<>[] then
            vi::nextlayer_ctrs
          else
            nextlayer_ctrs
        in

        (* proceed with tail *)
        process_layer unif_std unif_nonstd xor_eqns rem_ls ctr nextlayer
          nextlayer_ctrs vi_layer_ctrs tas tail
    | [] ->
      (* this layer is finished, proceed with nextlayer, if this is not empty
       *)
      if nextlayer=[] then
        tas
      else
        (* start with an empty nextlayer, the new vi_layer_ctrs are the former
         * nextlayer_ctrs *)
        process_layer unif_std unif_nonstd xor_eqns rem_ls ctr []
          nextlayer_ctrs nextlayer_ctrs tas nextlayer
;;


(* Unifies u and v with the given substitutions subs combining the unification
 * algorithms unif_std and unif_nonstd for two disjoint theories.
 *
 * If subs has only one element s then the empty list ([]) is returned, if u and
 * v are not unifiable. Otherwise the result is a complete set of unifiers for
 * u =? v, where each unifier is an extension of s.
 *
 * If to_purify is true it is assumed, that u and v are purified and normalized
 * with respect to subs.
 *)
let unif_combi
   unif_std
   unif_nonstd
  (to_purify   : bool        )
  (u           : term        )
  (v           : term        )
  (subs        : t_subst list)
               : t_subst list =

  let rec unif_combi_help (tas : t_subst list) : t_subst list -> t_subst list =
    function
    | sub::tail ->
        (* Purify u2 and v2. It is also ensured that they are of the same theory
         * afterwards.
         * If they are in the xor theory, at least one of them is Atm(0). *)
        let (u2,v2,sub) = purify_pair to_purify sub (u,v)
        in

        (* Split the substitution into three parts, std_ls, xor_eqns and rem_ls.
         *)
        let (u2,v2,(std_ls,xor_eqns,rem_ls,fv,ctr),std_vars,xor_vars) =
          split_subst sub u2 v2
        in

        (* shared_vars are the variables that occur in std_ls and xor_ls at the
         * same time *)
        let shared_vars = intersection std_vars xor_vars
        in

        (* If u2 and v2 are standard std_subs is the result of unifing u2 and v2
         * with the unif_std. Otherwise std_subs is [(std_ls,fv,[])].
         *
         * xor_eqns contains the remaining xor equations that have to be made
         * equal to Atm(0) by unification in the xor theory. *)
        let (xor_eqns, std_subs) =
          match (u2,v2) with
          | (Atm(0),Atm(0)) -> (xor_eqns, [(std_ls,fv,[])])
          | (Atm(0),Var(n))
          | (Var(n),Atm(0)) ->
              (* add 0 = Var(n) to xor_eqns *)
              ([Var(n)]::xor_eqns, [(std_ls,fv,[])])
          | (Atm(0),Xor(l))
          | (Xor(l),Atm(0)) ->
              (* add 0 = l to xor_eqns *)
              (l::xor_eqns, [(std_ls,fv,[])])
          | _ ->
              (* u2 and v2 are both standard,
               * (u2,v2) = (Xor(_),Xor(_)), (Atm(0),t) or (Xor(_),t) with t
               * standard cannot appear, due to purify_pair *)
              (xor_eqns, unif_std false u2 v2 [(std_ls,fv,[])])
        in

        (if std_subs = [] then
          (* solving the std problem failed, so u and v are not unifiable with
           * respect to sub. Proceed with tail. *)
          unif_combi_help tas tail
        else
          (* Start the breadth first search with the most general layer element
           * (std_subs,vi,[]), so with the trivial variable identif and without
           * layer_ctrs. See comments of process_layer for details.
           *
           * Add the results to tas. *)
          let vi = trivial_part shared_vars
          in
          let tas2 = process_layer unif_std unif_nonstd xor_eqns rem_ls ctr
            [] [] [] tas [(std_subs,vi,[])]
          in
          (* proceed with tail *)
          unif_combi_help tas2 tail
        )
    | [] -> tas
  in
  unif_combi_help [] subs
;;




(* ===============================
   Standard Unification Algorithms
   =============================== *)


(* Unification for exponentials. *)
let rec merge_expinv tas = function
  | (t,Pos)::fin -> merge_expinv ((t,Neg)::tas) fin
  | (t,Neg)::fin -> merge_expinv ((t,Pos)::tas) fin
  | [] -> tas ;;
let exp_equal unif l1 l2 subs =
  let rec tri tas1 tas2 = function
    | (t,Pos)::fin -> tri (t::tas1) tas2 fin
    | (t,Neg)::fin -> tri tas1 (t::tas2) fin
    | [] -> (tas1,tas2)  in
  let lg,lr = tri [] [] l1  in
  let lr,lg = tri lr lg l2  in
  let rec comb lref subs lst = 
    if subs = [] then []
    else match lst with
    | t::fin -> 
	let rec f lsubs lref_tas = function
	  | tt::lref_fin -> f (merge lsubs (comb (merge lref_fin lref_tas) (unif true t tt subs) fin))
		(tt::lref_tas) lref_fin
	  | [] -> lsubs
	in f [] [] lref
    | [] -> if lref = [] then subs else []
  in comb lg subs lr ;;
let rec unif_exp unif a l1 b l2 subs = 
  (match (a,b) with
  | (Exp(a,l3),_) -> (match merge_exp l3 l1 with 
                      | [] -> unif true a (Exp(b,l2)) subs
		      | l1 -> unif_exp unif a l1 b l2 subs)
  | (_,Exp(b,l3)) -> (match merge_exp l3 l2 with
                      | [] -> unif true (Exp(a,l1)) b subs
                      | l2 -> unif_exp unif a l1 b l2 subs)
  | (PInv(PInv(a)),b)
  | (a,PInv(PInv(b))) -> unif_exp unif a l1 b l2 subs
  | _ -> merge (merge
	  (unif true a b (exp_equal unif l1 l2 subs))   (* a bit redundant... *)
	  (unif true a (Exp(b, merge_expinv l2 l1)) subs))
	  (unif true b (Exp(a, merge_expinv l1 l2)) subs))
;;



(* Unification algorithm, standard theory, no associativity for pair
 *
 * TODO: check, where it may be to_purify and where it has to be true *)
let rec unif_uplet_light subs = function
  | (a::l1, b::l2) -> unif_uplet_light (unif_light true a b subs) (l1,l2)
  | ([],[]) -> subs
  | _ -> []
and unif_light (to_purify : bool) (u : term) (v : term) (subs : t_subst list) : t_subst list =
  match (u,v) with
    | (Var(n),t) | (t,Var(n))       -> add unif_light (unif_combi unif_light unif_xor) to_purify n t [] subs
    | (Atm(n1),Atm(n2))             -> if n1=n2 then subs else []
    | (Uplet(l1),Uplet(l2))         -> unif_uplet_light subs (l1,l2)
    | (PCrypt(m1,k1),PCrypt(m2,k2))
    | (SCrypt(m1,k1),SCrypt(m2,k2)) -> unif_light true m1 m2 (unif_light to_purify k1 k2 subs)
    | (PInv(t1),t2) | (t2,PInv(t1)) ->
	( match (t1,t2) with
	  | (_,PInv(t2)) -> unif_light to_purify t1 t2 subs
	  | (Var(_), _)  -> unif_light to_purify t1 (PInv t2) subs                  (* since t2 <> PInv(_) *)
	  | (PInv(t1),_) -> unif_light to_purify t1 t2 subs
	  | (Xor(_), _)  -> unif_combi unif_light unif_xor to_purify t1 (PInv t2) subs    (* t2 <> PInv(_) *)
	  | (_, Xor(_))  -> unif_combi unif_light unif_xor to_purify (PInv t1) t2 subs
	  | _            -> []  )
    | (Exp(a,l1),Exp(b,l2))         -> unif_exp unif_light a l1 b l2 subs
    | (Xor(_),_) | (_,Xor(_))       ->
        (* TODO: we could first test, if it is really a xor problem,
         * because 0 =? Xor[t1;t2] is not really xor,
         * since it is equivalent to t1 =? t2 *)
        unif_combi unif_light unif_xor to_purify u v subs
    | _                             -> []
;;


(* Unification algorithm, standard theory, limited associativity for pair
 *
 * TODO: check, where it may be to_purify and where it has to be true *)
exception NeedXor ;;
let rec unif_uplet_norm s = function
  | Uplet(l)::fin -> unif_uplet_norm s (l@fin)
  | (Var n)::fin -> 
      let t = valeur n s in 
      if t = (Var n) then t::(unif_uplet_norm s fin) else (unif_uplet_norm s (t::fin))
  | (Xor _)::_ -> raise NeedXor
  | t::fin -> t::(unif_uplet_norm s fin)
  | [] -> [] ;;
let rec unif_uplet_cmp_f z1 z2 ss = if ss=[] then ss else
  match (z1,z2) with
  | ([],[])                        -> ss
  | ((Var(_) as hd)::tl,(Var(_) as hd2)::tl2) -> 
      (unif_uplet_cmp_g ss hd tl [] z2)@(unif_uplet_cmp_g ss hd2 tl2 [hd] tl)
  | ((Var(_) as hd)::tl,_)         -> unif_uplet_cmp_g ss hd tl [] z2
  | (_,(Var(_) as hd)::tl)         -> unif_uplet_cmp_g ss hd tl [] z1
  | (t1::zz1,t2::zz2)              -> unif_uplet_cmp_f zz1 zz2 (unif_cmp true t1 t2 ss)
  | _                              -> []
and unif_uplet_cmp_g su d1h d1t d2 = function
  | []    -> []
  | t::d3 -> (unif_uplet_cmp_f d1t d3 (unif_cmp true(*false*) d1h
					 (if d2=[] then t
					  else Uplet (List.rev (t::d2))) su))
           @ (unif_uplet_cmp_g su d1h d1t (t::d2) d3)
and unif_uplet_cmp_h tas lu lv = function
  | s::fin -> 
      let tas = 
	try 
	  merge tas (unif_uplet_cmp_f (unif_uplet_norm s lu) (unif_uplet_norm s lv) [s])
	with
	  NeedXor -> unif_combi unif_cmp unif_xor true (Uplet lu) (Uplet lv) [s]
      in unif_uplet_cmp_h tas lu lv fin
  | [] -> tas
and unif_cmp (to_purify : bool) (u : term) (v : term) (subs : t_subst list) : t_subst list =
  match (u,v) with
    | (Var(n),t) | (t,Var(n))       -> add unif_cmp (unif_combi unif_cmp unif_xor) to_purify n t [] subs
    | (Atm(n1),Atm(n2))             -> if n1=n2 then subs else []
    | (Uplet(lu),Uplet(lv))         -> 
	if to_purify then 
	  unif_uplet_cmp_h [] lu lv subs
	else 
	  unif_uplet_cmp_f lu lv subs
    | (PCrypt(m1,k1),PCrypt(m2,k2))
    | (SCrypt(m1,k1),SCrypt(m2,k2)) -> unif_cmp true m1 m2 (unif_cmp to_purify k1 k2 subs)
    | (Uplet([t1]),_)               -> unif_cmp to_purify t1 v subs
    | (_,Uplet([t2]))               -> unif_cmp to_purify u t2 subs
    | (PInv(t1),t2) | (t2,PInv(t1)) ->
	( match (t1,t2) with
	    | (t1, PInv(t2))  | (PInv(t1), t2) 
	    | (Uplet([t1]),t2) -> unif_cmp to_purify t1 t2 subs
	    | (Xor(_),_)   -> unif_combi unif_cmp unif_xor to_purify t1 (PInv t2) subs
	    | (_, Xor(_))  -> unif_combi unif_cmp unif_xor to_purify (PInv t1) t2 subs
            | (Var(_), _)  -> unif_cmp true t1 (PInv t2) subs
            | _            -> [] )
    | (Xor(_),_) | (_,Xor(_))       ->
        (* TODO: we could first test, if it is really an xor problem,
         * because 0 =? Xor[t1;t2] is not really xor,
         * since it is equivalent to t1 =? t2 *)
        unif_combi unif_cmp unif_xor to_purify u v subs
    | (Exp(a,l1),Exp(b,l2))         -> unif_exp unif_cmp a l1 b l2 subs
    | _             -> []
;;

(* Unification choice *)
let unif (to_purify : bool) (u : term) (v : term) (subs : t_subst list) : t_subst list =
  if !light_unif then
    unif_light to_purify u v subs
  else
    unif_cmp to_purify u v subs
;;

(* Simulation of a standard unification algorithm, without to_purif and subs *)
let usual_unification (fv : int) (u : term) (v : term) : t_subst list =
  unif_light true u v [[],FreeVar fv,[]]
;;
