
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche, Thomas Genet and Erwan Houssay

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.5. Last modification 23/09/08
*)
open Types;;
open Globals;;
open Interface;;

(*
  \section{Functions for debugging}
*)

(*
   The following functions are used as debugging utilities. 
   They allow the programmar to check the content of the main globals tables.
   It is recommanded to set [Globals.debug_level] to a non null value to use them.
*)

(*
   \begin{center}
   type: [val prerr_type_table : unit -> unit]
   \end{center}
*)
let prerr_type_table () =
  prerr_endline "% TYPE TABLE:  [ident_id:ident_type]";
  Hashtbl.iter
    (fun i t ->
      prerr_string "[";
      prerr_int i;
      prerr_string ":"; 
      prerr_term_type t;
      prerr_string "]")
    type_table#get_tab;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val prerr_player_to_role : unit -> unit]
   \end{center}
*)
let prerr_player_to_role () =
  prerr_endline "% PLAYER TO ROLE list:  [player_id:role_id]";
  List.iter
    (fun (a,b) ->
      prerr_string "[";
      prerr_int a;
      prerr_string ":";
      prerr_int b;
      prerr_string "]")
    !Globals.player_to_role;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val prerr_string_id : unit -> unit]
   \end{center}
*)
let prerr_string_id () =
  prerr_endline "% STRING ID:  [ident_id:ident_string]";
  Hashtbl.iter
    (fun i st ->
      prerr_string "[";
      prerr_int i; prerr_string ":"; 
      prerr_string st;
      prerr_string "]")
    string_id#get_tab;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val prerr_id_string : unit -> unit]
   \end{center}
*)
let prerr_id_string () =
  prerr_endline "% STRING ID rev.:  [(ident_string,role_id):ident_id]";
  Hashtbl.iter
    (fun (st,i1) i2 ->
      prerr_string "[(";
      prerr_string st;
      prerr_string ",";
      prerr_int i1; 
      prerr_string "):"; 
      prerr_int i2;
      prerr_string "]")
    string_id#get_tab_string_to_id;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val prerr_global_ids : unit -> unit]
   \end{center}
*)
let prerr_global_ids () =
  prerr_endline "% GLOBAL VAR ID:  [var_num:(role_id,ident_id)]";
  Hashtbl.iter
    (fun i (j,k) ->
      prerr_string "["; prerr_int i; prerr_string ":"; 
      prerr_string "("; prerr_int j; prerr_string ",";
      prerr_int k; prerr_string ")";
      prerr_string "]")
    global_var_id#get_tab;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val prerr_mem_map : unit -> unit]
   \end{center}
*)
let prerr_mem_map () =
  prerr_endline "% MEMORY MAP:  [var_num:map_num]";
  Hashtbl.iter
    (fun i j ->
      prerr_string "["; prerr_int i;
      prerr_string ":"; prerr_int j;
      prerr_string "]")
    memory_map#get_tab;
  prerr_newline()
;;

(* 
   \subsection{Dumping the values in the tables}

*)

(*
   \begin{center}
   type: [val prerr_value_table : unit -> unit]
   \end{center}
*)
let prerr_value_table () =
  let prerr_value_couple mmap_id if_value =
    prerr_string "["; prerr_int mmap_id; 
    prerr_string ":"; Interface.prerr_ground_if_value if_value;
    prerr_string "]"
  in
  prerr_endline "% VALUE TABLE:  [map_num:if_value]";
  let table = value_table#get_tab in
  Hashtbl.iter prerr_value_couple table;
  prerr_newline()
;;

(*
   \begin{center}
   type: [val dump : unit -> unit]
   \end{center}
*)
let dump () =
  prerr_newline();
  prerr_endline "%%% Dumping the tables: ";
  prerr_string_id(); 
  prerr_id_string();  
  prerr_type_table();  
  prerr_global_ids ();  
  prerr_mem_map (); 
  prerr_value_table ();
  prerr_player_to_role()
;;

(*
   \begin{center}
   type: [val gvi_get_name : int -> int]
   \end{center}
*)
let gvi_get_name gid =
  try
    Globals.global_var_id#get_name_of gid
  with
    Not_found ->
      prerr_endline ("Utils.gvi_get_name: The integer "^(string_of_int gid)^" is not a key in the global_var_id table");
      raise Not_found
;;


(*
   \begin{center}
   type: [val affiche_equation : int * Types.atoms_type Types.term_structure -> unit]
   \end{center}
*)
let affiche_equation (a,b) =
  prerr_endline ("  variable: "^(Globals.string_id#get_name (global_var_id#get_name_of a)));
  prerr_string "      term: "; Interface.prerr_term b; prerr_newline();
  prerr_string "     value: ";
  (try 
    Interface.prerr_ground_if_value (value_table#get_value (memory_map#get_map a))
  with Not_found -> prerr_string "?");
  prerr_newline()
;;

(*
   \begin{center}
   type: [val affiche_debug : (int * Types.atoms_type Types.term_structure) list -> unit]
   \end{center}
*)
let affiche_debug l =
  prerr_endline "% final assignment list:";
  List.iter affiche_equation l
;;
