(**
Lexer generator for AVISPA output (OFMC + ATSE)
@author Erwan HOUSSAY
@since summer 2008
*)

{

(* Link to corresponding parser *)
open Atkifparser;;

(** Points out lexicon error in each section *)
exception Lexicon_summary of char;;
exception Lexicon_attack of char;;
(** Points out that no tool has been identified as the author of the parsed report *)
exception Tool_undefined;;

(** Possible lexing entry points *)
type entry_point =
	SECTION_SUMMARY
  |	SECTION_ATTACK
  |	SECTION_FILTER
;;
(** Current lexing entry point *)
let state = ref SECTION_FILTER;;

(** Lists sections run through by lexing *)
type section =
	RUN_NONE
  |	RUN_SUMMARY
  |	RUN_SKIP
  |	RUN_ATTACK
;;
(** Last interesting section run through *)
let last_run = ref RUN_NONE;;

(** Distinguishes AVSIPA output author tools in order to solve ambiguous lexemes *)
type avispa =
	TOOL_ATSE
  |	TOOL_OFMC
  |	TOOL_UNDEF
;;
(** Stores AVISPA tool identified as the author of the report being parsed *)
let tool = ref TOOL_UNDEF;;

(** Reset global variables used for lexing *)
let (lexing_reset: unit -> unit) = function () ->
	state := SECTION_FILTER ;
	tool := TOOL_UNDEF ;
	last_run := RUN_NONE
;;

(**
Builds tokens out of strings, singling out keywords of section "ATTACK_TRACE" among identifiers
@param string : lexeme
@return token : either keyword or identifier
*)
let (attack_trace_filter: string -> token) = function
	"start"     -> START
|	id          -> IDENT id
;;

(**
Builds tokens out of strings, singling out keywords of section "SUMMARY" among identifiers
@param string : lexeme
@return token : either keyword or identifier
*)
let (summary_filter: string -> token) = function
	"SAFE"					-> SAFE
|	"UNSAFE"				-> UNSAFE
|	"INCONCLUSIVE"				-> INCONCLUSIVE
|	"DETAILS"				-> DETAILS
|	"ATTACK_FOUND"				-> ATTACK_FOUND
|	"STRONGLY_TYPED_MODEL"			-> STRONGLY_TYPED_MODEL
|	"TYPED_MODEL"				-> TYPED_MODEL
|	"UNTYPED_MODEL"				-> UNTYPED_MODEL
|	"BOUNDED_NUMBER_OF_SESSIONS"		-> BOUNDED_NUMBER_OF_SESSIONS
|	"BOUNDED_NUMBER_OF_SYMBOLIC_SESSIONS"	-> BOUNDED_NUMBER_OF_SYMBOLIC_SESSIONS
|	"BOUNDED_SEARCH_DEPTH"			-> BOUNDED_SEARCH_DEPTH
|	"BOUNDED_MESSAGE_DEPTH"			-> BOUNDED_MESSAGE_DEPTH
|	"TIME_OUT"				-> TIME_OUT
|	"MEMORY_OUT"				-> MEMORY_OUT
|	"NOT_SUPPORTED"				-> NOT_SUPPORTED
|	"OVER_APPROXIMATION"			-> OVER_APPROXIMATION
|	"UNDER_APPROXIMATION"			-> UNDER_APPROXIMATION
|	id					-> IDENT id
;;

}

(** Regular expressions *)

(* Skipped characters *)
let skip = [' ''\t''\r''\n''%']+
(* Identifiers *)
let alphanum = ['a'-'z''A'-'Z''0'-'9''-''_']*
let var = ['A'-'Z'] alphanum
let const = ['a'-'z'] alphanum
let ident = var | const
(* Numbers *)
let number = ['0'-'9']+
(* Security-properties-linked events mentionned along with transitions in ATSE attack trace *)
let prop = ('&' [^'\n']* '\n' skip* )+
(* Headers of uninteresting parts of the report *)
let deb_inu = "PROTOCOL" (*ATSE+OFMC*) | "Reached State:" (*OFMC*)

(** Lexing entry points *)

(** Produces tokens for section "ATTACK TRACE" *)
rule section_attack = parse
	deb_inu					{ state := SECTION_FILTER ; section_filter lexbuf }
|	'('(const as a)','(number as n)')'	{ AGENT (a,(int_of_string n)) }
|	'n'(number as n)'('(var as s)')'	{ NONCE_ATSE (s,(int_of_string n)) }
|	(var as s)'('(number as n)')'		{ match !tool with TOOL_ATSE -> FREE_ATSE (s,(int_of_string n)) | TOOL_OFMC -> NONCE_OFMC (s,(int_of_string n)) | TOOL_UNDEF -> raise Tool_undefined }
|	'x'(number as n)			{ FREE_OFMC ("x",(int_of_string n)) }
|	prop					{ PROP }
|	ident as id				{ attack_trace_filter id }
|	number as nb				{ NUMBER (int_of_string nb) }
|	"->"					{ ARROW }
|	':'					{ COLON }
|	"()"					{ EMPTY }
|	'('					{ L_ROUND }
|	')'					{ R_ROUND }
|	'{'					{ L_CURLY }
|	'}'					{ R_CURLY }
|	','					{ COMMA }
|	'.'					{ DOT }
|	'_'					{ UNDERSCORE }
|	eof					{ lexing_reset() ; EOF }
|	skip					{ section_attack lexbuf }
|	_ as x					{ lexing_reset() ; raise (Lexicon_attack x) }
(** Produces tokens for section "SUMMARY" *)
and section_summary = parse
	deb_inu		{ state := SECTION_FILTER ; last_run := RUN_SKIP ; section_filter lexbuf }
|	ident as id	{ summary_filter id }
|	skip		{ section_summary lexbuf }
|	_ as x		{ lexing_reset() ; raise (Lexicon_summary x) }
(** Filters uninteresting sections *)
and section_filter = parse
	"SUMMARY"			{ state := SECTION_SUMMARY ; last_run := RUN_SUMMARY ; SUMMARY }
|	"ATTACK TRACE"			{ state := SECTION_ATTACK ; last_run := RUN_ATTACK ; ATTACK_TRACE }
|	"BACKEND" skip+ "CL-AtSe"	{ tool := TOOL_ATSE ; section_filter lexbuf }
|	"BACKEND" skip+ "OFMC"		{ tool := TOOL_OFMC ; section_filter lexbuf }
|	eof				{ EOF }
|	_				{ section_filter lexbuf }

{

(** Calls relevant lexing entry point according to the lexer's current state *)
let (adaptative_lexing: Lexing.lexbuf -> token) = function buf ->
	match !state with
		SECTION_SUMMARY	-> section_summary buf
	|	SECTION_ATTACK	-> section_attack buf
	|	_		-> section_filter buf
;;

}
