
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche, Thomas Genet and Erwan Houssay

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.5. Last modification 23/09/08
*)
open Types;;
open Globals;;
open Utils;;
open Error_handler;;

(*
   This module defines an interface with the system output.
   The functions work on data structures returned by the instantation module.
   They perform last formatting steps of the IF sections and print them on stdin.
 *)

(*
   When a rule is printed, we may have to add some [Contains] fact. We build the list
   of these facts as the states are printed.
  \begin{center}
  type:
   [val contains_term_list : (Types.if_value_type * Types.if_value_type) list ref]
  \end{center}
*)
let contains_term_list = ref ([]: (if_value_type * if_value_type) list);;

(* 
   The signature section contains state signatures such as:
    \begin{center}
   [state_Alice(agent,agent,public_key,public_key,nat,{3,1,2},nonce,nonce,nat) -> fact]
   \end{center}

   In order to print a signature only once, we store it in the following list.
  \begin{center}
  type:
   [val roles_printed : string list ref]
  \end{center}
*)
let roles_printed = ref ([] : string list);;


(*
  \begin{center}
  type:
   [val list_to_string : ('a -> string) -> string -> 'a list -> string]\\
  type:
   [val print_list : ('a -> string) -> string -> 'a list -> unit]\\
  type:
   [val prerr_list : ('a -> string) -> string -> 'a list -> unit]
  \end{center}
*)
let list_to_string f separator =
  let rec list_to_string_rec = function
      [] -> ""
    | [h] -> f h
    | h::t -> (f h)^separator^(list_to_string_rec t)
  in
  list_to_string_rec
;;
let print_list f separator l =
  prout_string (list_to_string f separator l)
;;
let prerr_list f separator l =
  prerr_string (list_to_string f separator l)
;;

(*
  \begin{center}
  type:
   [val list2_to_string : ('a -> string) -> string -> 'a list -> string]
  \end{center}
*)
let list2_to_string f separator =
  let rec list2_to_string_rec = function
      [] -> ""
    | [h] -> f h	
    | h::t ->
 	"pair("^(f h)^separator^(list2_to_string_rec t)^")"
  in
  list2_to_string_rec
;;


(*
   [print_term_type] is used for section types or signature to print
   recursively the types of terms.
  \begin{center}
  type:
   [val term_type_to_string : Types.basic_type Types.term_structure -> string]\\
  type:
   [val prerr_term_type : Types.basic_type Types.term_structure -> unit]
  \end{center}
*)
let rec term_type_to_string = function
    Base(term) | Prime(term) ->
      (match term with
	Agent -> "agent"
      | Public_key -> "public_key"
      | Symmetric_key -> "symmetric_key"
      | Text -> "nonce"
      | Protocol_id -> "protocol_id"
      | Message -> "message"
      | Channel -> "channel"
      | Channel_dy -> "channel(dy)"
      | Channel_ota -> "channel(ota)"
      | Bool -> "bool"
      | Hash_func -> "hash_func"
      | Nat -> "nat"
(*      | Pointer -> "pointer"*)
      | Enum(list) ->
	  "{"
	  ^(List.fold_left
	      (fun s num ->
	      	if not (num = List.hd list) then
	      	  s^","^(string_id#get_name num)
	      	else
		  s^(string_id#get_name num))
	      "" list)
	  ^"}")
	
  | Function(term,list)->
      (match list with
	[] -> "unit"
      | _ ->
	  (List.fold_left
 	    (fun s t -> 
	      s^(term_type_to_string t)^" -> ")
	    "" list)
	  ^(term_type_to_string term))
      
(*  | Lst(list)->
      (List.fold_left
	 (fun s t ->
	   s^(term_type_to_string t))
	 "" list)
      ^" list"*)
      
  | Set([])->
     "set"
  | Set(list)->
     (term_type_to_string (List.hd list))^" set"
      
  | Pair(term1,term2)->
      "pair("
      ^(term_type_to_string term1)^","^(term_type_to_string term2)
      ^")"
        
  | _ -> "nonce"
;;
let prerr_term_type t =
  prerr_string (term_type_to_string t)
;;


(*
  \section{Generic printing functions}

  \subsection{Printing [if_value_type] terms}
  \begin{center}
  type:
   [val if_value_to_string :
   (Types.if_value_type -> string) -> Types.if_value_type -> string]\\
  type:
   [val print_if_value :
   (Types.if_value_type -> string) -> Types.if_value_type -> unit]\\
  type:
   [val prerr_if_value :
   (Types.if_value_type -> string) -> Types.if_value_type -> unit]
  \end{center}
*)
let rec if_value_to_string atom_to_string if_value =
  let if_value_to_string_aux = if_value_to_string atom_to_string in
  match if_value with
    If_address _ | If_const _ ->
      atom_to_string if_value
  | If_pair(term1,term2) ->
      "pair("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_scrypt(term1,term2) ->
      "scrypt("^(if_value_to_string_aux term2)^","
      ^(if_value_to_string_aux term1)^")"
  | If_crypt(term1,term2) ->
      "crypt("^(if_value_to_string_aux term2)^","
      ^(if_value_to_string_aux term1)^")"
  | If_inv(term) ->
      "inv("^(if_value_to_string_aux term)^")"
  | If_function(term1,termlist) ->
      begin
	match term1 with
	    If_const(fun_id) ->
	      let fun_string = (string_id#get_name fun_id) in
		if (fun_string = "exp") || (fun_string = "xor") then
		  fun_string^"("
		  ^(list_to_string if_value_to_string_aux "," termlist)
		  ^")"
		else
		  "apply("
		  ^(if_value_to_string_aux term1)
		  ^","
		  ^(list2_to_string if_value_to_string_aux "," termlist)
		  ^")"
	  | _ -> 
	      "apply("^(if_value_to_string_aux term1)^","
	      ^(list2_to_string if_value_to_string_aux "," termlist)^")"
      end
  | If_cons(term1,term2) ->
      "cons("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_delete(term1,term2) ->
      "delete("^(if_value_to_string_aux term1)^","
      ^(if_value_to_string_aux term2)^")"
  | If_set(termlist) ->
      "set("^(list_to_string if_value_to_string_aux "," termlist)^")"
(*  | If_lst(termlist) ->
      "list("^(list_to_string if_value_to_string_aux "," termlist)^")"*)
;;
let print_if_value atom_fn if_value =
  prout_string (if_value_to_string atom_fn if_value)
;;
let prerr_if_value atom_fn if_value =
  prerr_string (if_value_to_string atom_fn if_value)
;;

(*
  \begin{center}
  type:
   [val atom_ground_value_to_string : Types.if_value_type -> string]
  \end{center}
*)
let rec atom_ground_value_to_string = function
  | If_address gid-> 
      (*
	 In the case of a variable, we get its initial value from
	 the [Globals.value_table] variable. 
       *)
      let map_id = memory_map#get_map gid in
      let if_value = value_table#get_value map_id in
      if if_value = If_address gid then
	(let vid=global_var_id#get_name_of gid in
	 let name = string_id#get_name vid in
	 Error_handler.displayWarning 4 ("Interface.atom_ground_value_to_string."^name);
	 "")
      else
	if_value_to_string atom_ground_value_to_string if_value (* \fix{is it possible to reuse [atom_value_to_string] } *)

  | If_const strid ->
      begin
	(string_id#get_name strid)^
	(*
	  If the constant is a pointer to a set, we also have to print the contains terms according to
	  the value of this set. To this end, we add the right terms in the [contains_term_list] variable.
	*)
	(try
	  match (Globals.type_table#get_type strid) with
(*	    Base(Pointer)
	  | Lst(_)*)
	    Set(_) ->
	      begin
		let pointer_global_var_id = Globals.global_var_id#get_id_of 0 strid in
		let pointer_mmap_id = 
		  Globals.memory_map#get_map pointer_global_var_id in
		let if_value_map = Globals.value_table#get_value pointer_mmap_id in 
		match if_value_map with
(*		  If_lst l *)
		  If_set l ->
		    contains_term_list := 
		      (List.map 
			 (fun element -> 
(*			   if (!debug_level >= 3) then
			     begin
			       prerr_string ("element in the set "
					     ^(Globals.string_id#get_name strid)
					     ^": ");
			       prerr_if_value atom_ground_value_to_string element;
			       prerr_newline();
			     end;*)
			   (element,If_const(strid))) 
			 l)
		      @ !contains_term_list;
		    ""
		| _ -> 
		    prerr_endline ("Interface.atom_ground_value_to_string: found a pointer to something else than a set or a list");
		    raise Not_found
	      end;
	  | _ -> ""
	with
	  _  -> "")
      end
     (*
        This last case should not be encountered as long as the [print_atom_ground_value]
	is called only in the previous cases.
      *)
  | _ -> prerr_endline "Interface.atom_ground_value_to_string: considering a non-atomic term as atomic"; ""
;;


(*
  \begin{center}
  type:
   [val ground_if_value_to_string : Types.if_value_type -> string]\\
  type:
   [val print_ground_if_value : Types.if_value_type -> unit]\\
  type:
   [val prerr_ground_if_value : Types.if_value_type -> unit]
  \end{center}
*)
let ground_if_value_to_string =
  if_value_to_string atom_ground_value_to_string
;;
let print_ground_if_value =
  print_if_value atom_ground_value_to_string
;;
let prerr_ground_if_value =
  prerr_if_value atom_ground_value_to_string
;;


(*
  \subsection{Printing [term] terms}
*)


(*
   This function prints input terms in IF format.
  \begin{center}
  type:
   [val term_to_string : Types.atoms_type Types.term_structure -> string]\\
  type:
   [val print_term : Types.atoms_type Types.term_structure -> unit]\\
  type:
   [val prerr_term : Types.atoms_type Types.term_structure -> unit]
  \end{center}
*)
let rec term_to_string = function
    Pair(term1,term2) ->
      "pair("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Scrypt(term1,term2) ->
      "scrypt("^(term_to_string term2)
      ^","^(term_to_string term1)^")"
      
  | Crypt(term1,term2) ->
      "crypt("^(term_to_string term2)
      ^ ","^(term_to_string term1)^")"

  | Inv(term) ->
      "inv("^(term_to_string term)^")"

  | Function(term1,termlist) ->
      begin
	match term1 with
	    Base(Const(fun_id)) ->
	      let fun_string = (string_id#get_name fun_id) in
		if (fun_string = "exp") || (fun_string = "xor") then
		  fun_string^"("
		  ^(list_to_string term_to_string "," termlist)
		  ^")"
		else
		  "apply("^(term_to_string term1)
		  ^","^(list2_to_string term_to_string "," termlist)
		  ^")"
	  | _ -> 
	      "apply("^(term_to_string term1)
	      ^","^(list2_to_string term_to_string "," termlist)
	      ^")"
      end

  | Cons(term1,term2) ->
      "contains("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Delete(term1,term2) -> (* to be contains... *)
      "delete("^(term_to_string term1)
      ^","^(term_to_string term2)^")"

  | Set(termlist) ->
      "set("^(list_to_string term_to_string "," termlist)^")"
(*BUG??? 15/03/2005
      list2_to_string term_to_string "," termlist *)

(*  | Lst(termlist) ->
      "list("^(list_to_string term_to_string "," termlist)^")"*)

  | Base(Var(id)) | Base(Const(id))
  | Prime(Var(id)) | Prime(Const(id)) ->
      Globals.string_id#get_name id
  | Hash(_) ->
      prerr_endline "Interface.term_to_string: Impossible error: invalid hash type/term"; 
      raise Not_found
;;
let print_term t =
   prout_string (term_to_string t)
;;
let prerr_term t =
   prerr_string (term_to_string t)
;;



(*
   This function is similar to the previous one except that it use HLPSL terms and not IF terms.
  \begin{center}
  type:
   [val term_value_to_string : int -> Types.atoms_type Types.term_structure -> string]
  \end{center}
*)
let rec term_value_to_string session_id =
  (* This sub-fonction prints the IF values of atoms.*)
  let atom_value_to_string hlpsl_id = 
    let gid = global_var_id#get_id_of session_id hlpsl_id in
    let map_id = memory_map#get_map gid in
    let if_value = value_table#get_value map_id in
    ground_if_value_to_string if_value
  in function
      Pair(term1,term2) ->
      	"pair("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Scrypt(term1,term2) ->
      	"scrypt("^(term_value_to_string session_id term2)^","
      	^(term_value_to_string session_id term1)^")"
    | Crypt(term1,term2) ->
      	"crypt("^(term_value_to_string session_id term2)^","
      	^(term_value_to_string session_id term1)^")"
    | Inv(term) ->
      	"inv("^(term_value_to_string session_id term)^")"
    | Function(term1,termlist) ->
      	(term_value_to_string session_id term1)^"("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"
    | Cons(term1,term2) ->
      	"cons("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Delete(term1,term2) ->
      	"delete("^(term_value_to_string session_id term1)^","
      	^(term_value_to_string session_id term2)^")"
    | Set(termlist) ->
      	"set("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"
(*BUG??? 15/03/2005
      	list2_to_string (term_value_to_string session_id) "," termlist *)
(*    | Lst(termlist) ->
      	"list("
      	^(list_to_string (term_value_to_string session_id) "," termlist)^")"*)
    | Prime(Var(id))
    | Prime(Const(id)) ->
      	(Globals.string_id#get_name id)^"'"
    | Base(Var(id)) ->
      	atom_value_to_string id
    | Base(Const(id)) ->
      	string_id#get_name id;
    | Hash(_) ->
	prerr_endline "Interface.term_value_to_string: Impossible error: invalid hash type/term"; 
	raise Not_found
;;


(*

   \subsection{Printing \textsf{Intermediate Format} facts}

   [print_if_term] prints IF facts. It also generates output for facts signatures.
  \begin{center}
   type: [val if_term_to_string : string -> Types.if_term -> string]\\
   type: [val print_if_term : string -> Types.if_term -> unit]
   type: [val prerr_if_term : string -> Types.if_term -> unit]
  \end{center}
*)
let rec if_term_to_string func =

  let type_to_string = function
      Base(Var(i))
    | Base(Const(i)) ->
	term_type_to_string (Globals.type_table#get_type i)
    | Set(_) ->
	"set"
    | _ ->
	"message";
  in function

      State(name,session_id,l) ->
      	if (func = "print_term_type") then
	  if not(List.mem name !roles_printed) then
	    (roles_printed := !roles_printed @ [name];
	     ("state_"^name^" : "^(list_to_string type_to_string " * " l)^" -> fact\n"))
	  else ""
      	else 
	  ("state_"^name^"("^
	   (if func = "print_term_value" then 
	     list_to_string (term_value_to_string session_id) "," l
	   else
	     list_to_string term_to_string "," l)
	   ^")")

    | Iknows t -> 
      	if not (func = "print_term_type") then
	  ("iknows("^(term_to_string t)^")")
	else ""

    | Ifcontains (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("contains("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifdelete (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("contains("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifequal (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("equal("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifleq (t1,t2) -> 
      	if not (func = "print_term_type") then
	  ("leq("^(term_to_string t1)
	   ^","^(term_to_string t2)^")")
	else ""

    | Ifnot t -> 
 	"not("^(if_term_to_string func t)^")"

(*    | Secret(t1,t2) -> 
     	if not (func = "print_term_type") then
	  ("secret("^(term_to_string t1)^","
	   ^(term_to_string t2)^")")
	else ""*)

(*    | Witness(authenticator, authenticated, const_term, var_term) ->
     	if not (func = "print_term_type") then
	  let pl_authenticator = find_player_in_role authenticator in
	  let pl_authenticated = find_player_in_role authenticated in
	  ("witness("
	   ^(Globals.string_id#get_name pl_authenticator)^","
	   ^(Globals.string_id#get_name pl_authenticated)^","
	   ^(term_to_string const_term)^","
	   ^(term_to_string var_term)
	   ^")")
	else ""*)
(*Event: *)
    | Unknown(t,lt) ->
      	(match t with
	  Base(Const(id)) ->
	    let nid = Globals.string_id#get_name id in
	    if (nid = "request") or (nid = "wrequest") then
	      (nid^"("^(list_to_string term_to_string "," lt)^",SID)")
	    else
	      (nid^"("^(list_to_string term_to_string "," lt)^")")
	| _ ->
	    (term_to_string t)^"("
 	    ^(list_to_string term_to_string "," lt)^")")

    | _ ->
 	Error_handler.displayWarning 5 "print_if_term"; ""
;;
let print_if_term func t =
  prout_string (if_term_to_string func t)
;;
let prerr_if_term func t =
  prerr_string (if_term_to_string func t)
;;


(*
   Printing the IF states using formatting function "func".
  \begin{center}
  type:
   [val print_if_states_list : string -> Types.if_term list -> unit]
  \end{center}
*)
let print_if_states_list func l =
  let rec print_if_states_list_rec = function
      [] -> Error_handler.displayWarning 5 "print_if_states"
    | [h] ->
      	if (func <> "print_term_type") then
	  prout_string " ";
	print_if_term func h
    | h::t ->
      	if (func <> "print_term_type") then
	  prout_string " "; 
      	print_if_term func h ;
      	if (func <> "print_term_type") then
	  prout_string func; (* ".\n" *)
      	print_if_states_list_rec t
  in
  print_if_states_list_rec l
(* useful only when goal predicates are automatically generated:
  print_if_states_list_rec 
    (List.filter
       (function
	   Witness _
	 | Secret _ -> false
	 | _ -> true)
       l) *)
;;


let rec ltl_formula_to_string = function
    Pred(Ifnot(p)) -> "~ "^(ltl_formula_to_string (Pred(p)))
  | Pred(p) -> if_term_to_string "" p
  | Op1(f) -> "[] "^(ltl_formula_to_string f)
  | Op2(f) -> "<-> "^(ltl_formula_to_string f)
  | Op3(f) -> "(-) "^(ltl_formula_to_string f)
  | Op4(f) -> "[-] "^(ltl_formula_to_string f)
  | Op5(f) -> "~ "^(ltl_formula_to_string f)
  | Conj(f1,f2) -> "("^(ltl_formula_to_string f1)^" /\\ "^(ltl_formula_to_string f2)^")"
  | Dij(f1,f2) -> "("^(ltl_formula_to_string f1)^" \\/ "^(ltl_formula_to_string f2)^")"
  | Impl(f1,f2) -> "("^(ltl_formula_to_string f1)^" => "^(ltl_formula_to_string f2)^")"
;;


(*
  \subsection{Translation from [atom_type] to [string]}
*)

(*
  \begin{center}
  type:
   [val atoms_type_to_string : Types.basic_type Types.term_structure -> string]
  \end{center}
*)
let rec atoms_type_to_string parentheses a_term =
    match a_term with
      Base(term) | Prime(term) ->
	(match term with
	  Agent -> "agent"
	| Public_key -> "public_key"
	| Symmetric_key -> "symmetric_key"
	| Text -> "text"
	| Message -> "message"
	| Protocol_id -> "protocol_id"
	| Channel -> "channel"
	| Channel_dy -> "channel_dy"
	| Channel_ota -> "channel_ota"
	| Bool -> "bool"
	| Hash_func -> "hash_func"
	| Nat  -> "nat"
(*	| Pointer -> "set"*)
	| Enum(list) ->
	    "{"
	    ^(List.fold_left
		(fun s num ->
	      	  if not (num = List.hd list) then
	      	    s^","^(string_id#get_name num)
	      	  else
		    s^(string_id#get_name num))
		"" list)
	    ^"}")
    | Inv t -> "inv("^(atoms_type_to_string false t)^")";
    | Hash t -> "apply(hash_func,"^(atoms_type_to_string false t)^")";
    | Set [] -> "set"
    | Set lt -> "set("^(atoms_type_to_string false (List.hd lt))^")"
(*    | Lst [] -> "list"
    | Lst lt -> "list("^(atoms_type_to_string false (List.hd lt))^")"*)
    | Pair(t1,t2) ->
	"pair("^(atoms_type_to_string false t1)^","^(atoms_type_to_string false t2)^")"
    | Crypt(t1,t2) -> "crypt("^(atoms_type_to_string false t2)^","^(atoms_type_to_string false t1)^")"
    | Scrypt(t1,t2) -> "scrypt("^(atoms_type_to_string false t2)^","^(atoms_type_to_string false t1)^")"
    | Function(t,lt) ->
	let string_of_args =
	  match lt with
	    [] -> "unit"
	  | [t1] -> atoms_type_to_string true t1
	  | t1::l ->
	      List.fold_left
		(fun s t ->
		  s^" * "^(atoms_type_to_string true t))
		(atoms_type_to_string true t1) l
	in
	if parentheses then
	  "("^string_of_args^" -> "^(atoms_type_to_string true t)^")"
	else
	  string_of_args^" -> "^(atoms_type_to_string true t)
    | _ -> "message"
;;

(*
  \begin{center}
  type:
   [val global_id_to_type_string : int -> string]
  \end{center}
*)
let global_id_to_type_string gid =
  atoms_type_to_string true
    (Globals.type_table#get_type
       (Globals.global_var_id#get_name_of gid))
;;


(*
  \section{Printing the output file}
*)


(*

  \subsection{Printing the prelude}
 
   The prelude from the compiler's point of view is a file
   [prelude.if] containing static data, containing some
   meta-information about the displayed IF sections.  It also contains
   the intruder section, describing its main actions.  This file is
   independant, and therefore not printed by the compiler; it has to
   be considered by the backends.
*)

(*

  \subsection{Printing the types section}

*)

(*
  \begin{center}
  type:
   [val print_section_types : unit -> unit]
  \end{center}
*
let print_section_types () =
  *
     The key  in the [identifiers_type_table] is a string representing 
     a type of an identifier as it will be printed. The [int] is the [string_id]
     corresponding to the identifier.
   *
  let identifiers_type_table = (Hashtbl.create 15 : (string,int) Hashtbl.t) in
  let identifiers_type_table_key_list = ref ([]: string list) in
  Hashtbl.iter
    (fun x -> function
	*
	   Put here the list of identifiers with a type you don't want to see
	 *
	Base(Channel|Channel_ota|Channel_dy)
      | Function(_)  * To be commented if functions are permitted in IF *
	-> ()
      | y ->
	  let y_string = (atoms_type_to_string false y) in
	  Hashtbl.add identifiers_type_table y_string x;
	  identifiers_type_table_key_list := 
	    Utils.insert_with_no_copy y_string !identifiers_type_table_key_list)
    Globals.type_table#get_tab;

  identifiers_type_table_key_list := 
    List.find_all 
      (function "channel" -> false | _ -> true)
      !identifiers_type_table_key_list;

  prout_endline "section types:";prout_newline ();
  List.iter
    (fun a ->
      let list_of_identifiers_of_type_a = 
	Utils.remove_duplicate
	  (List.map 
	     (Globals.string_id#get_name) 
	     (Hashtbl.find_all identifiers_type_table a))
      in
      print_list (fun s -> s) ", " list_of_identifiers_of_type_a; 
      prout_endline (" : "^a))
    !identifiers_type_table_key_list;
  prout_newline (); prout_newline ()
*)
let print_section_types () =
  (* Computation of the list of ids of basic roles *)
  let role_ids = 0::(List.map snd !Globals.player_to_role) in
  (*
     The key  in the [identifiers_type_table] is a string representing 
     a type of an identifier as it will be printed. The [int] is the [string_id]
     corresponding to the identifier.
   *)
  let identifiers_type_table = (Hashtbl.create 15 : (string,int) Hashtbl.t) in
  let identifiers_type_table_key_list = ref ([]: string list) in
  Hashtbl.iter
    (fun x y ->
      (* If x is a variable, it has to be declared in a basic role *)
	if (not_excluded_type y)
	   && (List.mem (Globals.string_id#get_role_id x) role_ids) then
	  (let y_string = (atoms_type_to_string false y) in
	   Hashtbl.add identifiers_type_table y_string x;
	   identifiers_type_table_key_list := 
	     Utils.insert_with_no_copy y_string !identifiers_type_table_key_list))
    Globals.type_table#get_tab;

  prout_endline "section types:";prout_newline ();
  List.iter
    (fun a ->
      let list_of_identifiers_of_type_a = 
	Utils.remove_duplicate
	  (List.map 
	     (Globals.string_id#get_name) 
	     (Hashtbl.find_all identifiers_type_table a))
      in
      print_list (fun s -> s) ", " list_of_identifiers_of_type_a; 
      prout_endline (": "^a))
    !identifiers_type_table_key_list;
  prout_newline (); prout_newline ()
;;

(*
  \begin{center}
  type:
   [val print_section_signature : role_instance list -> unit]
  \end{center}
*)
let print_section_signature role_instances = 
  (*
     The key in the [role_type_table] is an integer representing the 
     key in the [Globals.string_id] table of the name
     of a role.  The second argument is the string representing 
     the type of the role. The [ -> Fact] part is missing.
   *) 
  let role_type_table = (Hashtbl.create 10 : (int,string) Hashtbl.t) in
  let role_type_table_key_list = ref ([]: int list) in 
  let init_state_list = 
    List.map 
      (fun role_instance ->
(*	(role_instance#get_name, (List.hd role_instance#get_initial_state))*)
	(role_instance#get_name, role_instance#get_initial_state))
      role_instances 
  in
  List.iter
    (fun (a,lb) ->
      List.iter
	(function
	    (*
	       We are only interested in the list of actual arguments since we already
	       know the role identity.
	     *)
	    State(_,_,l) ->
	      let string_from_type =
		match l with
		  [] -> ""
		| Base(Var(head))::tail ->
		    List.fold_left 
		      (fun already_done -> function
			  Base(Var(elt)) -> 
			    (already_done^" * "^(global_id_to_type_string elt)) 
			| _ -> raise Not_found)
		      (global_id_to_type_string head) tail
		| _ -> raise Not_found
	      in
	      Hashtbl.add role_type_table a string_from_type;
	      role_type_table_key_list := Utils.insert_with_no_copy a !role_type_table_key_list
	  | _ -> ())
	lb)
    init_state_list;

  (*
     We now proceed to the actual phase of printing the terms.
   *)
  prout_endline "section signature:"; prout_newline ();
  List.iter
    (fun a ->
       prout_endline 
       	("state_"^(Globals.string_id#get_name a)^": "
	 ^(Hashtbl.find role_type_table a)^" -> fact"))
    !role_type_table_key_list;

  prout_newline (); prout_newline ()
;;

(*
  \begin{center}
  type:
   [val print_section_types_signatures : role_instance list -> unit]
  \end{center}
 *)
let print_section_types_signatures role_instances =
  print_section_signature role_instances;
  print_section_types ()
;;


(*
  \subsection{Printing the initial state}
*)

(*
  \begin{center}
  type:
   [val print_initial_state : Types.if_term list -> unit]
  \end{center}
 *)
let print_initial_state =
  let one_argument_to_string = function
      Base(Var(id)) -> if_value_to_string atom_ground_value_to_string (If_address(id))
    | _ -> "Interface.prout_initial_state: Not a variable"
  in
  let args_to_string =
    list_to_string one_argument_to_string ","
  in
  let one_state_to_string = function
      State(role_string,_,term_list) ->
	" state_"^role_string^"("
	^(args_to_string term_list)
	^")"
    | Iknows(l) ->  
	" iknows("
	^(term_to_string l)
	^")"
    | Unknown(t,lt) ->
	" "^(term_to_string t)^"("
 	^(list_to_string term_to_string "," lt)^")"
    | _ ->
	prerr_endline "Interface.print_initial_state: Warning: Predicate not handled!"; ""
  in
    print_list one_state_to_string ".\n"
;;

(*
  \begin{center}
  type:
   [val print_initial_state : Types.if_term list -> unit]
  \end{center}
 *)
let print_initial_predicates() =
  let one_predicate_to_string (p,lt) =
    " "^(ground_if_value_to_string p)^"("
    ^(list_to_string ground_if_value_to_string "," lt)^")"
  in
  print_list one_predicate_to_string ".\n" !init_predicates
;;

(*
  \begin{center}
  type:
   [val print_section_init : role_instance list -> unit]
  \end{center}
*)
let print_section_init role_instances = 
  let init_state =
    List.flatten
      (List.map 
	 (fun role_instance ->
	   role_instance#get_initial_state)
	 role_instances)
  in
  contains_term_list := [];
  prout_string "section inits:\n\ninitial_state init1 := \n";

(*  if !Globals.global_start_event then
    prout_string " start().\n";*)
  if !Globals.global_start_const then
    prout_string " iknows(start).\n";

  let iknows_list =
    List.map 
      (fun x ->
	" iknows("
	^(ground_if_value_to_string x)
	^")")
      !Globals.i_knowledge
  and preds_list =
    List.map
      (fun (p,lt) ->
	" "^(ground_if_value_to_string p)^"("
	^(list_to_string ground_if_value_to_string "," lt)^")")
      !init_predicates
  in
  print_list (fun x -> x) ".\n" (concatenate_with_no_copy iknows_list preds_list);

  if (init_state <> []) then
    (prout_string ".\n";
     print_initial_state init_state);
  (* temporary measure... *)
  contains_term_list := (Utils.remove_duplicate !contains_term_list);
  List.iter
    (fun (a,b) ->
      prout_string ".\n contains(";
      print_ground_if_value a;
      prout_string ",";
      print_ground_if_value b;
      prout_string ")")
    !contains_term_list;
  contains_term_list := [];
  prout_string "\n\n\n";
;;


(*
   \subsection{Printing the rules section}
*)
(*
  \begin{center}
  type:
   [val print_section_rules : role_instance list * string -> unit]
  \end{center}
let print_section_rules role_instances goal_type goal_cst =
  let filter_wanted_goals =
    List.filter (function
	Unknown(Base(Const(id)),lt) ->
	  let nid = Globals.string_id#get_name id in
	  if (nid = "request") then
            (goal_type = "request")
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else if (nid = "wrequest") then
            (goal_type = "wrequest")
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else if (nid = "witness") then
            ((goal_type = "request") or (goal_type = "wrequest"))
	    && (goal_cst = term_to_string (List.hd (List.tl (List.tl lt))))
	  else (nid <> "secret") or (goal_type = "secret")
      | _ -> true)
*)
let print_section_rules role_instances s_ids a_ids wa_ids =
  let filter_wanted_goals =
    let check_warning cond nid c_id =
      cond
	or (not(!Globals.flag_split)
	      && (Error_handler.displayWarning 11 (nid^" goal predicate ("
						   ^(term_to_string c_id)
						   ^") without corresp. goal property; dropped");
		  false))
    in
    List.filter (function
	Unknown(Base(Const(id)),lt) ->
	  let nid = Globals.string_id#get_name id in
	  if (nid = "request") then
(*	    List.mem (List.hd (List.tl (List.tl lt))) a_ids*)
	    let a_id = List.hd (List.tl (List.tl lt)) in
	    check_warning (List.mem a_id a_ids) nid a_id
	  else if (nid = "wrequest") then
(*	    List.mem (List.hd (List.tl (List.tl lt))) wa_ids*)
	    let wa_id = List.hd (List.tl (List.tl lt)) in
	    check_warning (List.mem wa_id wa_ids) nid wa_id
	  else if (nid = "witness") then
(*	    (List.mem (List.hd (List.tl (List.tl lt))) a_ids)
	      or (List.mem (List.hd (List.tl (List.tl lt))) wa_ids)*)
	    let a_id = List.hd (List.tl (List.tl lt)) in
	    check_warning ((List.mem a_id a_ids)
			     or (List.mem a_id wa_ids))
	      nid a_id
	  else if (nid = "secret") then
(*	    List.mem (List.hd (List.tl lt)) s_ids*)
	    let s_id = List.hd (List.tl lt) in
	    check_warning (List.mem s_id s_ids) nid s_id
	  else true
      | _ -> true)
  in
  let stepcounter = ref 0 in
  let role_list = ref [] in
  prout_string "section rules:\n\n";
  List.iter
    (fun role_instance ->
      if not (List.exists 
		(fun role -> role=role_instance#get_name) 
		!role_list) then
	(role_list:=(role_instance#get_name)::(!role_list);
	 List.iter
	   (function
	       (_,(_, basic_r)) ->
		 let if_var_list = basic_r#get_var_list in
		 prout_string ("step step_");
		 prout_int !stepcounter; incr stepcounter;
		 prout_string " (" ; 
		 print_list term_to_string "," if_var_list; 
		 prout_string ")";
		 prout_string (" :=\n");
		 print_if_states_list ".\n" basic_r#get_lhs;
		 if (basic_r#get_cond <> []) then
		   (prout_string " &\n";
		    print_if_states_list " &\n" basic_r#get_cond);

		 let if_exists_list=basic_r#get_exists_list in
		 if (if_exists_list <> []) then
		   (prout_string "\n=[exists ";
		    print_list term_to_string "," if_exists_list;
		    prout_string "]=>\n")
		 else
		   prout_string "\n =>\n";
		 print_if_states_list ".\n"
		   (if (s_ids<>[]) or (a_ids<>[]) or (wa_ids<>[]) then
		     filter_wanted_goals basic_r#get_rhs
		   else basic_r#get_rhs);
		 prout_string "\n\n")
	 role_instance#get_actions))
    role_instances;
  prout_newline()
;;


(* Verify if the LTL Formula is convertible. To be convertible a
   formula can not cantain other temporal operator than []
   val is_convertible_property : ltl_type -> bool *)
let rec is_convertible_property = function
  | Pred(str) -> true

  | Op1(f) | Op2(f) | Op3(f) | Op4(f) -> false

  (* ~f *)
  | Op5(f)    -> is_convertible_property f

  (* f1 /\ f2 *)
  | Conj(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)

  (* f1 \/ f2 *)
  | Dij(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)

  (* f1 -> f2 *) 
  | Impl(f1,f2) -> (is_convertible_property f1) && (is_convertible_property f2)
;;


(* Convert an LTL formula to a Disjunctive Normal Form (DNF) 
   val ltl_to_dnf : bool -> ltl_type -> ltl_type *)
let ltl_to_neg_dnf f =
  let lists_from_dnf f =

    let rec list_from_dij = function
      | Dij(f1,f2) -> (list_from_dij f1) @ (list_from_dij f2)
      | x -> [x]
    in
    let rec list_from_conj = function
      | Pred(str) -> [str]
      | Conj(f1,f2) -> (list_from_conj f1) @ (list_from_conj f2)
      | x -> failwith "list_from_conj: impossible case"
    in

    List.map list_from_conj (list_from_dij f)
  in

  let rec ltl_to_dnf neg = function
    | Pred(Ifnot(str)) ->
	if neg then
	  Pred(str)
	else
	  Pred(Ifnot(str))

    | Pred(str) ->
	if neg then
	  Pred(Ifnot(str))
	else
	  Pred(str)

	    (* ~f *)
    | Op5(f)    -> 
	ltl_to_dnf (not neg) f

	  (* f1 /\ f2 *)
    | Conj(f1,f2) ->
	if neg then
	  ltl_to_dnf false (Dij(Op5(f1),Op5(f2)))
	else
	  let f1' = ltl_to_dnf false f1
	  and f2' = ltl_to_dnf false f2
	  in
	  (match f1' with
	    Dij(f1a,f1b) ->
	      Dij(ltl_to_dnf false (Conj(f1a,f2')),
		  ltl_to_dnf false (Conj(f1b,f2')))
	  | _ ->
	      (match f2' with
		Dij(f2a,f2b) ->
		  Dij(ltl_to_dnf false (Conj(f1',f2a)),
		      ltl_to_dnf false (Conj(f1',f2b)))
	      | _ ->
		  Conj(f1',f2')))

	    (* f1 \/ f2 *)
    | Dij(f1,f2) ->
	if neg then
	  ltl_to_dnf false (Conj(Op5(f1),Op5(f2)))
	else
	  let f1' = ltl_to_dnf false f1
	  and f2' = ltl_to_dnf false f2
	  in
	  Dij(f1',f2')

	    (* f1 => f2 *)
    | Impl(f1,f2) ->
	ltl_to_dnf neg (Dij(Op5(f1),f2))

    | x -> x (* impossible case *)
  in

  lists_from_dnf (ltl_to_dnf true f)
;;


(*
   \subsection{Printing the goals section}
  \begin{center}
  type:
   [val print_section_goals : Types.goals_type list -> unit]
  \end{center}
*)
let print_section_goals print_previous_sections goals =
  (* filename: name of the output file, in case of goals splitting *)
  let filename = ref "" in
  (* secrecy_info: list of ID of informations used for secrecy *)
  let secrecy_info = ref [] in
  (* auth_info: list of ID of informations used for authentication *)
  let auth_info = ref [] in
  (* wauth_info: list of ID of informations used for weak authentication *)
  let wauth_info = ref [] in
  (* secrecy: number of secrecy goals *)
  let secrecy = ref 0 in
  (* auth: number of authentication goals *)
  let auth = ref 0 in
  (* wauth: number of weak authentication goals *)
  let wauth = ref 0 in
  (* ltl: number of LTL goals *)
  let ltl = ref 0 in

  let print_secret_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state secrecy_of_"^sk^" (MGoal,ASGoal) :=");
    prout_endline " iknows(MGoal).";
    prout_endline (" secret(MGoal,"^sk^",ASGoal) &");
    prout_endline " not(contains(i,ASGoal))";
    prout_newline()
  in
  let print_secret_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-secrecy-"^sk^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [Base(Const(xid))] [] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property secrecy_of_"^sk^" (MGoal,ASGoal) :=");
    prout_endline (" [] ((secret(MGoal,"^sk^",ASGoal) /\\ iknows(MGoal))");
    prout_endline "     => contains(i,ASGoal))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_secret_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_authentication_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID) &");
    prout_endline (" not(witness(A2Goal,A1Goal,"^sk^",MGoal)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_endline ("attack_state replay_protection_on_"^sk^" (A2Goal,A1Goal,MGoal,SID1,SID2) :=");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID1).");
    prout_endline (" request(A1Goal,A2Goal,"^sk^",MGoal,SID2) &");
    prout_endline (" not(equal(SID1,SID2)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_newline()
  in
  let print_authentication_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-auth-"^sk^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [Base(Const(xid))] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID,SID1,SID2) :=");
    prout_endline (" [] (((request(A1Goal,A2Goal,"^sk^",MGoal,SID)");
    prout_endline ("       /\\ ~ equal(A2Goal,i))");
    prout_endline ("      => witness(A2Goal,A1Goal,"^sk^",MGoal))");
    prout_endline ("     /\\ ((request(A1Goal,A2Goal,"^sk^",MGoal,SID1)");
    prout_endline ("           /\\ request(A1Goal,A2Goal,"^sk^",MGoal,SID2)");
    prout_endline ("           /\\ ~ equal(A2Goal,i))");
    prout_endline "          => equal(SID1,SID2)))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_authentication_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_weak_authentication_attack_state xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state weak_authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" wrequest(A1Goal,A2Goal,"^sk^",MGoal,SID) &");
    prout_endline (" not(witness(A2Goal,A1Goal,"^sk^",MGoal)) &");
    prout_endline (" not(equal(A2Goal,i))");
    prout_newline()
  in
  let print_weak_authentication_property xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-wauth-"^sk^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [] [Base(Const(xid))];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property weak_authentication_on_"^sk^" (A1Goal,A2Goal,MGoal,SID) :=");
    prout_endline (" [] ((wrequest(A1Goal,A2Goal,"^sk^",MGoal,SID)");
    prout_endline ("      /\\ ~ equal(A2Goal,i))");
    prout_endline ("     => witness(A2Goal,A1Goal,"^sk^",MGoal))");
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_weak_authentication_attack_state xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_auth_attack_state s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID) :=");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID) &");
    prout_endline (" not(witness("^s2^","^s1^","
		   ^sk^",MGoal)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_endline ("attack_state replay_protection_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID1,SID2) :=");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID1).");
    prout_endline (" request("^s1^","^s2^","
		   ^sk^",MGoal,SID2) &");
    prout_endline (" not(equal(SID1,SID2)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_newline()
  in
  let print_auth_property s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-auth-"^(string_of_int !auth)^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [Base(Const(xid))] [];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID,SID1,SID2) :=");
    prout_endline (" [] (((request("^s1^","^s2^","^sk^",MGoal,SID) /\\ ~ equal(A2Goal,i))");
    prout_endline ("      => witness("^s2^","^s1^","^sk^",MGoal))");
    prout_endline ("     /\\ ((request("^s1^","^s2^","^sk^",MGoal,SID1)");
    prout_endline ("           /\\ request("^s1^","^s2^","^sk^",MGoal,SID2)");
    prout_endline ("           /\\ ~ equal(A2Goal,i))");
    prout_endline "           => equal(SID1,SID2)))";
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_auth_attack_state s1 s2 xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_weak_auth_attack_state s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    prout_endline ("attack_state weak_authentication_on_"^sk^" ("
		   ^s2^","^s1^",MGoal,SID) :=");
    prout_endline (" wrequest("^s1^","^s2^","
		   ^sk^",MGoal,SID) &");
    prout_endline (" not(witness("^s2^","^s1^","
		   ^sk^",MGoal)) &");
    prout_endline (" not(equal("^s2^",i))");
    prout_newline()
  in
  let print_weak_auth_property s1 s2 xid =
    let sk = Globals.string_id#get_name xid in
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-wauth-"^(string_of_int !wauth)^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
       print_previous_sections [] [] [Base(Const(xid))];
       prout_endline "section properties:"; prout_newline ());
    prout_endline ("property weak_authentication_on_"^sk^" ("^s1^","^s2^",MGoal,SID) :=");
    prout_endline (" [] ((wrequest("^s1^","^s2^","^sk^",MGoal,SID)");
    prout_endline ("      /\\ ~ equal(A2Goal,i))");
    prout_endline ("     => witness("^s2^","^s1^","^sk^",MGoal))");
    prout_newline();
    if !Globals.flag_split then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_weak_auth_attack_state s1 s2 xid;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  let print_ltl_attack_states p lvc =
    let las = ltl_to_neg_dnf p in
    let ltl2 = ref 0 in
    List.iter
      (fun s ->
	incr ltl2;
	prout_endline ("attack_state ltl_"^(string_of_int !ltl)^"_"^(string_of_int !ltl2)^" ("
		       ^(list_to_string term_to_string "," lvc)^") :=");
	prout_string " ";
	let (s1,s2) =
	  List.partition
	    (function
		Ifnot(_)
	      | Ifequal(_)
	      | Ifleq(_) -> false
	      | _ -> true)
	    s
	in
	print_list (if_term_to_string "") ".\n " s1;
	if s2 <> [] then
	  (if s1 <> [] then prout_string " &\n ";
	   print_list (if_term_to_string "") " &\n " s2);
	prout_newline())
      las;
    prout_newline()
  in
  let print_ltl_property p lvc =
    if !Globals.flag_split then
      (filename := (!Globals.output_dir)^(!Globals.output_file)^"-ltl-"^(string_of_int !ltl)^".if";
       prerr_endline ("%% IF output in "^(!filename));
       Globals.tmp_channel := open_out !filename;
       prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");
(*** WARNING: all goal facts are printed!  to be changed when real LTL formulas supported? ***)
       print_previous_sections [] [] [];
       prout_endline "section properties:"; prout_newline ());
    prout_string "property ltl_";
    prout_int !ltl;
    prout_endline (" ("^(list_to_string term_to_string "," lvc)^") :=");
    prout_endline (" "^(ltl_formula_to_string p));
    prout_newline();
    if !Globals.flag_split
	&& (is_convertible_property p) then
      (prout_newline (); prout_endline "section attack_states:"; prout_newline ();
       print_ltl_attack_states p lvc;
       Globals.close_out_noerr !Globals.tmp_channel; Globals.tmp_channel:=stdout)
  in

  if not (!Globals.flag_split) then
    (let secrecy_ids = ref [] in
    let auth_ids = ref [] in
    let wauth_ids = ref [] in
    List.iter
      (function
	  Secrecy(lt) -> secrecy_ids:= !secrecy_ids @ (List.map (fun x -> Base(Const x)) lt)
	| Auth(_,_,lt)
	| Authentication(lt) -> auth_ids:= !auth_ids @ (List.map (fun x -> Base (Const x)) lt)
	| Weak_auth(_,_,lt)
	| Weak_authentication(lt) ->  wauth_ids:= !wauth_ids @ (List.map (fun x -> Base (Const x)) lt)
	| LTL_goal(_) -> ())
      goals;
    print_previous_sections !secrecy_ids !auth_ids !wauth_ids;
    prout_endline "section properties:"; prout_newline ());

  (* first pass: for printing properties, and if splitting is required, attack states *)
  List.iter
    (function
	Secrecy(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !secrecy_info)) then
		(secrecy_info := xid::!secrecy_info;
		 incr secrecy;
		 print_secret_property xid))
	    lt

      | Auth(r1,r2,lt) ->
	  let name_r1 =
	    Globals.string_id#get_name (find_player_in_role r1)
	  and name_r2 =
	    Globals.string_id#get_name (find_player_in_role r2)
	  in
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !auth_info)) then
		(auth_info := xid::!auth_info;
		 incr auth;
		 print_auth_property name_r1 name_r2 xid))
	    lt
      | Authentication(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !auth_info)) then
		(auth_info := xid::!auth_info;
		 incr auth;
		 print_authentication_property xid))
	    lt

      | Weak_auth(r1,r2,lt) ->
	  let name_r1 =
	    Globals.string_id#get_name (find_player_in_role r1)
	  and name_r2 =
	    Globals.string_id#get_name (find_player_in_role r2)
	  in
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !wauth_info)) then
		(wauth_info := xid::!wauth_info;
		 incr wauth;
		 print_weak_auth_property name_r1 name_r2 xid))
	    lt
      | Weak_authentication(lt) ->
	  List.iter
	    (fun xid ->
	      if (!Globals.flag_split || not (List.mem xid !wauth_info)) then
		(wauth_info := xid::!wauth_info;
		 incr wauth;
		 print_weak_authentication_property xid))
	    lt

      | LTL_goal(s,lvc) -> 
	  incr ltl;
	  print_ltl_property s (List.filter (function Base(Var(_)) -> true | _ -> false) lvc))
    goals;

  (* second pass: for printing attack states if no splitting required *)
  if (not !Globals.flag_split)
      && (List.exists
	    (function
		LTL_goal(Op1(s),_) -> is_convertible_property s
	      | _ -> true)
	    goals) then
    (prout_newline ();
     prout_endline "section attack_states:";
     prout_newline ();
     secrecy_info := [];
     auth_info := [];
     wauth_info := [];
     ltl := 0;
     List.iter
       (function
	   Secrecy(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !secrecy_info) then
		   (secrecy_info := xid::!secrecy_info;
		    incr secrecy;
		    print_secret_attack_state xid))
	       lt

	 | Auth(r1,r2,lt) ->
	     let name_r1 =
	       Globals.string_id#get_name (find_player_in_role r1)
	     and name_r2 =
	       Globals.string_id#get_name (find_player_in_role r2)
	     in
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !auth_info) then
		   (auth_info := xid::!auth_info;
		    incr auth;
		    print_auth_attack_state name_r1 name_r2 xid))
	       lt
	 | Authentication(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !auth_info) then
		   (auth_info := xid::!auth_info;
		    incr auth;
		    print_authentication_attack_state xid))
	       lt

	 | Weak_auth(r1,r2,lt) ->
	     let name_r1 =
	       Globals.string_id#get_name (find_player_in_role r1)
	     and name_r2 =
	       Globals.string_id#get_name (find_player_in_role r2)
	     in
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !wauth_info) then
		   (wauth_info := xid::!wauth_info;
		    incr wauth;
		    print_weak_auth_attack_state name_r1 name_r2 xid))
	       lt
	 | Weak_authentication(lt) ->
	     List.iter
	       (fun xid ->
		 if not (List.mem xid !wauth_info) then
		   (wauth_info := xid::!wauth_info;
		    incr wauth;
		    print_weak_authentication_attack_state xid))
	       lt

	 | LTL_goal(Op1(s),lvc) -> 
	     incr ltl;
	     if is_convertible_property s then
	       (print_ltl_attack_states s
		  (List.filter (function Base(Var(_)) -> true | _ -> false) lvc))

	 | _ -> ())
       goals)
;; 

(*

   \subsection{Main function}
  
   The [print_if] function is the main Intermediate Format printing function. 
   It displays IF sections according the the command line arguments. 
   By default, this function displays all sections.
  \begin{center}
  type:
   [val print_if : role_instance list -> Types.goals_type list -> unit]
  \end{center}
*)
let print_if role_instances goals =
  if (not (!Globals.flag_types || !Globals.flag_init ||
          !Globals.flag_rules || !Globals.flag_goals)) then
    (Globals.flag_rules:=true; 
     Globals.flag_init:=true; 
     Globals.flag_goals:=true; 
     Globals.flag_types:=true);

  if not (!Globals.flag_stdout || !Globals.flag_split) then
    Globals.tmp_channel := open_out !Globals.tmp_file;
  if not (!Globals.flag_split) then
    prout_endline ("%% IF specification of "^(!Globals.input_file)^"\n");

  if (!Globals.flag_goals && (goals <> [])) then
    print_section_goals
      (fun s a wa ->
	if (!Globals.flag_types) then
	  print_section_types_signatures role_instances;
	if (!Globals.flag_init) then
	  print_section_init role_instances;
	if (!Globals.flag_rules) then
	  print_section_rules role_instances s a wa)
      goals
  else
    (if (!Globals.flag_types) then
      print_section_types_signatures role_instances;
     if (!Globals.flag_init) then
       print_section_init role_instances;
     if (!Globals.flag_rules) then
       print_section_rules role_instances [] [] []);

  if not (!Globals.flag_stdout || !Globals.flag_split) then
    (let filename = (!Globals.output_dir)^(!Globals.output_file)^".if" in
     prerr_endline ("%% IF output in "^filename);
     Globals.close_out_noerr !Globals.tmp_channel);

  (* 
     Indicating the total number of warnings, if any.
  *)
  if (!Error_handler.warningsCounter <> 0) then
    (flush stdout;
     prerr_string "\n%% Number of warnings: "; 
     prerr_int !Error_handler.warningsCounter; 
     prerr_newline();
     flush stderr)
;;
