%{
open Lexing;;
open Types;;
open Globals;;
open Error_handler;;
open Semantic_check;;


let parse_error msg =
  handleParseError();
  exit(-1)
;;

let newSetsInfo = ref [];;
let newSetsInit = ref [];;
let newSetsDecl = ref [];;

let line_number () =
  string_of_int !Error_handler.lineCounter
;;
%}

/* 

   \section{The HLPSL parser}

   \subsection{Declaration of Tokens}
*/

%token <string> VAR_IDENT
%token <string> CONST_IDENT
%token <string> NAT_IDENT

%token ROLE
%token INIT
%token ACCEPT
%token IKNOWLEDGE

%token END
%token LOCAL
%token OWNS
%token CONST
%token CONS
%token DELETE
%token NEW
%token TRANSITION
%token GOAL

%token LOOP
%token AND

%token PRIME
%token UNDERSCORE

%token IMPLIES
%token OR
%token LTL_op1
%token LTL_op2
%token LTL_op3
%token LTL_op4
%token LTL_op5

/*
%token LIST
*/
%token AGENT
%token CHANNEL
%token PUBLIC_KEY
%token TEXT
%token MESSAGE
%token HASH
%token HASH_FUNC
%token NAT
%token PROTOCOL_ID

%token BOOL
%token SYMMETRIC_KEY

%token IN
%token NOT

%token ACTION_ARROW
%token REACTION_ARROW
%token FUNCTION_ARROW

%token L_PAR
%token R_PAR
/*
%token L_HOOK
%token R_HOOK
*/
%token L_BRACE
%token R_BRACE
%token COMMA
%token COLON
%token SEMI_COLON
%token DOT


%token DEF_EQUAL

%token PLUS
%token EQUAL
%token ASSIGN
%token LEQ
%token NEQ

%token PLAYED_BY
%token COMPOSITION
%token SECRECY_OF
%token AUTHENTICATES
%token AUTHENTICATION_ON
%token WEAKLY
%token WEAK_AUTHENTICATION_ON
%token ON
%token DY
%token OTA
%token SET
%token INV
%token START


/* 
   We define the associativity of operators. The order matters.
*/

%left FUNCTION_ARROW
%nonassoc SET /* LIST */
%right IMPLIES
%right AND
%right OR
%right DOT
%nonassoc MSG_COMPOSE
%nonassoc MSG_ENCRYPT


/*  
    Starting point of the grammar, and type of the final parsing result
*/

%start specification         
%type <Types.specification_class> specification
%%


/*
  The specification of a protocol consists in a list of roles. 
  They are followed by a goal declaration and a main protocol call.

*/

specification:
  err01
  Role_definitions
  Maybe_goal_declaration
  Role_instantiation
  { new Types.specification_class $2 $3 $4 }
;


Role_definitions:
  Role_definition
  { Globals.current_role := 0;
    [$1] }
| Role_definition Role_definitions
  { $1::$2 }
;

/*
  A role definition consists in a sequence of declarations. Note that
  it is possible to define syntactically correct, but semantically
  unsound, roles in this way. The exact meaning of each part is
  explained later.
*/
Role_definition:
  Basic_role
  { Types.Basic ($1) }
| Composition_role
  { Types.Composition($1) }
;

Basic_role:
  ROLE Role_header err33
  Player
  Role_declarations err29
  Maybe_transition_declaration
  END ROLE err01
  { let (name, formal_args) = $2 in
    let player = $4 in
    let (local, owns, init, accept, knowledge) = $5 in
    let role = new Types.basic_hlpsl_role in
    role#set_name name;
    role#set_player player;
    role#set_param formal_args;
    role#set_owned owns;
    role#set_local (local@(!newSetsDecl));
    role#set_init (init@(!newSetsInit));
    role#set_accept accept;
    role#set_knowledge knowledge;
    role#set_actions $7;
    Globals.player_to_role:=(player,name)::(!Globals.player_to_role);
    check_player_of_role player name formal_args;
    role
  }
;

Player:
  PLAYED_BY err03 VAR_IDENT
  { Globals.string_id#register_name $3 !Globals.current_role }
;


Composition_role:
  ROLE Role_header
  Role_declarations err30
  Maybe_composition_declaration
  END ROLE err01
  { let (name, formal_args) = $2 in
    let (local, owns, init, accept, knowledge) = $3 in
    let role = new Types.composition_role $5 in
    role#set_name name;
    role#set_param formal_args;
    role#set_owned owns;
    role#set_local (local@(!newSetsDecl));
    role#set_init (init@(!newSetsInit));
    role#set_accept accept;
    role#set_knowledge knowledge;
    role
  }
;


Role_header:
  Role_name err04 L_PAR err202 Maybe_formal_arguments err11 R_PAR 
  { newSetsInfo:=[];
    newSetsInit:=[];
    newSetsDecl:=[];
    ($1,$5) }
;

Role_name:
  err37 CONST_IDENT
  { Globals.current_role := Globals.string_id#register_name $2 0;
    !Globals.current_role
  }
;

/*
  The formal arguments are a list of variables
  declarations. If composed terms are to be given to a role, the
    variables shall be matched either in the [Init] part of the
  specfication or in the left-hand side of a rule. For example, if a
  role [A] is given a certificate from the certification authority [CA]
  for a public key [K], [CA] having public key [KCA], the list of
  parameters of the role will be:
  \begin{center}
  [ROLE alice
  (A:agent,Cert:{agent.public_key}public_key,KCA:public_key,
  K:public_key),CA:agent]\fix{PHD:  This is not currently allowed by the definition of Subtype\_of.  Should this be fixed?}
  \end{center}
  And in the initial part, the user states:
  \begin{center}
  $\ldots\wedge$[Cert={A.K}_inv(KCA)]$\wedge\ldots$
  \end{center}
  
*/
    
Maybe_formal_arguments:
  Formal_arguments
  { $1 }
| { [] }
;

Formal_arguments:
| Variable_declaration err35
  { $1 }
| Variable_declaration err35 COMMA Formal_arguments
  { $1@$4 }
;


Role_declarations:
  err06 DEF_EQUAL
  err13 Maybe_local_declaration
  err12 Maybe_owns_declaration
  err36 Maybe_const_declaration
  err14 Maybe_init_declaration
  err15 Maybe_accept_declaration
  err32 Maybe_iknowledge_declaration
  { ($4,$6,$10,$12,$14) (* WARNING: const declarations not taken into account *) }
;


/*
  The [local] part permits to specify either new
  variables or new constants. First, we need to declare both variables 
  (to designate memory places shared between sessions of an agent, for 
  example (and maybe even synchronous communications)) and constants
  (for nonces and initial knowledge). An identifier is considered to
  be a variable or a constant depending on its first letter.

  The use of constants is rather tricky for the compiler. Consider a
  role [A] with local section:
  \begin{center}
  [LOCAL na : text]
  \end{center}
  and suppose this role is instantiated by:
  \begin{center}
  [LOOP A(...)]
  \end{center}
  Then, for each session of [A], a \emph{new constant} [na] will be
  created. If the user wants to have constants shared over multiple
  role instances, it has to declare them in the [local] part of a
  meta-role, [env] for example, to ensure they will be created only
  once.
*/
Maybe_local_declaration:
  LOCAL Local_declaration
  { $2 }
| { [] }
;

Local_declaration:
  Variables_declaration_list
  { $1 }
;

Variables_declaration_list:
  Variable_declaration err351
  { $1 }
| Variable_declaration err351 COMMA Variables_declaration_list
  { $1@$4 }
;


/*
  I have not fully understood the exact role of the [owns] part. It
  seems it permits to specify that some variables can only be changed
  by this role. However, as long as we do not use double arrows, the
  usefulness of it is obscure. Can someone correct (if) I am wrong ?
*/
Maybe_owns_declaration:
  OWNS Owns_declaration
  { $2 }
| { [] }
;

Owns_declaration:
  Variables_list
  { List.fold_left
      (fun lv var -> 
	try
	  (lv@[new Types.hlpsl_variable var (Globals.type_table#get_type var)])
	with
	  Not_found ->
	    (Semantic_check.add_sem_error
	     (19,
	      Globals.string_id#get_name var,
	      line_number(),
	      Globals.string_id#get_name !Globals.current_role,
	      "","");
	     lv))
      [] $1 
  }
;


Maybe_const_declaration:
  CONST Const_declaration
  {}
| {}
;

Const_declaration:
  Constants_declaration_list
  {}
;

Constants_declaration_list:
  Constant_declaration err352
  {}
| Constant_declaration err352 COMMA Constants_declaration_list
  {}
;


/* 
   The initial state permits to give initial values to variables
   declared in the [local] part of the specification. The value of
   parameters may also be changed, although it does not look like a good
   programming way of doing.
     */
Maybe_init_declaration:
  INIT Init_declarations
  { List.iter
      (fun (xid,l) ->
	let setType =
	  let rec find_applied = function
	      t::lt ->
		(try
		  Set([Utils.term_to_type t])
		with Not_found -> find_applied lt)
	    | [] -> Set([])
	  in
	  find_applied l
	in
	ignore (Globals.type_table#register_type xid setType);
	newSetsDecl := !newSetsDecl @ [new Types.hlpsl_variable xid setType];
	newSetsInit := !newSetsInit @ [Types.Equal(Types.Base(Types.Var(xid)),Set(l))])
      !newSetsInfo;
    newSetsInfo:=[];
    $2 }
| {[]}
;

Init_declarations:
  Init_declaration
  { [$1] }
| Init_declaration AND Init_declarations
  { $1::$3 }
;

Init_declaration:
  VAR_IDENT err243 ASSIGN err312 Stutter_expression 
  { match $5 with
    Types.Base(Types.Var(xid)) ->
      if List.mem_assoc xid !newSetsInfo then
	(let l = List.assoc xid !newSetsInfo in
	newSetsInfo := List.remove_assoc xid !newSetsInfo;
	Types.Equal(Types.Base(Types.Var(Globals.string_id#register_name $1 !Globals.current_role)), Set(l)))
      else
	Types.Equal(Types.Base(Types.Var(Globals.string_id#register_name $1 !Globals.current_role)), $5)
  | _ -> Types.Equal(Types.Base(Types.Var(Globals.string_id#register_name $1 !Globals.current_role)),$5) }
| CONST_IDENT err04 L_PAR err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),[]) }
| CONST_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),$5) }
/*
| AND err07 UNDERSCORE err09 L_BRACE Parameters_instance err08 R_BRACE Init_declaration
  { $9 (* \fix{To be implemented} *) }
*/
;

Parameters_instance :
  Concatenated_variables_list
  { (* \fix{To be implemented} *) 
    (new 
       Types.call_over_declaration 0
         (dummy_val_of_type (Types.Base(Types.Text))))
  }
| IN err04 L_PAR Concatenated_variables_list COMMA err311 Stutter_expression err11 R_PAR
  {
    let string_var = 
      match $7 with
        Types.Base(Types.Var i) -> i
      | _ -> 
	  Error_handler.errorNumber := 20;
	  raise Parse_error()
    in
    new Types.call_over_declaration string_var $4
  }
;

Concatenated_variables_list :
  Concatenated_variables
  { $1 }
| L_PAR Concatenated_variables err11 R_PAR
  { $2 }
;

/* was: VAR_IDENT Var_param */
Concatenated_variables:
  err03 VAR_IDENT
  { Types.Base (Types.Var (Globals.string_id#register_name $2 !Globals.current_role)) }
| err03 VAR_IDENT err19 DOT Concatenated_variables
  { Pair(Types.Base (Types.Var (Globals.string_id#register_name $2 !Globals.current_role)), $5) }
;


/*
  The [ACCEPT] declaration is useful only in the case of roles
  compositions. It permits to detect when a role is over, and thus
  when, during analysis of the protocol, we can go on to the next
  role. 

  A compiler issue about this is to decide whether the role ends once
  it has reached an acceptation state, or if reaching an acceptation
  state give the role a non-deterministic choice to end or not the
  session.  This case probably won't appear in the first modeled
  protocols, and I remember the issue was raised in
  Zurich. Unfortunately, I could not hear what the conclusion was.
*/

Maybe_accept_declaration:
  ACCEPT Accept_declaration
  { $2 }
| { [] }
;
 
Accept_declaration:
  err21 Predicates
  { $2 }
;


/* removed on April 6, 2005
Maybe_knowledge_declaration:
  Knowledge_declaration 
  Maybe_knowledge_declaration
  { $1::$2 }
| { [] }
;
 
Knowledge_declaration:
  KNOWLEDGE err04 L_PAR Variable_or_constant err11 R_PAR
      err242 EQUAL err09 L_BRACE err316 Stutter_expressions_list err08 R_BRACE
  { ($4,$12) }
| KNOWLEDGE err04 L_PAR Variable_or_constant err11 R_PAR
      err242 EQUAL err09 L_BRACE err08 R_BRACE 
  { ($4,[]) }
;
*/

Maybe_iknowledge_declaration:
  IKnowledge_declaration 
  { let newSetKnlg =
      (List.map
	 (fun (xid,l) ->
	   let setType =
	     let rec find_applied = function
		 t::lt ->
		   (try
		     Set([Utils.term_to_type t])
		   with Not_found -> find_applied lt)
	       | [] -> Set([])
	     in
	     find_applied l
	   in
	   ignore (Globals.type_table#register_type xid setType);
	   newSetsInit := !newSetsInit @ [Types.Equal(Types.Base(Types.Var(xid)),Set(l))];
	   Types.Base(Types.Var(xid)))
	 !newSetsInfo)
    in
    newSetsInfo:=[];
    $1 @ newSetKnlg }
| { [] }
;
 
IKnowledge_declaration:
  IKNOWLEDGE err242 EQUAL err09 L_BRACE err316 Stutter_expressions_list err08 R_BRACE
  { $7 }
| IKNOWLEDGE err242 EQUAL err09 L_BRACE err08 R_BRACE
  { [] }
;

/*
  The [transition_declaration] part permits to declare rules changing
  the state of a role. In the preliminar version of the compiler, we
  handle only simple-arrow rules. I also propose, for the compiler
  issue, to not consider the case of multiple rules firing at the same 
  time. Note that this is very problematic wrt TLA modelisation. 
*/
    
Maybe_transition_declaration:
  TRANSITION err16 Transitions_list err02
  { $3 }
| TRANSITION err02
  { [] }	
;

Transitions_list:
  err161 Transition
  { [$2] }
| err161 Transition Transitions_list
  { $2::$3 }
;

Transition:
  Label err19 DOT err21 Predicates err18 ACTION_ARROW err38 Actions
  { let newSetCons =
      (List.map
	 (fun (xid,l) ->
	   let setType =
	     let rec find_applied = function
		 t::lt ->
		   (try
		     Set([Utils.term_to_type t])
		   with Not_found -> find_applied lt)
	       | [] -> Set([])
	     in
	     find_applied l
	   in
	   ignore (Globals.type_table#register_type xid setType);
	   newSetsDecl := !newSetsDecl@[new Types.hlpsl_variable xid setType];
	   newSetsInit := !newSetsInit@[Types.Equal(Types.Base(Types.Var(xid)),Set([]))];
	   let rec makeCons = function
	       [] -> Types.Base(Types.Var(xid))
	     | t::lt ->
		 Types.Cons(t, makeCons lt)
	   in
	   Types.Equal(Types.Prime(Types.Var(xid)), makeCons l))
	 !newSetsInfo)
    in
    newSetsInfo:=[];
    (true, new Types.hlpsl_transition $1 $5 ($9 @ newSetCons)) }
| Label err19 DOT err21 Events err18 REACTION_ARROW err38 Reactions
  { let newSetCons =
      (List.map
	 (fun (xid,l) ->
	   let setType =
	     let rec find_applied = function
		 t::lt ->
		   (try
		     Set([Utils.term_to_type t])
		   with Not_found -> find_applied lt)
	       | [] -> Set([])
	     in
	     find_applied l
	   in
	   ignore (Globals.type_table#register_type xid setType);
	   newSetsDecl := !newSetsDecl@[new Types.hlpsl_variable xid setType];
	   newSetsInit := !newSetsInit@[Types.Equal(Types.Base(Types.Var(xid)),Set([]))];
	   let rec makeCons = function
	       [] -> Types.Base(Types.Var(xid))
	     | t::lt ->
		 Types.Cons(t, makeCons lt)
	   in
	   Types.Equal(Types.Prime(Types.Var(xid)), makeCons l))
	 !newSetsInfo)
    in
    newSetsInfo:=[];
    (false, new Types.hlpsl_transition $1 $5 ($9 @ newSetCons)) }
;

Label:
  CONST_IDENT
  { Globals.string_id#register_name $1 0 }
| NAT_IDENT
  { Globals.string_id#register_name $1 0 }
;


Predicates:
  Predicate err22
  { if !newSetsInfo <> [] then
      (Error_handler.errorNumber := 251;
       raise Parse_error());
    [$1] }
| Predicate err22 AND err21 Predicates
  { $1::$5 }
;

Predicate:
  Stutter_formula
  { $1 }
| CONST_IDENT err04 L_PAR err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),[]) }
| CONST_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),$5) }
| VAR_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Event(Types.Var (Globals.string_id#register_name $1 !Globals.current_role),$5) }
| NOT err04 L_PAR err31 Predicate err11 R_PAR
  { Types.Not($5) }
/*
| START err04 L_PAR err11 R_PAR
  { Globals.global_start_event := true;
    Globals.register_start();
    Types.Event(Types.Const(!Globals.global_start),[]) }
*/
;


Events:
  Predicate err22
  { if !newSetsInfo <> [] then
      (Error_handler.errorNumber := 251;
       raise Parse_error());
    [$1] }
| Event err22
  { if !newSetsInfo <> [] then
      (Error_handler.errorNumber := 251;
       raise Parse_error());
    [$1] }
| Predicate err22 AND err21 Events
  { $1::$5 }
| Event err22 AND err21 Events
  { $1::$5 }
;

Event:
  Non_stutter_formula
  { $1 }
| CONST_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),$5) }
| VAR_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Event(Types.Var (Globals.string_id#register_name $1 !Globals.current_role),$5) }
| VAR_IDENT err04 L_PAR START err11 R_PAR
  { Globals.global_start_const := true;
    Globals.register_start();
    Types.Event
      (Types.Var (Globals.string_id#register_name $1 !Globals.current_role),
       [Types.Base
	  (Types.Const 
	     !Globals.global_start)]) }
| NOT err04 L_PAR err31 Event err11 R_PAR
  { Types.Not($5) }
;


Actions:
  Action err16
  { [$1] }
| Action err16 AND Actions
  { $1::$4 }
;

Reactions:
  Actions
  { $1 }
;

Action:
  VAR_IDENT err23 PRIME err243 ASSIGN err312 Expression
  { 
    Types.Equal
      (Types.Prime
	 (Types.Var
	    (Globals.string_id#register_name $1 !Globals.current_role)),
       $7)
  }
| VAR_IDENT err23 PRIME err243 ASSIGN err312 NEW err23 L_PAR err11 R_PAR
  { 
    Types.New
      (Globals.string_id#register_name $1 !Globals.current_role)
  }
/* TO DO:
| Variable_or_constant err23 PRIME L_PAR err31 Expressions_list err11 R_PAR err243 ASSIGN err312 Expression
*/
| CONST_IDENT err04 L_PAR err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),[]) }
| CONST_IDENT err04 L_PAR err31 Expressions_list err11 R_PAR
  { Types.Event(Types.Const (Globals.string_id#register_name $1 0),$5) }
| VAR_IDENT err04 L_PAR err31 Expressions_list err11 R_PAR
  { Types.Event(Types.Var (Globals.string_id#register_name $1 !Globals.current_role),$5) }
;


Maybe_composition_declaration:
  COMPOSITION err28 Compositions_list err02
  { List.iter
      (fun (xid,l) ->
	let setType =
	  let rec find_applied = function
	      t::lt ->
		(try
		  Set([Utils.term_to_type t])
		with Not_found -> find_applied lt)
	    | [] -> Set([])
	  in
	  find_applied l
	in
	ignore (Globals.type_table#register_type xid setType);
	newSetsDecl := !newSetsDecl@[new Types.hlpsl_variable xid setType];
	newSetsInit := !newSetsInit@[Types.Equal(Types.Base(Types.Var(xid)),Set(l))])
      !newSetsInfo;
    newSetsInfo:=[];
    $3 }
| COMPOSITION err02
  { Types.Empty }
;

Compositions_list:
  Composition err201
  { $1 }
| Composition err201 AND Bracketed_par_compositions_list
  { Types.Parallel($1,$4) }
| Composition err201 SEMI_COLON Bracketed_seq_compositions_list
  { Types.Sequential($1,$4) }
| err04 L_PAR Compositions_list err11 R_PAR
  { $3 }
;

Composition:
  Role_instance
  { $1 }
| AND err07 UNDERSCORE err09 L_BRACE Parameters_instance err08 R_BRACE Bracketed_compositions_list
  {
   Types.MultiPar($6,$9)
     (*
	match $8 with
	  Types.Parallel(r1,r2) ->
	    Types.MultiPar($5,$8)
	| Types.Sequential(r1,r2) ->
	    Types.MultiSeq($5,$8)
	| _ ->
	  Types.MultiPar($5,$8)
      *)
  }
;

Bracketed_par_compositions_list:
  Composition err201
  { $1 }
| Composition err201 AND Bracketed_par_compositions_list
  { Types.Parallel($1,$4) }
| err04 L_PAR Compositions_list err11 R_PAR
  { $3 }
;

Bracketed_seq_compositions_list:
  Composition err201
  { $1 }
| Composition err201 SEMI_COLON Bracketed_seq_compositions_list
  { Types.Sequential($1,$4) }
| err04 L_PAR Compositions_list err11 R_PAR
  { $3 }
;

Bracketed_compositions_list:
  Composition err201
  { $1 }
| err04 L_PAR Compositions_list err11 R_PAR
  { $3 }
;


Role_instance :
  Role_instantiation
  { $1 }
/*| LOOP Role_instantiation
  { Types.Sequential($2,$2) (* \fix{postponed} *) }
| LOOP err04 L_PAR Role_instantiation err11 R_PAR
  { Types.Sequential($4,$4) (* \fix{postponed} *) }*/
;

Role_instantiation:
  err37 CONST_IDENT err04 L_PAR Maybe_effective_arguments err11 R_PAR err201
  { (* \fix{what if no params?} *)
    Types.Call(new Types.call_type (Globals.string_id#register_name $2 0) $5)
  }
;

Maybe_effective_arguments :
  Expressions_list
  { $1 }
| { [] }
;


Maybe_goal_declaration:
  Goal_declaration
  { $1 }
| { [] }
;

Goal_declaration:
  GOAL Goal_formulas_list END GOAL
  { $2 }
;

Goal_formulas_list:
  Goal_formula
  { [$1] }
| Goal_formula Goal_formulas_list
  { $1::$2 }
;

Goal_formula:
/*** OLD goals:
      * We use expressions list in the following because we expect that goals will not only use variables ***
  SECRECY_OF err27 Stutter_expressions_list
  { let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    let agoal = Globals.string_id#register_name "AGoal" 0 in
    let agoal_type = Globals.type_table#register_type agoal (Types.Base Types.Agent) in
    if agoal_type <> (Types.Base Types.Agent) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable AGoal has to be of type agent","","","","");
    Types.Secrecy $3 }
***/
| CONST_IDENT err05 AUTHENTICATES err37 CONST_IDENT ON err37 Constants_list
  { let rid1 = Globals.string_id#register_name $1 0 in
    let rid2 = Globals.string_id#register_name $5 0 in
    let sid1 = Globals.string_id#register_name "SID1" 0 in
    let sid1_type = Globals.type_table#register_type sid1 (Types.Base Types.Nat) in
    if sid1_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID1 has to be of type nat","","","","");
    let sid2 = Globals.string_id#register_name "SID2" 0 in
    let sid2_type = Globals.type_table#register_type sid2 (Types.Base Types.Nat) in
    if sid2_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID2 has to be of type nat","","","","");
    let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    Types.Auth (rid1,rid2,$8) }
| CONST_IDENT err05 WEAKLY AUTHENTICATES err37 CONST_IDENT ON err37 Constants_list
  { let rid1 = Globals.string_id#register_name $1 0 in
    let rid2 = Globals.string_id#register_name $6 0 in
    let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    Types.Weak_auth (rid1,rid2,$9) }

/*** NEW goals ***/
| SECRECY_OF err27 Constants_list
  { let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    let asgoal = Globals.string_id#register_name "ASGoal" 0 in
    let asgoal_type = Globals.type_table#register_type asgoal (Types.Set [Types.Base Types.Agent]) in
    if asgoal_type <> (Types.Set [Types.Base Types.Agent]) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable ASGoal has to be of type agent set","","","","");
    Types.Secrecy $3 }
| AUTHENTICATION_ON err37 Constants_list
  { let a1goal = Globals.string_id#register_name "A1Goal" 0 in
    let a1goal_type = Globals.type_table#register_type a1goal (Types.Base Types.Agent) in
    if a1goal_type <> (Types.Base Types.Agent) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable A1Goal has to be of type agent","","","","");
    let a2goal = Globals.string_id#register_name "A2Goal" 0 in
    let a2goal_type = Globals.type_table#register_type a2goal (Types.Base Types.Agent) in
    if a2goal_type <> (Types.Base Types.Agent) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable A2Goal has to be of type agent","","","","");
    let sid = Globals.string_id#register_name "SID" 0 in
    let sid_type = Globals.type_table#register_type sid (Types.Base Types.Nat) in
    if sid_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID has to be of type nat","","","","");
    let sid1 = Globals.string_id#register_name "SID1" 0 in
    let sid1_type = Globals.type_table#register_type sid1 (Types.Base Types.Nat) in
    if sid1_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID1 has to be of type nat","","","","");
    let sid2 = Globals.string_id#register_name "SID2" 0 in
    let sid2_type = Globals.type_table#register_type sid2 (Types.Base Types.Nat) in
    if sid2_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID2 has to be of type nat","","","","");
    let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    Types.Authentication ($3) }
| WEAK_AUTHENTICATION_ON err37 Constants_list
  { let a1goal = Globals.string_id#register_name "A1Goal" 0 in
    let a1goal_type = Globals.type_table#register_type a1goal (Types.Base Types.Agent) in
    if a1goal_type <> (Types.Base Types.Agent) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable A1Goal has to be of type agent","","","","");
    let a2goal = Globals.string_id#register_name "A2Goal" 0 in
    let a2goal_type = Globals.type_table#register_type a2goal (Types.Base Types.Agent) in
    if a2goal_type <> (Types.Base Types.Agent) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable A2Goal has to be of type agent","","","","");
    let sid = Globals.string_id#register_name "SID" 0 in
    let sid_type = Globals.type_table#register_type sid (Types.Base Types.Nat) in
    if sid_type <> (Types.Base Types.Nat) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable SID has to be of type nat","","","","");
    let mgoal = Globals.string_id#register_name "MGoal" 0 in
    let mgoal_type = Globals.type_table#register_type mgoal (Types.Base Types.Message) in
    if mgoal_type <> (Types.Base Types.Message) then
      Semantic_check.add_sem_error
	(18,"pre-defined variable MGoal has to be of type message","","","","");
    Types.Weak_authentication ($3) }
| LTL_op1 LTL_unary_formula 
  { if !newSetsInfo <> [] then
      (Error_handler.errorNumber := 252;
       raise Parse_error());
    Types.LTL_goal (Op1($2), Utils.vars_consts_from_LTL_goal $2) }
;

LTL_unary_formula :
  LTL_unary_predicate
  {Pred($1)}
| LTL_op2 LTL_unary_formula
  {Op2($2)}
| LTL_op3 LTL_unary_formula
  {Op3($2)}
| LTL_op4 LTL_unary_formula
  {Op4($2)}
| LTL_op5 LTL_unary_formula
  {Op5($2)}
| L_PAR err31 LTL_formula err11 R_PAR
  {$3}
;

LTL_formula :
  LTL_predicate
  {Pred($1)}
| LTL_op2 LTL_unary_formula
  {Op2($2)}
| LTL_op3 LTL_unary_formula
  {Op3($2)}
| LTL_op4 LTL_unary_formula
  {Op4($2)}
| LTL_op5 LTL_unary_formula
  {Op5($2)}
| LTL_formula AND LTL_formula
  {Conj($1,$3)}
| LTL_formula OR LTL_formula
  {Dij($1,$3)}
| LTL_formula IMPLIES LTL_formula
  {Impl($1,$3)}
| L_PAR err31 LTL_formula err11 R_PAR
  {$3}
;

LTL_unary_predicate :
  CONST_IDENT err04 L_PAR err11 R_PAR
  { Types.Unknown(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),[]) }
| CONST_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Unknown(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),$5) }
| IN err04 L_PAR err31 Stutter_expression COMMA err311 Variable_or_constant err11 R_PAR
  { Types.Ifcontains($5,$8) }
| NOT err04 L_PAR LTL_predicate err11 R_PAR
  { Types.Ifnot($4) }
;

LTL_predicate :
  LTL_unary_predicate
  {$1}
| Stutter_expression err241 EQUAL Stutter_expression
  { Types.Ifequal($1,$4) }
| Stutter_expression err241 LEQ Stutter_expression
  { Types.Ifleq($1,$4) }
| Stutter_expression err241 NEQ Stutter_expression
  { Types.Ifnot(Types.Ifequal($1,$4)) }
;


/*
  \section{Auxiliary rules}

  The following rules are often explicit. Thus, no comment is given.
  Feel free to ask for some comments if some points are obscure.
*/

Variable_declaration:
  Variables_list COLON err10 Type_of
  { List.map 
      (fun var ->
        if (Globals.type_table#register_type var $4) <> $4 then
	  (let error_string =
	    ("variable "^(Globals.string_id#get_name var)^
	     " redeclared with incompatible type")
	  in
	  Semantic_check.add_sem_error
	    (16, error_string, line_number(), "","",""));
        new Types.hlpsl_variable var $4)
      $1
  }
;

/* was: VAR_IDENT Var_param err702 */
Variables_list:
  err03 VAR_IDENT err702
  { [Globals.string_id#register_name $2 !Globals.current_role] }
| err03 VAR_IDENT err702 COMMA Variables_list
  { (Globals.string_id#register_name $2 !Globals.current_role)::$5 }
;
/*
Var_param:
  err04 L_PAR err03 VAR_IDENT err11 R_PAR 
  { Globals.string_id#register_name $4 !Globals.current_role }
| { 0 (* \fix{To be handled} *) }
;
*/


Type_of:
  Subtype_of
  { $1 }
| Subtype_of FUNCTION_ARROW Type_of
  { match $3 with
      Types.Function(r,lt) -> Types.Function(r,$1::lt)
    | _ -> Types.Function($3,[$1])
  }
;

Subtype_of:
  Simple_type
  { Types.Base $1 }
| L_PAR Subtype_of err11 R_PAR
  { $2 }
| Subtype_of DOT Subtype_of
  { Types.Pair($1,$3) }
/*
| Subtype_of LIST
  { (*no more supported: Types.Lst [$1] *)
    Error_handler.displayWarning 99 ("line "^(string_of_int !lineCounter)^": Lists not supported; automatically transformed into sets ");
    Types.Set [$1] }
*/
| Subtype_of SET
  { Types.Set [$1] }
| L_BRACE Subtype_of err08 R_BRACE err07 UNDERSCORE Bracketed_subtype_of
  { match $7 with
    Base(Public_key)
  | Inv(Base(Public_key)) ->
      Types.Crypt($2,$7)
  | _ ->
      Types.Scrypt($2,$7) }
| INV err04 L_PAR Subtype_of err11 R_PAR
  { Types.Inv($4) }
| HASH err04 L_PAR Subtype_of err11 R_PAR
  { Types.Hash($4) }
;

Bracketed_subtype_of:
  Simple_type
  { Types.Base $1 }
| INV err04 L_PAR Subtype_of err11 R_PAR
  { Types.Inv($4) }
| L_PAR Subtype_of err11 R_PAR
  { $2 }
;

Constant_declaration:
  Constants_list COLON err10 Simple_type_of
  { List.iter
      (fun cst ->
         if (Globals.type_table#register_type cst $4) <> $4 then
	   let error_string =
	     ("constant "^(Globals.string_id#get_name cst)^
	      " redeclared with incompatible type")
	   in
	   Semantic_check.add_sem_error
	     (16, error_string, line_number(), "","",""))
      $1
  }
;

Constants_list:
  err37 CONST_IDENT err701
  { [Globals.string_id#register_name $2 0] }
| err37 CONST_IDENT err701 COMMA Constants_list
  { (Globals.string_id#register_name $2 0)::$5 }
;


Simple_type_of:
  Simple_subtype_of
  { $1 }
| Simple_subtype_of FUNCTION_ARROW Simple_type_of
  { match $3 with
      Types.Function(r,lt) -> Types.Function(r,$1::lt)
    | _ -> Types.Function($3,[$1])
  }
;

Simple_subtype_of:
  Simple_type
  { Types.Base $1 }
| L_PAR Simple_type_of err11 R_PAR
  { $2 }
;

/*
  The [public_key] type stands for both public and private keys.
  The private key associated to a public key [K] is [inv(K)].
*/

Simple_type:
  AGENT
  { Types.Agent }
| CHANNEL
  { Types.Channel }
| CHANNEL L_PAR err34 DY err11 R_PAR
  { Types.Channel_dy }
| CHANNEL L_PAR err34 OTA err11 R_PAR
  { Types.Channel_ota }
| PUBLIC_KEY
  { Types.Public_key }
| SYMMETRIC_KEY
  { Types.Symmetric_key }
| TEXT
  { Types.Text }
| MESSAGE
  { Types.Message }
| PROTOCOL_ID
  { Types.Protocol_id }
| NAT
  { Types.Nat }     /* may be useful later for inequations */
| BOOL
  { Types.Bool }      /* may be useful later for Boolean properties */
| HASH_FUNC
  { Types.Hash_func }
| L_BRACE Constants_or_nat_list err08 R_BRACE
  { Types.Enum $2 }
;

Constants_or_nat_list:
  CONST_IDENT
  { [Globals.string_id#register_name $1 0] }         /* to be defined (numbers, names,...) */
| NAT_IDENT
  { let name = Globals.string_id#register_name  $1 0 in
    ignore (Globals.type_table#register_type name (Types.Base Types.Nat));
    [name] }
| CONST_IDENT COMMA Constants_or_nat_list
  { (Globals.string_id#register_name $1 0)::$3 }
| NAT_IDENT COMMA Constants_or_nat_list
  { let name = Globals.string_id#register_name  $1 0 in
    ignore (Globals.type_table#register_type name (Types.Base Types.Nat));
    name::$3 }
;


Stutter_formula:
  Stutter_expression err241 EQUAL err312 Stutter_expression
  { Types.Equal($1,$5) }
| Stutter_expression err241 LEQ err313 Stutter_expression
  { Types.Leq($1,$5) }
| IN err04 L_PAR err31 Stutter_expression COMMA err311 Stutter_expression err11 R_PAR
  { Types.In($5,$8) }
| IN err04 L_PAR err31 Non_stutter_expression COMMA err311 Stutter_expression err11 R_PAR
  { Types.In($5,$8) }
| Stutter_expression err241 NEQ err314 Stutter_expression
  { Types.Not(Types.Equal($1,$5)) }
/*
| NOT err04 L_PAR err31 Stutter_formula err11 R_PAR
  { Types.Not($5) }
*/
| L_PAR err31 Stutter_formula err11 R_PAR
  { $3 }
;

Non_stutter_formula:
  Non_stutter_expression err241 EQUAL err312 Stutter_expression
  { Types.Equal($1,$5) }
| Stutter_expression err241 EQUAL err312 Non_stutter_expression
  { Types.Equal($1,$5) }
| Non_stutter_expression err241 EQUAL err312 Non_stutter_expression
  { Types.Equal($1,$5) }
| Non_stutter_expression err241 LEQ err313 Stutter_expression
  { Types.Leq($1,$5) }
| Stutter_expression err241 LEQ err313 Non_stutter_expression
  { Types.Leq($1,$5) }
| Non_stutter_expression err241 LEQ err313 Non_stutter_expression
  { Types.Leq($1,$5) }
| IN err04 L_PAR err31 Non_stutter_expression COMMA err311 Non_stutter_expression err11 R_PAR
  { Types.In($5,$8) }
| IN err04 L_PAR err31 Stutter_expression COMMA err311 Non_stutter_expression err11 R_PAR
  { Types.In($5,$8) }
| Non_stutter_expression err241 NEQ err314 Stutter_expression
  { Types.Not(Types.Equal($1,$5)) }
| Stutter_expression err241 NEQ err314 Non_stutter_expression
  { Types.Not(Types.Equal($1,$5)) }
| Non_stutter_expression err241 NEQ err314 Non_stutter_expression
  { Types.Not(Types.Equal($1,$5)) }
/*
| NOT err04 L_PAR err31 Non_stutter_formula err11 R_PAR
  { Types.Not($5) }
*/
| L_PAR err31 Non_stutter_formula err11 R_PAR
  { $3 }
;


Stutter_expressions_list:
  err27 Stutter_expression
  { [$2] }
| err27 Stutter_expression COMMA err311 Stutter_expressions_list
  { $2::$5 }
;

Non_stutter_expressions_list:
  err27 Non_stutter_expression
  { [$2] }
| err27 Non_stutter_expression COMMA err311 Expressions_list
  { $2::$5 }
| err27 Stutter_expression COMMA err311 Non_stutter_expressions_list
  { $2::$5 }
;

Expressions_list:
  err27 Expression
  { [$2] }
| err27 Expression COMMA err311 Expressions_list
  { $2::$5 }
;


Stutter_expression:
  L_PAR err31 Stutter_expression err11 R_PAR
  { $3 }
| INV err04 L_PAR err31 Stutter_expression err11 R_PAR
  { Types.Inv $5 }
| Stutter_expression DOT err317 Stutter_expression /* %prec MSG_COMPOSE */
  { Types.Pair($1,$4) }
/*
| L_HOOK err26 R_HOOK
  { (*no more supported: Types.Lst [] *
    Error_handler.displayWarning 99 ("line "^(string_of_int !lineCounter)^": Lists not supported; automatically transformed into sets ");
    Types.Set []*)
    Types.Lst [] }
| LTL_op1
  { (*no more supported: Types.Lst [] *
    Error_handler.displayWarning 99 ("line "^(string_of_int !lineCounter)^": Lists not supported; automatically transformed into sets ");
    Types.Set []*)
    Types.Lst [] }
| L_HOOK err27 Stutter_expressions_list err26 R_HOOK
  { (*no more supported: Types.Lst($3) *
    Error_handler.displayWarning 99 ("line "^(string_of_int !lineCounter)^": Lists not supported; automatically transformed into sets ");
    Types.Set($3)*)
    Types.Lst($3)
      (* \fix{TODO: to check the types of all elements.
	 But not valid for function definitions!} *) }
*/
| CONST_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),$5) }
| VAR_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)),$5) }
| Variable_or_constant_or_nat
  { $1 }
| L_BRACE err08 R_BRACE
  {
(*prerr_string "***SET: "; Interface.prerr_term (Types.Set([])); prerr_newline();
     Types.Set([]) *)
   let xid = Globals.string_id#register_blind_var "Set" !Globals.current_role in
   newSetsInfo := !newSetsInfo@[(xid,[])];
   Types.Base(Types.Var(xid)) }
| L_BRACE err316 Stutter_expressions_list err08 R_BRACE err07
  {
(*prerr_string "***SET: "; Interface.prerr_term (Types.Set($3)); prerr_newline();
   Types.Set($3) *)
   let xid = Globals.string_id#register_blind_var "Set" !Globals.current_role in
   newSetsInfo := !newSetsInfo@[(xid,$3)];
   Types.Base(Types.Var(xid)) }
| L_BRACE err316 Stutter_expressions_list err08 R_BRACE err07 UNDERSCORE err315 Bracketed_stutter_expression
  { match $3 with
      [e] ->
	(match $9 with
	  Types.Inv _ ->
 	    Types.Crypt(e,$9)
	| Types.Base (Types.Var i | Types.Const i) ->
	    (try
	      let t =
		(try
		  Globals.type_table#get_type i
		with Not_found ->
		  if !Globals.current_role = 0 then
		    Globals.get_any_type_of i
		  else
		    raise Not_found)
	      in
	      match t with
		Base(Public_key)
	      | Inv(Base(Public_key)) ->
		  Types.Crypt(e,$9)
	      | _ ->
		  Types.Scrypt(e,$9)
	    with Not_found ->
	      (Semantic_check.add_sem_error
		 (16,
		  ("cannot infer type of variable/constant "^(Globals.string_id#get_name i)),
		  line_number (),"","","");
	       Types.Scrypt(e,$9)))
(*TODO: The key may be the result of a function call
   Rmk: this can be avoided by writing:  K'=F(A) and {...}K' *)
	| _ -> Types.Scrypt(e,$9))
    | _ ->
	Error_handler.errorNumber := 901;
	raise Parsing.Parse_error
  }
;

Non_stutter_expression:
  L_PAR err31 Non_stutter_expression err11 R_PAR
  { $3 }
| VAR_IDENT err23 PRIME
  { Types.Prime (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)) }
| INV err04 L_PAR err31 Non_stutter_expression err11 R_PAR
  { Types.Inv $5 }
| Non_stutter_expression DOT err317 Stutter_expression
  { Types.Pair($1,$4) }
| Stutter_expression DOT err317 Non_stutter_expression
  { Types.Pair($1,$4) }
| Non_stutter_expression DOT err317 Non_stutter_expression
  { Types.Pair($1,$4) }
/*
| L_HOOK err27 Non_stutter_expressions_list err26 R_HOOK
  { (*no more supported: Types.Lst($3) *
    Error_handler.displayWarning 99 ("line "^(string_of_int !lineCounter)^": Lists not supported; automatically transformed into sets ");
    Types.Set($3)*)
    Types.Lst($3) }
*/
| CONST_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),$5) }
| VAR_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)),$5) }
| CONS err04 L_PAR err31 Expression COMMA err311 Expression err11 R_PAR
  { Types.Cons($5,$8) }
| DELETE err04 L_PAR err31 Expression COMMA err311 Expression err11 R_PAR
  { Types.Delete($5,$8) }
| L_BRACE err316 Non_stutter_expressions_list err08 R_BRACE err07
  { 
(*prerr_string "***SET: "; Interface.prerr_term (Types.Set($3)); prerr_newline();
    Types.Set($3)  *)
   let xid = Globals.string_id#register_blind_var "Set" !Globals.current_role in
   newSetsInfo := !newSetsInfo@[(xid,$3)];
   Types.Base(Types.Var(xid)) }
| L_BRACE err316 Non_stutter_expressions_list err08 R_BRACE err07 UNDERSCORE err315 Bracketed_expression
  { match $3 with
      [e] ->
	(match $9 with
	  Types.Inv _ ->
 	    Types.Crypt(e,$9)
	| Types.Base (Types.Var i | Types.Const i)    
	| Types.Prime (Types.Var i | Types.Const i)  ->
	    (try
	      match (Globals.type_table#get_type i) with
		Base(Public_key)
	      | Inv(Base(Public_key)) ->
		  Types.Crypt(e,$9)
	      | _ ->
		  Types.Scrypt(e,$9)
	    with Not_found ->
	      (Semantic_check.add_sem_error
		 (16,
		  ("cannot infer type of variable/constant "^(Globals.string_id#get_name i)),
		  line_number (),"","","");
	       Types.Scrypt(e,$9)))
(*TODO: The key may be the result of a function call
   Rmk: this can be avoided by writing:  K'=F(A) and {...}K' *)
	| _ -> Types.Scrypt(e,$9))
    | _ ->
	Error_handler.errorNumber := 901;
	raise Parsing.Parse_error
  }
| L_BRACE err316 Stutter_expressions_list err08 R_BRACE err07 UNDERSCORE err315 Bracketed_non_stutter_expression
  { match $3 with
      [e] ->
	(match $9 with
	  Types.Inv _ ->
 	    Types.Crypt(e,$9)
	| Types.Base (Types.Var i | Types.Const i)    
	| Types.Prime (Types.Var i | Types.Const i)  ->
	    (try
	      match (Globals.type_table#get_type i) with
		Base(Public_key)
	      | Inv(Base(Public_key)) ->
		  Types.Crypt(e,$9)
	      | _ ->
		  Types.Scrypt(e,$9)
	    with Not_found ->
	      (Semantic_check.add_sem_error
		 (16,
		  ("cannot infer type of variable/constant "^(Globals.string_id#get_name i)),
		  line_number (),"","","");
	       Types.Scrypt(e,$9)))
(*TODO: The key may be the result of a function call
   Rmk: this can be avoided by writing:  K'=F(A) and {...}K' *)
	| _ -> Types.Scrypt(e,$9))
    | _ ->
	Error_handler.errorNumber := 901;
	raise Parsing.Parse_error
  }
;

Expression:
  Stutter_expression
  { $1 }
| Non_stutter_expression
  { $1 }
;

Bracketed_stutter_expression:
  INV err04 L_PAR err31 Stutter_expression err11 R_PAR
  { Types.Inv $5 }
| CONST_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),$5) }
| VAR_IDENT err04 L_PAR err31 Stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)),$5) }
| Variable_or_constant_or_nat
  { $1 }
| L_PAR err31 Stutter_expression err11 R_PAR
  { $3 }
;

Bracketed_non_stutter_expression:
  VAR_IDENT err23 PRIME
  { Types.Prime 
      (Types.Var 
	 (Globals.string_id#register_name $1 !Globals.current_role))
  }
| INV err04 L_PAR err31 Non_stutter_expression err11 R_PAR
  { Types.Inv $5 }
| CONST_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Const (Globals.string_id#register_name $1 0)),$5) }
| VAR_IDENT err04 L_PAR err31 Non_stutter_expressions_list err11 R_PAR
  { Types.Function(Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)),$5) }
| L_PAR err31 Non_stutter_expression err11 R_PAR
  { $3 }
;

Bracketed_expression:
  Bracketed_stutter_expression
  { $1 }
| Bracketed_non_stutter_expression
  { $1 }
;


Variable_or_constant:
  VAR_IDENT
  { Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)) }
| CONST_IDENT
  { Types.Base (Types.Const (Globals.string_id#register_name $1 0)) }
;

Variable_or_constant_or_nat:
  VAR_IDENT
  { Types.Base (Types.Var (Globals.string_id#register_name $1 !Globals.current_role)) }
| CONST_IDENT
  { let name = Globals.string_id#register_name $1 0 in
    (try ignore (Globals.type_table#get_type name)
    with _ -> ());
    Types.Base (Types.Const name) }
| NAT_IDENT
  { let name = Globals.string_id#register_name $1 0 in
    ignore (Globals.type_table#register_type name (Types.Base Types.Nat));
    Types.Base (Types.Const name)
  }
;


err01:
  { Error_handler.errorNumber := 1 }
;

err02:
  { Error_handler.errorNumber := 2 }
;

err201:
  { Error_handler.errorNumber := 201 }
;

err03:
  { Error_handler.errorNumber := 3 }
;

err04:
  { Error_handler.errorNumber := 4 }
;

err05:
  { Error_handler.errorNumber := 5 }
;

err06:
  { Error_handler.errorNumber := 6 }
;

err07:
  { Error_handler.errorNumber := 7 }
;

err701:
  { Error_handler.errorNumber := 701 }
;

err702:
  { Error_handler.errorNumber := 702 }
;

err08:
  { Error_handler.errorNumber := 8 }
;

err09:
  { Error_handler.errorNumber := 9 }
;

err10:
  { Error_handler.errorNumber := 10 }
;

err11:
  { Error_handler.errorNumber := 11 }
;

err12:
  { Error_handler.errorNumber := 12 }
;

err13:
  { Error_handler.errorNumber := 13 }
;

err14:
  { Error_handler.errorNumber := 14 }
;

err15:
  { Error_handler.errorNumber := 15 }
;

err16:
  { Error_handler.errorNumber := 16 }
;

err161:
  { Error_handler.errorNumber := 161 }
;

err18:
  { Error_handler.errorNumber := 18 }
;

err19:
  { Error_handler.errorNumber := 19 }
;

err202:
  { Error_handler.errorNumber := 202 }
;

err21:
  { Error_handler.errorNumber := 21 }
;

err22:
  { Error_handler.errorNumber := 22 }
;

err23:
  { Error_handler.errorNumber := 23 }
;

err241:
  { Error_handler.errorNumber := 241 }
;
err242:
  { Error_handler.errorNumber := 242 }
;
err243:
  { Error_handler.errorNumber := 243 }
;
/*
err26:
  { Error_handler.errorNumber := 26 }
;
*/
err27:
  { Error_handler.errorNumber := 27 }
;

err28:
  { Error_handler.errorNumber := 28 }
;

err29:
  { Error_handler.errorNumber := 29 }
;

err30:
  { Error_handler.errorNumber := 30 }
;

err31:
  { Error_handler.errorNumber := 31 }
;

err311:
  { Error_handler.errorNumber := 311 }
;

err312:
  { Error_handler.errorNumber := 312 }
;

err313:
  { Error_handler.errorNumber := 313 }
;

err314:
  { Error_handler.errorNumber := 314 }
;

err315:
  { Error_handler.errorNumber := 315 }
;

err316:
  { Error_handler.errorNumber := 316 }
;

err317:
  { Error_handler.errorNumber := 317 }
;

err32:
  { Error_handler.errorNumber := 32 }
;

err33:
  { Error_handler.errorNumber := 33 }
;

err34:
  { Error_handler.errorNumber := 34 }
;

err35:
  { Error_handler.errorNumber := 35 }
;

err351:
  { Error_handler.errorNumber := 351 }
;

err352:
  { Error_handler.errorNumber := 352 }
;

err36:
  { Error_handler.errorNumber := 36 }
;

err37:
  { Error_handler.errorNumber := 37 }
;

err38:
  { Error_handler.errorNumber := 38 }
;
/*
  We have a small problem since [STRING] currently match anything. We
  propose to enclose the goal in quotes, in order to ease the
  parsing.
*/
