
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche, Thomas Genet and Erwan Houssay

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.5. Last modification 23/09/08
*)
(*
  \section{Contents}

   This  chapter contains definitions of the types
   and classes used in the compiler. 
   
   \section{Overall Architecture of the compiler}

  \subsection{Generalities}
   While all features  are not supported, the internal 
   data structures and compilation strategy can be easily 
   adapted for future versions. As this is a first release,
   several optimization would be possible.
   
   \subsection{Compilation steps}
   
   \begin{enumerate}
   \item\textsf{Parsing:} During this pass, the HLPSL
  file is read, and the internal data structures for each
  role is generated. The result of this pass is a structure 
  of type [hlpsl_specification] containing all pertaining
  information.

  \item\textsf{Semantic check:}
   Here are some of errors that can currently be detected:
   \begin{itemize}
   \item undeclared variables
   \item duplicated variables declaration
   \item intialized identifiers are not variables
   \item type mismatch
   \item missing or unused call arguments
   \item ...
   
   \end{itemize}
   
   The exhaustive list of semantic checks are implemented in
   the "Semantic\_check" module and handled in "Error\_handler".
   
  \item\textsf{Instantiation:} The compiler 
  has to \emph{decorate} each instance of a role by a unique
  [session ID] and has to build a global list of variables
  and constants referenced to in the protocol.

  \item\textsf{States:} During this phase, the compiler
  computes the \emph{accept} and \emph{initial} states
  of each roles.

  \item\textsf{Transitions:} The compiler 
  generates a list of \emph{action} transitions and \emph{reaction}
  transitions. 
   The result of this phase is a pair of action list and reaction
  list.
  \item\textsf{Init:} The compiler generates the initial state
   of the protocol.

  \item\textsf{Output:} The compiler outputs
  the set of rewrite rules representing the protocol instance. 
  It also outputs the other section specified in the IF.
  \end{enumerate}

  
  \section{Global Types Declarations}
 
*)

(*
  \subsection{Replacement of strings}

   While building the identifiers table,
   we replace all string identifiers by naturals.
   
   We use two association tables for faster search.
   When a string is encountered, it should be registered in both tables.
   We associate string identifiers with the roles they were declared in (their scope). In this way,
   we avoid naming conflicts. However, role names or constants are considered as global identifiers,
   so their scope is ignored.
*)

class ident_table =
object(this)
  val mutable ident_number = 1
  (* string_to_id: ((ident_string,role_id),ident_id) *)
  val string_to_id = (Hashtbl.create 50: ((string * int),int) Hashtbl.t)
  (* id_to_string: (ident_id,ident_string) *)
  val id_to_string = (Hashtbl.create 50: (int,string) Hashtbl.t)

  (* 
     getting the [id_to_string] table, used for debugging reasons 
  *)
  method get_tab = id_to_string;

  (* 
     LC: getting the [string_to_id] table, used for debugging reasons 
  *)
  method get_tab_string_to_id = string_to_id;
    
  (* 
     looking for a natural identifier given a string identifier, 
     and registration if not found 
  *)
  method register_name s role_id =
    if (Hashtbl.mem string_to_id (s,role_id)) then
      Hashtbl.find string_to_id (s,role_id)
    else
      let string_id = ident_number in 
      (Hashtbl.add string_to_id (s,role_id) string_id ;
       Hashtbl.add id_to_string string_id s;
       ident_number <- ident_number+1;
       string_id)

   (*
      We have added the method [register_blind_*] if one wants to register 
      a new identifier that has no meaningful name.
    *)
  method register_blind_const name role_id =
    let new_string = (name^"_"^(string_of_int ident_number)) in
      this#register_name new_string role_id

  method register_blind_var name role_id =
    let new_string = (name^"_"^(string_of_int ident_number)) in
      this#register_name new_string role_id


  method get_role_id sid =
    let i = ref 0 in
    Hashtbl.iter
      (fun (s,role_id) sid2 ->
	if sid = sid2 then
	  i := role_id)
      string_to_id;
    !i

     (*
	The converse method, [get_name], may throw an exception.
	The programmar have to add the appropriate exception handling
	when using it.
      *)
  method get_name id =
    Hashtbl.find id_to_string id

  method get_id name =
    let i = ref 0 in
    Hashtbl.iter
      (fun (a,_) id ->
	if (a=name) then i:=id)
      string_to_id;
    !i
end
;;
(*

  \subsection{Global Variables}

  The [global_var_ident_table] class permits to associate to each natural 
   identifier (refered to as "variable" identifier in several sections) a
   unique identifier,  depending on its name and on the \emph{instance} of
  the role it appears in. We refer to this identifier as "global" identifier
*)

(*
  First, let us define an exception for the case of a variable having no 
  value in the table.
*)

exception Variable_has_no_gid of int * int
exception Gid_not_in_table of int

class global_var_ident_table =
object(this)
  val mutable var_number = 0
  val to_id   = (Hashtbl.create 50: ((int*int),int) Hashtbl.t)
  val from_id = (Hashtbl.create 50: (int,(int*int)) Hashtbl.t)
  method get_tab = from_id

  (*
    This method permits to find the variable associated with a 
    global identifier.
  *)
  method get_variable global_id = 
    (try
      Hashtbl.find from_id global_id
    with
      Not_found -> raise (Gid_not_in_table global_id))

  (*
    Given a [global_id], the following methods permit to find
    the name of a variable (more precisely, the key to the name in
    the [string_id] table and the session number of the role the
    variable is declared in.
  *)  
  method get_name_of global_id =
    snd (this#get_variable global_id)
  method get_role_instance_of global_id =
    fst (this#get_variable global_id)
  
  method unsafe_get_id_of role_id string_id =
    (try
      Hashtbl.find to_id (role_id,string_id)
    with
      Not_found -> raise (Variable_has_no_gid(role_id,string_id)))

  (*
    The following method permits to register a variable found in a
    role, and later to find its global identifier.
  *)
  method get_id_of role_id string_id =
    try
      Hashtbl.find to_id (role_id,string_id)
    with 
	Not_found -> 
	  (let var_num = var_number in
	  Hashtbl.add to_id (role_id,string_id) var_num ;
	  Hashtbl.add from_id var_num (role_id,string_id);
	  var_number <- var_number+1 ;
	  var_num)
end
;;

(*
  \subsection{Locating variables}
   We define an intermediate table, called [memory_map] which associates the global ids
   to their IF values.
*)


(*
  First, let us define an exception for when trying to get the [mmap_id]
  of a variable that does not currently have a value in the table.
*)

exception Variable_has_no_mmap of int


class memory_map_table =
object(this)
  val mutable map_number = 0
  val mem_map = (Hashtbl.create 50: (int,int) Hashtbl.t)
  method get_tab = mem_map

  (*
    The next method permits to know the index in "memory" of a variable.
    A [Not_found] exception is raised if the map does not exist.
  *)
  method get_map global_id = 
    (try
      Hashtbl.find mem_map global_id
    with
      Not_found -> raise (Variable_has_no_mmap global_id))

  (*
    This method can be  used for new variables, such as those in the
    [local] part of the declaration of a role, those created
    when evaluating the composition over a set, and those created
    when passing parameters.
  *)
  method map_new_var global_id =
    (map_number <- map_number + 1;
     Hashtbl.add mem_map global_id map_number;
     map_number)

  (*
    The next method checks if a variable is already in the memory, and
    if not allocates a new place for it.
  *)
  method check_map_var global_id =
    if not (Hashtbl.mem mem_map global_id) then
      (map_number <- map_number + 1;
       Hashtbl.add mem_map global_id map_number)

  (*
    This method is used when we want to map a newly encountered
    variable to the place where an old variable already is.
    This happens, for example, in the case of parameters if they are
    passed by reference.
  *)
  method map_var_to new_var_id old_var_id =
    let map_result = this#get_map old_var_id in
      Hashtbl.add mem_map new_var_id map_result
end
;;


(*
  \subsection{Values of variables}

  It is important to store, during the compilation process, the value
  of the variables. This can be used, for example, when instances
  are created over a list or a set. First, we define what can be a \emph{value}
  with regard to the IF.

  \subsubsection{Values}
  
  We define the type [if_value_type] to denote the values of terms. Note that the 
  integer parameter does not have the same meaning if we consider an address or
  a constant.
*)

type if_value_type =
    
    (*
      In the case of an address, the [int] is a key in the 
      [global_var_id] table. There is no constructor for variables. The
      [Address] or [Const] are used, for example, during the call of a role.
      I suggest the following way of choosing between call by value or
      call by reference. 

      If the role [Alice], declared as [role Alice(A: agent)], is instantiated
      with:
      \begin{center}
      \parbox{0.6\linewidth}{[Alice(a)]}
      \end{center}
      Then the variable [A] will have its own location, and the value [a].
      On the other hand, if the calling sequence is something like:
      \begin{center}
      \parbox{0.6\linewidth}{%
      Local B : agent\\
      Init   B = a\\
      Alice(B)
      }
      \end{center}
      Then the \emph{value} of [A] will be [Address(n)], where $n$ is the
      global identifier of [B].
    *)

  | If_address of int

    (*
      In contrast, the [int] parameter of a [Const] is a string identifier.
    *)
 
  | If_const of int

    (*
      The value of the following constructors is syntactically defined. 
      This is not the case for functions, whose value is a variable.
      The [int] refers to a value in the [global_variable_id] table.
    *) 
  | If_crypt of if_value_type * if_value_type
  | If_scrypt of if_value_type * if_value_type
  | If_pair of if_value_type * if_value_type
  | If_cons of if_value_type * if_value_type
  | If_delete of if_value_type * if_value_type
  | If_inv of if_value_type
  | If_set of if_value_type list
(*
  | If_lst of if_value_type list
*)
    (*
      The first argument of the [Function] constructor must
      be of type ['a -> 'b], the second argument should be of type
      ['a], and the third one should be of type ['b].

      CORRECTION: With the following declaration, the [If_function] 
      constructor takes as input one [If_value_type], presumably 
      the function name, and a list of arguments. It therefore is suited 
      for the \emph{hash functions}. 
    *)
  | If_function of if_value_type * if_value_type list
;;
(*
  We can now define the [value_map_table] which holds the value (for constants)
  or the address of variables. This table maps global identifiers to [if_value].
*)

class value_map_table =
object
  val value_map = (Hashtbl.create 50: (int,if_value_type) Hashtbl.t)
  method get_tab = value_map
  method get_value mem_id = Hashtbl.find value_map mem_id
  method is_defined mem_id = Hashtbl.mem value_map mem_id
  method set_value mem_id t =
    Hashtbl.add value_map mem_id t
end
;;
(*
  \subsection{Types for Messages}
*)
(*
  \subsubsection{Terms-related Types}

  We first define the type [term_structure] to handle
  term constructors. This type is then instantiated
  by either [atoms_type], for terms, or [basic_type],
  for terms type.
*)


type 't term_structure =
    Pair of 't term_structure * 't term_structure
  | Scrypt of 't term_structure * 't term_structure
  | Crypt of 't term_structure * 't term_structure
  | Inv of 't term_structure
  | Function of 't term_structure * ('t term_structure list)
  | Cons of 't term_structure * 't term_structure
  | Delete of 't term_structure * 't term_structure
  | Hash of 't term_structure
  | Set of 't term_structure list
(*
  | Lst of 't term_structure list
*)
  | Prime of 't
  | Base of 't
;;

type basic_type =
    Agent
  | Channel 
  | Channel_dy
  | Channel_ota
  | Text
  | Message
  | Public_key 
  | Symmetric_key
  | Protocol_id
  | Bool
  | Nat
  | Hash_func
  | Enum of int list 
(*  | Pointer*)
;;


type atoms_type =
    Var of int
  | Const of int
;;

(*
  The types used to handle terms and term types are
  now given as instantiation of the [term_structure]
  type.
*)

type term = 
  atoms_type term_structure
;;

type term_type = 
  basic_type term_structure
;;


(*
  We define now a structure mapping each variable or constant to a type. 
   The follwing types table is built before the roles instantiation
   because it is used during the semantic check.
*)
class type_table_type =
object(this)
  val tab = (Hashtbl.create 50: (int,term_type) Hashtbl.t)
  method get_tab = tab
  method register_type id tp =
    if (Hashtbl.mem tab id) then
      Hashtbl.find tab id
    else
      (Hashtbl.add tab id tp ;
       tp)
  method get_type var_id = Hashtbl.find tab var_id
end

(*
   \subsection{HLPSL Variables}
*)
(*
  We define a class to represent HLPSL variables as it is more
  natural, in some cases, to write:
  \begin{center}
  [let x = new hlpsl_variable id typ]
  \end{center}
  \label{types:hlpsl:variable}
*)

class hlpsl_variable ~id:s ~htyp:v =
object
  val mutable id = (s:int)
  val mutable t = (v:term_type)
  method set_type u = t <- u
  method set_id my_id =
    id <- my_id
  method get_id = id
  method get_type = t
end
;;

(*
  \subsection{Terms in the IF syntax}

*)
type if_term = 
  (*
    knowledge of the intruder
   *)
  | Iknows of term
 (*
    goal terms
  *
  | Secret of term * term
  | Witness of int * int * term * term *)

  (*
    States of principals: first, the name of the role, then the identifier
    of the session (unique), and finally a list of terms. Note that using
    a \termif{state} constructor in the IF is a bad idea, as it would not
    be of constant arity.
   *)

  | State of string * int * term list

  (*
    Now, suppose that [PK(A)] stands for the public key of [A], and may
    change over time. In other words, we have a function whose value
    can be changed pointwise.
    We define a \termif{func} predicate of arity 3. For example,
    $\termif{func}(a,b,c)$ stands for the function whose name is stored
    at address $a$ and whose argument is stored at address $b$.
    The value of the function is stored at address $c$.
   *)

  | Func of int * int * int  (* not used!!! *)

  | Ifcontains of term * term
      
  | Ifdelete of term * term
      
  | Ifnot of if_term
      
  | Ifequal of term * term

  | Ifleq of term * term

  | Ifnew of int

  (* Event: unknown term, copied directly *)
  | Unknown of term * term list
;;

(*
  We define an [if_state] type as a list of [if_term]s.
  Note that an [if_state] can either be the state of an action
  or a reaction rule.
*)
type if_state = 
  if_term list
;;

(*
   We define the structure that is used for IF protocol rules
*)

class basic_if_rule 
  (l:if_state)
  (r:if_state) 
  (c:if_state)
  (vl:term list)
  (el: atoms_type list)
  = 
object
  val mutable lhs = l
  val mutable rhs = r
  val mutable cond = c
  val mutable var_list = vl
  val mutable exists_list = el

  method get_lhs = lhs
  method get_rhs = rhs
  method get_cond = cond
  method set_lhs l = lhs <- l
  method set_rhs r = rhs <- r
  method set_cond c = cond <- c
  method get_var_list = var_list
  method set_var_list vl = var_list <- vl
  method get_exists_list = List.map (fun e-> Base e) exists_list
  method set_exists_list el = exists_list <- el

end
;;

(*
  The first [int] is the [string identifier] of the role,
  and the second [int] is the label of this rule in the role.
*)

type if_rule = 
    int * (int * basic_if_rule)
;;

(*
  \section{Parsing related types}
*)

(*
  With the following grammar, a [Not(Event(a,b))] 
  predicate is possible. However, this case is avoided
  during the parsing of the HLPSL specification.
*)
type predicate =
    Not of predicate 
  | Equal of term * term
  | Leq of term * term
  | In of term * term
  | Event of atoms_type * (term list)
  | New of int
;;

(*
  We use the same structure for state predicates 
  and transition predicates. This means that, during 
  the parsing, and in case of a transition predicate, we
   semantically check that it contains at least one event.
*)
type hlpsl_state = 
    predicate list
;;

(*
  We use the same data structures for action rules and
  reaction rules. This means that we have to store
  separately the two kinds of rules in a basic role. 
*)

class hlpsl_transition ~name:s ~lhs:l ~rhs:r =
object
  val mutable label = (s:int)
  val mutable lhs = (l:hlpsl_state)
  val mutable rhs = (r:hlpsl_state)
  method set_label my_id =
    label <- my_id
  method get_label = label
  method get_lhs = lhs
  method set_lhs l = lhs <- l
  method get_rhs = rhs
  method set_rhs r = rhs <- r
end
;;

(*

  \subsubsection{Generalities}

  We have separated the roles in two type of roles. The 
  \emph{Basic Roles} are those in which transitions are
  defined and that can be \emph{played} by an agent. 
   The \emph{Composition Roles} are used to combine the basic
  roles to design complex scenarii of execution. Since these
  roles may combine sessions between agent as well as sessions
  of a single agent, the [player] item may have no meaning, and
  is thus not available.
*)

(*
  \subsubsection{An Example}

  In order to be more accurate on the meaning of each field 
  of the [role] type, we give, as example, a HLPSL
  specification of a basic role. We will refer to this example
  during the explanation of the field. 

  \begin{example}{Basic Role Specification}
  \begin{verbatim}
  role alice (A,B: agent,
              Ka,Kb: public_key,
              L: text set,
              SND,RCV : channel(dy),
              SID: text) played_by A def=
  local 
      St: {0,1,2}, Na,Nb: text
  init
      St:=0
  trans
      St=0 /\ RCV(start) =>
      St':=1 /\ Na':= new() /\ L':=cons(Na',L) /\ SND({Na'.A}_Kb)
  end role
  \end{verbatim}
  \end{example}

*)

(*
  \subsubsection{Generic HLPSL roles}

   We first declare a new class of roles called [generic_hlpsl_role].
  This class wil then be inherited by the [basic_hlpsl_role] and 
  [composition_hlpsl_role]. The aim is to have a type that permits to check
  that composition roles and basic roles are well defined.
*)

class generic_hlpsl_role =
object
  (*
    The [name] corresponds to the name of the role, \emph{i.e.}
    [Alice], in the example.
  *)
  val mutable name = 0
  method set_name my_id = name <- my_id
  method get_name = name

  (*
    The [param] field contains the declaration in parenthesis
    as a list of typed variables
  *)
  val mutable param = ([]: hlpsl_variable list)
  method get_param = param
  method set_param l = param <- l
    
  (*
    The list of variables in the owned declaration.
  *)
  val mutable owned = ([]: hlpsl_variable list)
  method get_owned = owned
   method set_owned l = owned <- l

(* 
   The [local] field contains the list of variables, together
   with their type, declared in the [local] section.
*)
  val mutable local = ([]: hlpsl_variable list) 
  method get_local = local
  method set_local l = local <- l  

(*
  The [init] field contains the initial state of the agent 
  playing the role as a list of predicates.
*)
  val mutable init_state = ([]: predicate list)
  method get_init = init_state
  method set_init l = init_state <- l  

(*
  The [accept] field contains the acceptation state
  as a list of predicates. This section should be
  mandatory. Put, in the example, 
  \begin{center}
  accept   St=2
  \end{center}
*)
  val mutable accept_state = ([]: predicate list)
  method get_accept = accept_state
  method set_accept l = accept_state <- l  

 (*
   The [knowledge] section contains the list of knowledge
   of each given agent
 *)
(* changed on April 6, 2005 (only intruder knowledge)
  val mutable knowledge = ([]: (term * term list) list) *)
  val mutable knowledge = ([]: term list)
  method get_knowledge = knowledge
  method set_knowledge l = knowledge <- l  

end
;;


(*
  \subsubsection{Basic roles}

*)
class basic_hlpsl_role = 
object 
  inherit generic_hlpsl_role

  (*
    In the example, the player is [A]. The corresponding string
    identifier is put in the [player] field.
  *)
  val mutable player = 0

  method get_player = player
  method set_player a = player <- a

(*
  The following lists contain the transitions of the
  role.  If the Boolean is true, the transition is an action,
  otherwise it is a reaction.
*)
  val mutable actions = ([]: (bool * hlpsl_transition) list)

  method add_action a = actions <- [a]@actions
  method set_actions l = actions <- l
  method get_actions = actions
end
;;

(*
  \subsubsection{Composition Roles}

  These roles are used to define scenarii, either between
  different agents, or an agent doing several sub-protocols.
*)

(*
  Before we define the [composition_role] type, we define
  the types handling a composition. Basically, we first need
  to define a type to describe the call of a role with a list
  of arguments.
*)

class call_type ~id ~args_list =
object
  val mutable role = (id:int)
  val mutable args = (args_list: (atoms_type term_structure) list)

  method set_role a = role <- a
  method get_role = role
  method set_args a = args <- a
  method get_args = args
end;;

(*
  In the [MultiPar] or [MultiSeq] constructor of a composition
  declaration, a list of parameters is given.
*)

class call_over_declaration variable te =
object

  (*
    The [term] should have variable identifiers that
    are relevant in the [string_id] table. These variables
    are not relevant now, they are only relevant once
    the instance is known.
  *)
  val mutable t = (te:term)

  method get_term = t
  method set_term u = t <- u
  
  (*
    The string must refer to a list or set of variables.
  *)
  val mutable set = (variable:int)

  method get_set = set
  method set_set a = set <- a
end
;;

(*
  We now define how roles can be composed. We first use a
  parametric type, as the declaration may be re-used for the 
  roles in phase 3 of the compilation.
*)

type 't composition_of =
    Call of 't
  | MultiPar of call_over_declaration * 't composition_of
(*  | MultiSeq of call_over_declaration * 't composition_of *)
  | Sequential of 't composition_of * 't composition_of
  | Parallel of 't composition_of * 't composition_of
  | Empty
;;


type composition_type = 
    call_type composition_of
;;


class composition_role (composition:composition_type) = 
object
  inherit generic_hlpsl_role
      
  val mutable composition_relation = composition

  method set_composition = composition_relation <- composition
  method get_composition = composition_relation
end
;;

type role_type =
    Basic of basic_hlpsl_role
  | Composition of composition_role
;;

(*
  \section{Instantiation related types}

*)

(*
  \subsection{Roles during instantiation}

  We now define the data structures to use during
  the translation from internal data structures to IF.
   We should be able to:
  \begin{itemize}
  \item retrieve the variables location from the
  roles calls and parameters instantiation;
  \item give a unique session ID to each role. This
  ID changes during, \emph{e.g.} a loop statement. 
  \end{itemize}
  We consider that all variables belong to a global memory. 
   The location of a variable in this memory is given by a unique
   identifier (an integer, for example). 

   We consider the case of a finite number of instances. 
   In this case, the compiler can explicitely give a session ID to each
   instance of a role. We put in the [state] term the name of
   the role, its session ID, and the list of identifiers, in the
   global memory, of the variables declared in this role.
*)


(*
  \subsection{General types}
*)

type ltl_type =
    Pred of if_term
  | Op1 of ltl_type
  | Op2 of ltl_type
  | Op3 of ltl_type
  | Op4 of ltl_type
  | Op5 of ltl_type
  | Conj of ltl_type * ltl_type
  | Dij of ltl_type * ltl_type
  | Impl of ltl_type * ltl_type
;;

(* We define the main type for goals handling *)
type goals_type =
(*    Secrecy of term list*)
    Secrecy of int list
  | Authentication of int list
  | Weak_authentication of int list
  | Auth of int * int * int list
  | Weak_auth of int * int * int list
    (* LTL goal: LTL formula + list of used variables/constants *)
  | LTL_goal of (ltl_type * term list)
;;

(* 
   Here is the specification class that will contain all the 
   parsed HLPSL information
*)
class specification_class roles_list goals main =
object
  val mutable role_definitions_list = (roles_list:role_type list)
  val mutable goals = (goals:goals_type list)
  val mutable main_role = (main:composition_type)
      
  method set_roles lst =
    role_definitions_list = lst
  method get_roles =
    role_definitions_list
  method get_main_role =
    main_role
  method get_goals =
    goals
end
;;


(*
  It is not possible to define a class ['c] depending on a type ['a] depending 
  on the class ['c]. Therefore, we first define a parameterized type 
  [_type_of_composition]
  and a parameterized class [_role_instance]. Then, we instantiate
  the parameters at the same time in a recursive type definition.
  The resulting \emph{types} are what we want. The only problem is
  that [role_instance] is no longer a \emph{class}, but a 
  \emph{type}. As a result, we cannot use the [new] function.
*)
type 't _type_of_composition = 
    If_parallel of 't _type_of_composition * 't _type_of_composition
  | If_sequential of 't _type_of_composition * 't _type_of_composition
  | If_call of 't 
  | If_empty
;;


(*
  When creating a [role_instance] object, we give as parameters
  the name of the role (\emph{e.g.} Alice) and the list of 
  variables instances for the parameters. This information is
  stored in a [call_type] object. We also have to give the
  role structure.
*)
class role_instance =
  let n = ref 0 in
  fun (call: call_type) ->
    object(this: 's)

  (*
    The [name] is the [id] of the name of the role (\emph{e.g.} 
    Alice) the structure is an instance of. It is set correctly in
    the [initializer] part of the role.
  *)
  val mutable name = 0

  method set_name a = name <- a
  method get_name = name

  (*
    The session ID field is used to represent unambiguously each
    instance of a role. One possible use of this representation is
    the creation of nonces or the authenticate goals. The value
    of [session_Id] is set correctly in the [initializer] part of the
    role.
  *)
  val mutable session_Id = 0
  method set_session_Id a = session_Id <- a
  method get_session_Id = session_Id

  (*
    The following list is the list of global variable
    identifiers for this role.
  *)
  val mutable variable_instances = 
    ([]: int list)

  method set_variable_instances l = variable_instances <- l
  method get_variable_instances = variable_instances

  (*
    The initial and accept state of a role are used to generate
    transitions for composition roles. They thus have to be generated
    \emph{before} the generation of transition rules.
  *)
  val mutable initial_state = ([]: if_state)

  method get_initial_state = initial_state
  method set_initial_state a = initial_state <- a

  val mutable accept_state = ([]: if_state)

  method get_accept_state = accept_state
  method set_accept_state a = accept_state <- a

  (*
    The two following fields store the list of actions and reactions
    in the IF format (\emph{i.e} as terms). Note that this is not the final
    step before printing, since the compiler should first pre-compile
     reactions in order to have only actions to output in the IF.\fix{To be implemented}
  *)
  val mutable actions   = ([]: if_rule list)

  method get_actions = actions
  method set_actions a = actions <- a
  method add_to_actions a = actions <- [a]@actions

  (*
    Finally, for composition roles, we define the sub-roles.
   *)
  val mutable composition = (If_empty: 's _type_of_composition)

  method set_composition (a:'s _type_of_composition) = composition <- a
  method get_composition = composition

  initializer 
  (*
    [role_id] is an integer whose value is the key to find the
    role name in the [string_id] table or the role internal data
    structure in the role instances list
  *)
    let role_id = call#get_role in
    (name <- role_id;
     session_Id <- !n; incr n)
end

type type_of_composition = 
    role_instance _type_of_composition
;;


type composition_schema =
  | Par of composition_schema list
  | Seq of composition_schema list
  | Inst of int
  | None
;;
